/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.algorithms;

import ikr.simlib.algorithms.queuingDisciplines.StackableQueuingDiscipline;
import ikr.simlib.messages.Message;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable2;

import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;

/**
 * {@link MessageClassificator} is used to manage a set of handlers for
 * aggregates of packets, i.e. flows, and their mapping to these flows.
 * {@link MessageClassificator}s are used to identify the proper sub-
 * {@link StackableQueuingDiscipline} in hierarchical queuing disciplines. Any
 * class implementing this interface MUST implement a constructor taking
 * {@link Parameters}, {@link SimNode} and {@link String} as parameters to allow
 * for dynamic configuration. Most implementations will be part of the IKR
 * ProtocolSupport library since they are relevant for DataMessages only. A
 * {@link MessageClassificator} never creates new Consumers.
 * 
 * @author dwagner
 */
public interface MessageClassificator<Key, Consumer> extends ReflectionConstructable2<SimNode, Parameters> {

   /**
    * This method returns the Consumer object for the given {@link Message}
    * according to its classification mapping. If such Consumer, e.g. a
    * {@link StackableQueuingDiscipline}, does not exist, it returns null. It
    * does not create new Consumers.
    */
   public abstract Consumer getConsumer(Message msg);

   /**
    * This method maps a consumer c to the flow the {@link Message} msg belongs
    * to, according to the {@link MessageClassificator}'s internal mapping
    * algorithms. Unfortunately, Java does not allow to define derived Classes
    * to expose methods accepting the Generic Parameter, so implementations will
    * accept {@link Object}s here and have to cast them.
    */
   public abstract void insertConsumer(Message msg, Consumer c);

   public abstract void removeConsumer(Message msg);

   /**
    * This method returns a set if the consumers existing in / managed by this
    * {@link MessageClassificator}. <br>
    * BE AWARE that this may return a collection backed by the used data
    * structure itself, so changes in the returned {@link Collection} result in
    * changes in the {@link MessageClassificator} itself.<br>
    * Its main purpose is to deliver an {@link Iterator} to access the stored
    * consumer for tracing. Therefore, any implementation should return that
    * {@link Collection} in the order of insertion. This is the case for
    * {@link LinkedHashMap}s for example.
    */
   public abstract Collection<Consumer> getConsumers();

   /**
    * This method shall give access to the key-value pairs to allow for
    * manipulation. The set is backed by the internal data structure, so changes
    * to the map are reflected in the set, and vice-versa.
    */
   public abstract Set<Map.Entry<Key, Consumer>> getEntrySet();

}
