/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.algorithms.queuingDisciplines;

import ikr.simlib.messages.Message;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;

/**
 * This class implements a bounded FIFO (First In First Out) queuing discipline,
 * i.e., it has a finite buffer size. If a new message arrives and there is not
 * enough buffer space left in the queue to store the message, the arriving
 * message is lost, i.e., it is deleted by the HandleLossMessage()-method. You
 * can overwrite this method if you want something else to be done with the
 * message.
 */
public class BoundedFIFOQDisc extends FIFOQDisc implements BoundedQDisc {

   protected int maxSizeInUnits;

   /** as required by {@link ReflectionConstructable} */
   public static BoundedFIFOQDisc createInstance(SimNode ownNode, Parameters pars) {
      return new BoundedFIFOQDisc(ownNode, pars);
   }

   public static BoundedFIFOQDisc createInstance(SimNode ownNode, @SuppressWarnings("unused") Parameters pars, Integer size) {
      return new BoundedFIFOQDisc(ownNode, size);
   }
   
   public BoundedFIFOQDisc(SimNode ownNode, int size) {
      super(ownNode);
      this.maxSizeInUnits = size;
   }

   public BoundedFIFOQDisc(SimNode ownNode, Parameters pars) {
      this(ownNode, pars.get(ownNode, "MaxSizeInUnits").asInteger());
   }

   @Override
   public int getMaxSizeInUnits() {
      return maxSizeInUnits;
   }

   /**
    * return true if message fits into maxSizeInUnits, else false
    */
   @Override
   public boolean hasSpaceFor(Message msg) {
      return msg.getLength() <= (maxSizeInUnits - getCurrentNumberOfUnits());
   }
}
