/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.algorithms.queuingDisciplines;

import ikr.simlib.messages.Message;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;

/**
 * This class implements a bounded LIFO (Last In First Out) queuing discipline,
 * i.e., it has a finite buffer size. If a new message arrives and there is not
 * enough buffer space left in the queue to store the message, messages are
 * deleted from the head of the queue until there is enough buffer space in the
 * drop-head queue to hold the new message. The messages are deleted by the
 * HandleLossMessage()-method. You can overwrite this virtual method if you want
 * something else to be done with the message. If the arriving message is larger
 * than the maximum available buffer space in the queue, then it is discarded
 * immediately and no messages within the queue are deleted.
 */
public class BoundedLIFOQDisc extends LIFOQDisc implements BoundedQDisc {

   protected int maxSizeInUnits;

   /** as required by {@link ReflectionConstructable} */
   public static BoundedLIFOQDisc createInstance(SimNode ownNode, Parameters pars) {
      return new BoundedLIFOQDisc(ownNode, pars);
   }

   public static BoundedLIFOQDisc createInstance(SimNode ownNode, Parameters pars, Integer size) {
      return new BoundedLIFOQDisc(ownNode, size);
   }

   public BoundedLIFOQDisc(SimNode ownNode, int size) {
      super(ownNode);
      this.maxSizeInUnits = size;
   }

   public BoundedLIFOQDisc(SimNode ownNode, Parameters pars) {
      this(ownNode,pars.get(ownNode, "MaxSizeInUnits").asInteger());
   }

   @Override
   public boolean wouldStore(Message msg) {
      return msg.getLength() < (maxSizeInUnits - getCurrentNumberOfUnits());
   }

}
