/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.algorithms.queuingDisciplines;

import ikr.simlib.messages.Message;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;

import java.util.LinkedList;

/**
 * This class implements a fair queuing on the basis of Virtual Finish Times,
 * and therefore achieves fairness regarding the number of units (e.g. Bytes)
 * transmitted.
 * 
 * @author dwagner
 * 
 */
public class DeficitRoundRobinFairQueuing extends FairQueuing {
   class DeficitQDiscContainer extends SortableQDiscContainer {
      long deficit;
      QueueEntry headOfQDisc = null;

      public DeficitQDiscContainer(StackableQueuingDiscipline q) {
         qdisc = q;
      }

      public void incrementDeficit(long inc) {
         deficit += inc;
      }

      public void updateDeficit(double cvt) {
         if (!qdisc.isEmpty() && qdisc.peek() != headOfQDisc) {
            // head changed, need to update sortIndex = deficit
            deficit = (long) cvt;
            headOfQDisc = qdisc.peek();
         }
      }

      @Override
      public long getLengthInMetric() {
         return qdisc.getCurrentNumberOfUnits();
      }

      @Override
      public double getsortIndex() {
         // irrelevant for DRR
         return 0;
      }
   }

   protected LinkedList<SortableQDiscContainer> activeQs = new LinkedList<SortableQDiscContainer>();
   protected int activeQIndex = 0;
   protected final long deficitIncrement;

   public DeficitRoundRobinFairQueuing(SimNode ownNode, Parameters pars) {
      super(ownNode, pars);
      deficitIncrement = pars.get(simNode, "DeficitIncrement").asLong();
   }

   /** as required by {@link ReflectionConstructable} */
   public static DeficitRoundRobinFairQueuing createInstance(SimNode ownNode, Parameters pars) {
      return new DeficitRoundRobinFairQueuing(ownNode, pars);
   }

   /**
    * This is just a simplified option to implement the dropping estimation. It
    * just removes one message from all queues longer than the queue the message
    * would be inserted in but in fact, it might be necessary to remove several
    * messages from one of these queues.
    * 
    * When overriding it, keep in mind that with a stochastic classifier packets
    * at the head of the queue might belong to a completely different flow
    */
   @Override
   public boolean wouldStore(Message msg) {
      if (getCurrentNumberOfUnits() + msg.getLength() > getMaxSizeInUnits()) {
         long tooMuch = getCurrentNumberOfUnits() + msg.getLength() - getMaxSizeInUnits();
         if (queues.getConsumer(msg) != null) {
            for (long index : longestQList.tailMap(queues.getConsumer(msg).getQDisc().getCurrentNumberOfUnits(), false).descendingKeySet()) {
               for (SortableQDiscContainer sq : longestQList.get(index)) {
                  tooMuch -= sq.getQDisc().peek().getMessage().getLength();
                  if (tooMuch <= 0)
                     return true;
               }
            }
         } else {
            return true;
         }
         return false;
      } else {
         return true;
      }
   }

   @Override
   protected void addToActiveQList(SortableQDiscContainer sq) {
      activeQs.add(sq);
      if (activeQ == null) {
         activeQ = sq;
         assert (activeQIndex == 0); // shouldn't be neccessary
      }
   }

   @Override
   protected void updateActiveQList(SortableQDiscContainer sq, double oldPos) {
      // DRR is just round robin, so the list itself is normally not
      // touched on pushs and pops.
      // It just needs to remove sq from activeQs if sq is empty
      if (sq.getQDisc().getCurrentNumberOfMessages() == 0) {
         if (!activeQs.get(activeQIndex).equals(activeQ))
            throw new RuntimeException("activeQ does not point to activeQIndex!");
         activeQs.remove(activeQIndex);
      }
   }

   @Override
   protected void advanceActiveQ() {
      if (activeQs.size() > 0) {
         activeQIndex = (activeQIndex + 1) % activeQs.size();
         activeQ = activeQs.get(activeQIndex);
      } else {
         activeQIndex = 0;
         activeQ = null;
      }
   }

   @Override
   public SortableQDiscContainer createSortableQDisc(StackableQueuingDiscipline q) {
      return new DeficitQDiscContainer(q);
   }

   @Override
   protected boolean activeQListIsEmpty() {
      return activeQs.isEmpty();
   }
}
