/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.algorithms.queuingDisciplines;

import ikr.simlib.exceptions.SimulationException;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;

import java.util.Deque;
import java.util.LinkedList;

/**
 * This is the base class for all Last In First Out (LIFO) queuing disciplines.
 * There is an unbounded version ({@link UnboundedLIFOQDisc UnboundedLIFOQDisc})
 * and a bounded version ({@link BoundedLIFOQDisc BoundedLIFOQDisc}).
 * 
 */
public abstract class LIFOQDisc extends StackableQueuingDiscipline {

   protected final Deque<QueueEntry> queue;

   public LIFOQDisc(SimNode ownNode) {
      super(ownNode);
      queue = new LinkedList<QueueEntry>();
   }

   public LIFOQDisc(SimNode ownNode, Parameters pars) {
      this(ownNode);
   }

   @Override
   public int getMaxSizeInUnits() {
      return Integer.MAX_VALUE;
   }

   @Override
   public boolean isEmpty() {
      return queue.isEmpty();
   }

   @Override
   protected void push(QueueEntry qe) {
      if (qe.getMessage().getLength() > this.getMaxSizeInUnits()) {
         // Message too large for queue
         handleLossMessage(qe);
      } else {
         // Deleting messages from top of list until there is enough free
         // space for the newly arriving message.
         // dwagner: Not sure if this is the only and correct way to handle
         // push. Here implemented the same way as in the old LIFOQueue
         while (!wouldStore(qe.getMessage())) {
            handleLossMessage(pop());
         }
         queue.addLast(qe);
      }
   }

   @Override
   protected QueueEntry peekInternal() {
      if (queue.isEmpty()) {
         throw new SimulationException("Get Message from empty queue");
      }
      return queue.getLast();
   }

   @Override
   protected QueueEntry pop() {
      if (queue.isEmpty()) {
         throw new SimulationException("Get Message from empty queue");
      }

      QueueEntry qe = queue.removeLast();
      if (getCurrentNumberOfUnits() < 0) {
         throw new SimulationException("Actual number of units in the queue is less than 0");
      }
      return qe;
   }
}
