/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.algorithms.queuingDisciplines;

import ikr.simlib.algorithms.queuingDisciplines.BoundedFIFOQDisc;
import ikr.simlib.exceptions.SimulationException;
import ikr.simlib.messages.Message;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.ports.input.InputPort;
import ikr.simlib.ports.output.OutputPort;

/**
 * This is QDisc intended to be used as "glue queue", i.e. to adapt from a
 * pushing {@link OutputPort} to a pulling {@link InputPort}.
 * 
 * @author dwagner
 *
 */
public class NoDropBoundedFIFOQDisc extends BoundedFIFOQDisc {
   private final int maxNumberOfMessages;

   public NoDropBoundedFIFOQDisc(SimNode ownNode, int size) {
      super(ownNode, size);
      maxNumberOfMessages = -1;
      assert size >= 1;
   }

   public NoDropBoundedFIFOQDisc(SimNode ownNode, Parameters pars) {
      this(ownNode, pars, 0);
   }

   public NoDropBoundedFIFOQDisc(SimNode ownNode, Parameters pars, int size) {
      super(ownNode, size);
      if (!pars.hasParameter(simNode, "MaxSizeInUnits"))
         maxNumberOfMessages = pars.getOrUseDefault(simNode, "MaxNumberOfMessages", "1").asInteger();
      else
         maxNumberOfMessages = pars.getOrUseDefault(simNode, "MaxNumberOfMessages", "-1").asInteger();
      assert maxNumberOfMessages >= 1 || maxSizeInUnits >= 1;
   }

   /**
    * as required by
    * {@link ikr.simlib.parameters.reflection.ReflectionConstructable}
    */
   public static NoDropBoundedFIFOQDisc createInstance(SimNode ownNode, Parameters pars) {
      return new NoDropBoundedFIFOQDisc(ownNode, pars);
   }

   public static NoDropBoundedFIFOQDisc createInstance(SimNode ownNode, Parameters pars, Integer size) {
      return new NoDropBoundedFIFOQDisc(ownNode, pars, size);
   }

   @Override
   protected void statsUpdateLoss(QueueEntry qe) {
      throw new SimulationException(simNode.getFullName() + ": can't store packet in " + this.getClass().getName() + ". This should never happen!!!");
   }

   @Override
   public boolean wouldStore(Message msg) {
      if (maxNumberOfMessages >= 1) {
         if (getCurrentNumberOfMessages() + 1 >= maxNumberOfMessages)
            return false;
         if (maxSizeInUnits > 0)
            return super.wouldStore(msg);
         return true;
      }
      return super.wouldStore(msg);
   }

   public int getMaxNumberOfMessages() {
      return maxNumberOfMessages;
   }
}
