/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.algorithms.queuingDisciplines;

import ikr.simlib.events.calendar.Calendar;
import ikr.simlib.events.time.Duration;
import ikr.simlib.events.time.PointInTime;
import ikr.simlib.messages.Message;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.result.ResultKeywords;
import ikr.simlib.statistics.integral.IntegralStatistic;
import ikr.simlib.statistics.integral.StdIntegralStatistic;
import ikr.simlib.statistics.probability.ProbabilityStatistic;
import ikr.simlib.statistics.probability.StdProbabilityStatistic;
import ikr.simlib.statistics.sample.SampleStatistic;
import ikr.simlib.statistics.sample.StdSampleStatistic;

public abstract class StackableQueuingDiscipline implements QueuingDiscipline {

   protected final SimNode simNode;
   protected final SimNode parent;

   // State
   /**
    * The actual size of the queue in units in the messages kept in the buffer.
    * Equals the sum of getLength() of the messages in the buffer. This
    * attribute is private since it should only be touched by enqueue(msg) and
    * dequeue()
    */
   private long currentNumberOfUnits;

   /**
    * The number of messages kept in the buffer. This attribute is private since
    * it should only be touched by enqueue(msg) and dequeue().
    */
   private int currentNumberOfMessages;

   private QueueEntry lastPeeked;

   // Statistics
   protected final SampleStatistic waitingTime;
   protected final SampleStatistic transferTime;
   protected final IntegralStatistic numberOfUnits;
   protected final ProbabilityStatistic lossProbability;
   protected final IntegralStatistic numberOfMessages;

   public StackableQueuingDiscipline(SimNode ownNode, @SuppressWarnings("unused") Parameters pars) {
      this(ownNode);
   }

   public StackableQueuingDiscipline(SimNode ownNode) {
      this.simNode = ownNode;
      this.parent = ownNode.getParent();
      this.waitingTime = new StdSampleStatistic(ownNode.createChildNode("WaitingTime"));
      this.waitingTime.addResultTag(ResultKeywords.UNIT, "s");
      this.transferTime = new StdSampleStatistic(ownNode.createChildNode("TransferTime"));
      this.transferTime.addResultTag(ResultKeywords.UNIT, "s");
      this.numberOfMessages = new StdIntegralStatistic(ownNode.createChildNode("NumberOfMessages"));
      this.numberOfUnits = new StdIntegralStatistic(ownNode.createChildNode("NumberOfUnits"));
      this.lossProbability = new StdProbabilityStatistic(ownNode.createChildNode("LossProbability"));
      this.currentNumberOfUnits = 0;
      this.currentNumberOfMessages = 0;
      this.lastPeeked = null;
   }

   // get configuration information
   /**
    * @return maximum size of queue or <em>Integer.MAX_VALUE</em> if queue is
    *         unbounded.
    */
   @Override
   public abstract int getMaxSizeInUnits();

   @Override
   public String getName() {
      return simNode.getName();
   }

   @Override
   public SimNode getNode() {
      return simNode;
   }

   @Override
   public SimNode getParent() {
      return parent;
   }

   // get state information
   @Override
   public abstract boolean isEmpty();

   /**
    * @return true if the queue would store the given message. May return false
    *         due to a bounded buffer size or a probabilistic decision, for
    *         {@link ProbabilisticQueuingDiscipline}s. For these, two subsequent
    *         calls with the same message may return different results!
    */
   @Override
   public boolean wouldStore(Message msg) {
      return hasSpaceFor(msg);
   }

   /**
    * A simple method to check for "physical" space. Therefore, in contrast to
    * wouldStore() this method is deterministic and two subsequent calls will
    * return the same result. <br>
    * Any override implementation should be independent of the actual message
    * itself, but just depend from msg.getLength().
    * 
    * @param msg
    * @return true if the {@link QueuingDiscipline} has space for that message.
    * 
    */
   public boolean hasSpaceFor(Message msg) {
      return true;
   }

   @Override
   public final int getCurrentNumberOfMessages() {
      return currentNumberOfMessages;
   }

   @Override
   public final long getCurrentNumberOfUnits() {
      return currentNumberOfUnits;
   }

   @Override
   public final QueueEntry dequeue() {
      QueueEntry qe = pop();
      assert lastPeeked == null || lastPeeked == qe;
      lastPeeked = null;
      currentNumberOfUnits -= qe.getMessage().getLength();
      currentNumberOfMessages--;

      statsUpdateDequeue(qe);
      return qe;
   }

   @Override
   public boolean handleMessage(Message msg) {
      if (wouldStore(msg)) {
         enqueue(msg);
         return true;
      } else {
         QueueEntry qe = new QueueEntry(msg);
         handleRejectedMessage(qe);
         statsUpdateReject(qe);
         return false;
      }
   }

   @Override
   public QueueEntry dropScheduledNext() {
      QueueEntry qe = pop();
      handleLossMessage(qe);
      statsUpdateLoss(qe);
      return qe;
   }

   // //considered useful at the beginning, but maybe just not neccessary?
   // public QueueEntry dropScheduledLast(){
   // QueueEntry qe = removeLast();
   // handleLossMessage(qe);
   // currentNumberOfUnits -= qe.message.getLength();
   // currentNumberOfMessages--;
   //
   // statsUpdateLoss(qe);
   // return qe;
   // }

   public final void enqueue(Message msg) {
      QueueEntry qe = new QueueEntry(msg);
      push(qe);
      currentNumberOfUnits += qe.getMessage().getLength();
      currentNumberOfMessages++;

      statsUpdateEnqueue(qe);
   }

   /**
    * Whatever to be done at rejected packets, e.g. update the statistics.
    * 
    * @param qe
    */
   protected void handleRejectedMessage(QueueEntry qe) {
      // providing a slim interface for overriding
   }

   protected final void handleLossMessage(QueueEntry qe) {
      currentNumberOfUnits -= qe.getMessage().getLength();
      currentNumberOfMessages--;
      handleLossMessageInternal(qe);
   }

   /**
    * This method is intended to be overriden to implement any desired loss
    * handling while ensuring that numberOfMessages and -Units is decremented
    * accordingly.
    * 
    * @param qe
    */
   protected void handleLossMessageInternal(QueueEntry qe) {
   }

   /**
    * Whatever to be done at enqueue, e.g. update the statistics.
    * 
    * @param qe
    */
   protected void statsUpdateEnqueue(QueueEntry qe) {
      lossProbability.update(1, false); // ok
      numberOfMessages.update(getCurrentNumberOfMessages());
      numberOfUnits.update(getCurrentNumberOfUnits());
   }

   protected void statsUpdateReject(QueueEntry qe) {
      statsUpdateLoss(qe);
   }

   /**
    * Whatever to be done on loss, e.g. update the statistics.
    * 
    * @param qe
    */
   protected void statsUpdateLoss(QueueEntry qe) {
      lossProbability.update(1, true);
   }

   protected void statsUpdateDequeue(QueueEntry qe) {
      PointInTime sysTime = Calendar.getInstance().getSystemTime();
      Duration _waitingTime = qe.getArrivalTime().until(sysTime);
      transferTime.update(_waitingTime.toSeconds());
      numberOfMessages.update(getCurrentNumberOfMessages());
      numberOfUnits.update(getCurrentNumberOfUnits());

      if (_waitingTime.isGreaterThan(Duration.ZERO)) {
         waitingTime.update(_waitingTime.toSeconds());
      }
   }

   /**
    * Push message into the data structure managed by the queuing discipline.
    */
   protected abstract void push(QueueEntry qe);

   /**
    * Retrieve the QueueEntry that would be delivered next, without actually
    * removing it. This is trivial for enqueue-/tail-dropping queues, but not
    * for dequeue-/head-dropping queues such as CoDel. Anyway, this method MUST
    * return the same {@link Message} object than the next pop(). Therefore only
    * peekInternal() may be overriden.
    */
   public final QueueEntry peek() {
      QueueEntry ret = peekInternal();
      assert lastPeeked == null || lastPeeked == ret;
      lastPeeked = ret;
      return ret;
   }

   protected abstract QueueEntry peekInternal();

   /**
    * Retrieve and remove the next QueueEntry according to the queuing
    * discipline.
    * 
    * @return the QueueEntry
    */
   protected abstract QueueEntry pop();
}
