/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.control;

import ikr.simlib.control.simulationcontrol.SimControlTracer;
import ikr.simlib.events.calendar.Calendar;
import ikr.simlib.events.calendar.OrderedTreeSetCalendar;
import ikr.simlib.events.calendar.SimpleOrderedCalendar;
import ikr.simlib.events.calendar.SimpleRandomCalendar;
import ikr.simlib.events.calendar.StdOrderedCalendar;
import ikr.simlib.exceptions.ExitExceptionHandler;
import ikr.simlib.exceptions.SimulationException;
import ikr.simlib.model.Model;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.fileparser.FileParser;
import ikr.simlib.random.RandomNumberGenerator;
import ikr.simlib.random.StdRandomNumberGenerator;
import ikr.simlib.support.cmdline.ArgParser;
import ikr.simlib.support.cmdline.ArgumentException;

import java.io.PrintWriter;
import java.util.EnumSet;

/**
 * Base class for <code>SimTree</code> applications. Applications derived from
 * this class support the following command line arguments:
 * 
 * <dl>
 * <dt><code>-s seed</code></dt>
 * <dd>Seed index for the system RNG</dd>
 * <dt><code>-n numberOfBatches</code></dt>
 * <dd>The number of batches to run. If <code>0</code>, the simulation is not
 * run at all.</dd>
 * <dt><code>-p par</code></dt>
 * <dd>Parameter file (default: <code>sim.par</code>)</dd>
 * <dt><code>-f fmt</code></dt>
 * <dd>Format file</dd>
 * <dt><code>-l log</code></dt>
 * <dd>Result file (default: <code>sim.log</code>). If the log file is
 * <code>none</code>, no results are written.</dd>
 * <dt><code>-b batchLog(s)</code></dt>
 * <dd>Batch result files. If only one argument is given and is equal to
 * <code>default</code>, the default batch logs are created.</dd>
 * <dt><code>-e exportLog(s)</code></dt>
 * <dd>Export result files. If only one argument is given and is equal to
 * <code>default</code>, the default export logs are created.</dd>
 * <dt><code>-i importLog(s)</code></dt>
 * <dd>Import result files.</dd>
 * <dt><code>-g parseLog</code></dt>
 * <dd>Log file for the parser</dd>
 * <dt><code>-t testLog</code></dt>
 * <dd>Test result file</dd>
 * <dt><code>-z compression type</code></dt>
 * <dd>Valid compression types are "bzip2", "lzma", or "gzip". The default
 * behavior is to not compress files.
 * <dt><code>-c calendar type</code></dt>
 * <dd>Valid calendar types are "std", "simple", or "treeset". The default
 * calendar is the "simple" calendar.
 * </dl>
 * 
 * <p>
 * A typical application may be started like this:
 * </p>
 * 
 * <pre>
 * public static void main(String[] args) {
 *    SimTreeApplication app = new MyApp(args);
 * 
 *    app.run();
 * }
 * </pre>
 * 
 * <p>
 * Applications have to implement the <code>createModel</code> method to provide
 * the simulation model. Further control over the application components is
 * given to deriving classes by the <code>createCalendar</code>,
 * <code>createSystemRNG</code> and <code>createSimulation</code> methods, which
 * have sensible default implementations.
 * </p>
 * 
 * <p>
 * The constructor might exit the VM with <code>System.exit</code> to be
 * compatible to the <code>SimTree</code> semantics.
 * </p>
 * 
 * <p>
 * The <code>SimTreeApplication</code> can also be executed directly from the
 * command line. It requires the class name of the model as first argument. The
 * model constructor must look like this:
 * </p>
 * 
 * <pre>
 * MyModel(ParseEnvrionment env)
 * </pre>
 * 
 * @see SimulationEnvironment
 * 
 */
public abstract class SimTreeApplication {

   /**
    * Simulation environment for the default simulations.
    */
   protected SimulationEnvironment simEnvironment;

   /**
    * Simulation model.
    */
   protected Model model;

   /**
    * The simulation. Depending on the command line arguments it is initialized
    * by either <code>createSimulation</code> or
    * <code>createImportSimulation</code>.
    */
   protected Simulation simulation;

   /**
    * Creates a new <code>SimTree</code> application. This constructor is solely
    * used by deriving applications which have to call <code>init</code>
    * themselves.
    */
   protected SimTreeApplication() {
   }

   /**
    * Creates a new <code>SimTree</code> application. This constructor might
    * exit the VM with <code>System.exit</code> to adhere to the
    * <code>SimTree</code> semantics.
    * 
    * @param name
    *           of the <code>SimTree</code> application
    * @param args
    *           command line arguments
    */
   protected SimTreeApplication(String name, String[] args) {
      init(name, args);
   }

   /**
    * Parses the arguments and creates the simulation environment. This method
    * might exit the VM with <code>System.exit</code> to adhere to the
    * <code>SimTree</code> semantics.
    */
   protected void init(String name, String[] args) {
      // 1. parse the command line arguments
      ArgParser argParser = new ArgParser(name, SimulationEnvironment.USAGE);
      try {
         simEnvironment = new SimulationEnvironment(argParser, args);

      } catch (ArgumentException ae) {
         ae.printStackTrace();
         argParser.printUsage();
         System.exit(1);

      } catch (SimulationException se) {
         se.printStackTrace();
         argParser.printError(se.getMessage());
         System.exit(1);
      }

      // 2. set the exception handler
      ExitExceptionHandler.setAsDefaultUncaughtExceptionHandler();

      // 3. create the calendar
      createCalendar(simEnvironment.getCalendarType());

      // 4. create the random number generator
      RandomNumberGenerator.setSystemRNG(createSystemRNG());

      // 5. create the model and parse the parameter file
      PrintWriter parserLogStream = simEnvironment.getParserLogStream();
      Parameters parameters = new Parameters(parserLogStream);
      FileParser.parseInputStream(simEnvironment.getParameterStream(), parameters, parserLogStream);
      model = createModel(parameters);

      // 6. create the respective simulation
      simulation = simEnvironment.isImportSimulation() ? createImportSimulation() : createSimulation();

      // 7. create the simulation control tracer
      SimControlTracer tracer = new SimControlTracer();
   }

   /**
    * Creates the model for the simulation. The <code>calendar</code> field is
    * initialized before this method is called.
    * 
    * @param pars
    *           an object representing the entire parameter tree
    * @return the simulation model
    */
   protected abstract Model createModel(Parameters pars);

   /**
    * Creates the calendar for the simulation.
    */
   protected void createCalendar(String name) {
      try {
         CalendarType.setCalendarInstanceByShortName(name);
      } catch (ArgumentException ae) {
         ae.printStackTrace();
         System.err.println(CalendarType.getAllowedTypes());
         System.exit(1);
      }
   }

   /**
    * Creates the system RNG. By default a <code>StdRandomNumberGenerator</code>
    * is returned.
    * 
    * @return the system RNG
    */
   protected RandomNumberGenerator createSystemRNG() {
      return new StdRandomNumberGenerator(simEnvironment.getSeedIndex());
   }

   /**
    * Creates the simulation. By default a <code>StdSimulation</code> is
    * returned.
    */
   protected Simulation createSimulation() {
      return new StdSimulation(model, simEnvironment);
   }

   /**
    * Creates the import simulation. By default a <code>ImportSimulation</code>
    * is returned.
    */
   protected Simulation createImportSimulation() {
      return new ImportSimulation(model, simEnvironment);
   }

   /**
    * Runs the simulation.
    */
   public void run() {
      simulation.run();
   }

   private enum CalendarType {

      std {
         @Override
         public String getShortName() {
            return "std";
         }

         @Override
         public String getName() {
            return StdOrderedCalendar.class.getSimpleName();
         }
      },
      simple {
         @Override
         public String getShortName() {
            return "simple";
         }

         @Override
         public String getName() {
            return SimpleOrderedCalendar.class.getSimpleName();
         }
      },
      treeset {
         @Override
         public String getShortName() {
            return "treeset";
         }

         @Override
         public String getName() {
            return OrderedTreeSetCalendar.class.getSimpleName();
         }
      },
      simplerandom {
         @Override
         public String getShortName() {
            return "simplerandom";
         }

         @Override
         public String getName() {
            return SimpleRandomCalendar.class.getSimpleName();
         }
      };

      public String getShortName() {
         throw new AbstractMethodError();
      }

      public String getName() {
         throw new AbstractMethodError();
      }

      public static void setCalendarInstanceByShortName(String name) {
         if (name.equalsIgnoreCase(std.getShortName())) {
            Calendar.setInstance(new StdOrderedCalendar());
         } else if (name.equalsIgnoreCase(simple.getShortName())) {
            Calendar.setInstance(new SimpleOrderedCalendar());
         } else if (name.equalsIgnoreCase(treeset.getShortName())) {
            Calendar.setInstance(new OrderedTreeSetCalendar());
         } else if (name.equalsIgnoreCase(simplerandom.getShortName())) {
            Calendar.setInstance(new SimpleRandomCalendar());
         } else if (name.equalsIgnoreCase("")) {
            Calendar.setInstance(new SimpleRandomCalendar());
         } else {
            throw new ArgumentException("Requested calendar is unknown.");
         }
      }

      public static String getAllowedTypes() {
         final EnumSet<CalendarType> enumSet = EnumSet.allOf(CalendarType.class);
         final StringBuilder sb = new StringBuilder();
         sb.append("allowed calendars:\n");
         for (CalendarType t : enumSet) {
            sb.append(String.format("\t%-14s - sets new %s\n", t.getShortName(), t.getName()));
//            sb.append("\t" + t.getShortName() + ":\t\t" + t.getName() + "\n");
         }
         return sb.toString();
      }
   }
}
