/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.control;

import ikr.simlib.exceptions.SimulationException;
import ikr.simlib.result.PrintResultWriter;
import ikr.simlib.result.ResultFileType;
import ikr.simlib.result.ResultReader;
import ikr.simlib.result.ResultWriter;
import ikr.simlib.result.compacting.CompactingResultWriter;
import ikr.simlib.result.exceptions.ImportException;
import ikr.simlib.result.filtering.Filter;
import ikr.simlib.result.filtering.PrintResultFilterWriter;
import ikr.simlib.result.xml.XMLResultReader;
import ikr.simlib.result.xml.XMLResultWriter;
import ikr.simlib.support.DelayedOutputStream;
import ikr.simlib.support.cmdline.ArgParser;
import ikr.simlib.support.compression.CompressionManager;
import ikr.simlib.support.compression.CompressionType;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.io.Reader;
import java.util.ArrayList;
import java.util.List;

/**
 * The <code>SimulationEnvironment</code> provides a standard set of command
 * line parameters which are automatically parsed. It is fully compatible with
 * <em>SimTree</em>.<br>
 * <br>
 * The class {@link ikr.simlib.control.Simulation} makes use of
 * <code>SimulationEnvironment</code> in order to read command line parameters.
 * The possible parameters are given in the following:<br>
 * <br>
 * <table border="1">
 * <tr>
 * <th>Parameter</th>
 * <th>Range</th>
 * <th>Description</th>
 * </tr>
 * <tr>
 * <td>n</td>
 * <td>>0</td>
 * <td>number of batches</td>
 * </tr>
 * <tr>
 * <td>s</td>
 * <td>0-127</td>
 * <td>index of seed for random number generator</td>
 * </tr>
 * <tr>
 * <td>p</td>
 * <td></td>
 * <td>parameter file</td>
 * </tr>
 * <tr>
 * <td>f</td>
 * <td></td>
 * <td>filter file</td>
 * </tr>
 * <tr>
 * <td>l</td>
 * <td></td>
 * <td>log file</td>
 * </tr>
 * <tr>
 * <td>b</td>
 * <td></td>
 * <td>batch log file</td>
 * </tr>
 * <tr>
 * <td>e</td>
 * <td></td>
 * <td>export file for batch results</td>
 * </tr>
 * <tr>
 * <td>i</td>
 * <td></td>
 * <td>import file for batch results</td>
 * </tr>
 * <tr>
 * <td>g</td>
 * <td></td>
 * <td>parse log file</td>
 * </tr>
 * <tr>
 * <td>t</td>
 * <td></td>
 * <td>test log file</td>
 * </tr>
 * <tr>
 * <td>c</td>
 * <td></td>
 * <td>type of calendar</td>
 * </tr>
 * </table>
 * <br>
 * If specified the number of given export and import files needs to match to
 * the number of batches. Simultaneous definition of import and export files is
 * not possible.
 * 
 * 
 */
public class SimulationEnvironment {

   public static final String USAGE = "[-n numberOfBatches] [-s seed]\n [-p parfile] [-l logfile] [-b batchlogfile(s)] [-e exportlogfile(s)] [-i importlogfile(s)] [-g parselogfile] [-f filterfile] [-c calendartype] [-z compressiontype]";

   protected int numberOfBatches;

   protected int seedIndex;

   protected InputStream parameterStream;

   protected PrintWriter parserLogStream;

   protected Reader filterStream;

   protected PrintResultWriter logWriter;

   protected List<PrintResultWriter> batchLogWriters;

   protected List<ResultWriter> exportWriters;

   protected List<ResultReader> importReaders;

   protected String calendarType;

   protected CompressionType compressionType;

   protected SimulationEnvironment() {
      batchLogWriters = new ArrayList<PrintResultWriter>();
      exportWriters = new ArrayList<ResultWriter>();
      importReaders = new ArrayList<ResultReader>();
      logWriter = null;
      filterStream = null;
   }

   public SimulationEnvironment(ArgParser argParser, String[] args) {
      this();

      addOptions(argParser);

      argParser.parse(args);

      init(argParser);

      if (filterStream != null && logWriter != null) {
         Filter.getInstance().readFilterStream(filterStream);
      }
   }

   public SimulationEnvironment(String name, String[] args) {
      this(new ArgParser(name, USAGE), args);
   }

   protected void addOptions(ArgParser argParser) {
      argParser.addOption('n', 1);
      argParser.addOption('s', 1);
      argParser.addOption('p', 1);
      argParser.addOption('l', 1);
      argParser.addOption('b', 1, -1);
      argParser.addOption('e', 1, -1);
      argParser.addOption('i', 1, -1);
      argParser.addOption('g', 1);
      argParser.addOption('f', 1);
      argParser.addOption('c', 1);
      argParser.addOption('z', 1);
   }

   protected void init(ArgParser argParser) {
      // Set compression type
      if (argParser.hasOption('z')) {
         try {
            compressionType = CompressionType.valueOf(argParser.getArgument('z', ""));
         } catch (IllegalArgumentException e) {
            throw new SimulationException("specified compression type unknown (valid types: " + CompressionType.listAll() + ")", e);
         }
      } else {
         compressionType = CompressionType.none;
      }

      numberOfBatches = argParser.getIntArgument('n', 10);

      // remember seed index, default is zero
      seedIndex = argParser.getIntArgument('s', 0);

      String parameterName = argParser.getArgument('p', "sim.par");
      try {
         parameterStream = new FileInputStream(parameterName);
      } catch (FileNotFoundException fnfe) {
         throw new SimulationException("Couldn't open parameter file \"" + parameterName + "\"");
      }

      if (argParser.hasOption('g')) {
         parserLogStream = new PrintWriter(getOutputStream(argParser.getArgument('g')), true);
      }

      if (argParser.hasOption('f')) {
         try {
            filterStream = new FileReader(argParser.getArgument('f', "sim.fil"));
         } catch (FileNotFoundException fnfe) {
            throw new SimulationException("Couldn't open filter file \"" + argParser.getArgument('f') + "\"");
         }
      }

      // final result
      String logName = argParser.getArgument('l', "sim.log");
      if (!logName.equalsIgnoreCase("none")) {
         PrintStream logStream = new PrintStream(getOutputStream(logName));
         logWriter = new PrintResultFilterWriter(new CompactingResultWriter(new XMLResultWriter(logStream, ResultFileType.FINAL, seedIndex, Integer.MAX_VALUE, numberOfBatches, logName)));
      }

      // Creating batch files
      if (argParser.getArgumentCount('b') >= 1) {

         // Reading batch log file names from command line
         if (argParser.getArgumentCount('b') != numberOfBatches) {
            throw new SimulationException("Number of given batch log files is " + argParser.getArgumentCount('b') + " which mismatches given desired number of batches " + numberOfBatches + ".");
         }
         for (int i = 0; i < numberOfBatches; i++) {
            String fileName = argParser.getNthArgument('b', i);
            batchLogWriters.add(new PrintResultFilterWriter(new CompactingResultWriter(new XMLResultWriter(getOutputStream(fileName), ResultFileType.BATCH, seedIndex, i + 1, numberOfBatches, fileName))));
         }
      }

      // Check for import option
      if (argParser.hasOption('i') && argParser.hasOption('e')) {
         throw new SimulationException("Cannot export simulation results while importing others.");
      }

      // Check if batch number is specified
      if (argParser.hasOption('e') && !argParser.hasOption('n')) {
         throw new SimulationException("Specify number of batches for exporting simulation.");
      }
      if (argParser.hasOption('i') && !argParser.hasOption('n')) {
         throw new SimulationException("Specify number of batches for importing simulation.");
      }

      // Creating export files
      if (argParser.getArgumentCount('e') >= 1) {

         // Reading batch log file names from command line
         if (argParser.getArgumentCount('e') != numberOfBatches) {
            throw new SimulationException("Number of given export log files is " + argParser.getArgumentCount('e') + " which mismatches given desired number of batches " + numberOfBatches + ".");
         }
         for (int i = 0; i < numberOfBatches; i++) {
            String fileName = argParser.getNthArgument('e', i);
            exportWriters.add(new XMLResultWriter(getOutputStream(fileName), ResultFileType.EXPORT, seedIndex, i + 1, numberOfBatches, fileName));
         }
      }

      // Import log files
      if (argParser.hasOption('i')) {

         // Reading import log file names from command line
         if (argParser.getArgumentCount('i') != numberOfBatches) {
            throw new SimulationException("Number of given import log files is " + argParser.getArgumentCount('i') + " which mismatches given desired number of batches " + numberOfBatches + ".");
         }
         for (int i = 0; i < numberOfBatches; i++) {
            String fileName = argParser.getNthArgument('i', i);
            try {
               importReaders.add(new XMLResultReader(getInputStream(fileName), fileName));
            } catch (ImportException e) {
               throw new RuntimeException("error while opening import file " + fileName, e);
            }
         }
      }

      // Set calendar type
      if (argParser.hasOption('c')) {
         calendarType = argParser.getArgument('c', "");
      } else {
         calendarType = "";
      }

   }

   public int getNumberOfBatches() {
      return numberOfBatches;
   }

   public int getSeedIndex() {
      return seedIndex;
   }

   public InputStream getParameterStream() {
      return parameterStream;
   }

   public PrintWriter getParserLogStream() {
      return parserLogStream;
   }

   public PrintResultWriter getLogWriter() {
      return logWriter;
   }

   public List<PrintResultWriter> getBatchLogWriters() {
      return batchLogWriters;
   }

   public List<ResultWriter> getExportWriters() {
      return exportWriters;
   }

   public List<ResultReader> getImportReaders() {
      return importReaders;
   }

   public boolean isImportSimulation() {
      return !importReaders.isEmpty();
   }

   public String getCalendarType() {
      return calendarType;
   }

   /**
    * Creates a new input stream accessing a file in compressed or uncompressed
    * manner
    */
   private static InputStream getInputStream(String fileName) {
      try {
         return CompressionManager.getInputStream(new File(fileName));
      } catch (FileNotFoundException e) {
         throw new SimulationException("Could not open file " + fileName);
      }
   }

   /**
    * Creates a new output stream accessing a file in compressed or uncompressed
    * manner
    */
   private OutputStream getOutputStream(String fileName) {
      final String fileNameWithExtension = CompressionManager.getFilenameWithExtension(fileName, compressionType);
      final File file = new File(fileNameWithExtension);
      return new DelayedOutputStream() {
         @Override
         protected OutputStream openUnderlying() {
            try {
               return CompressionManager.getOutputStream(file, compressionType);
            } catch (FileNotFoundException e) {
               throw new SimulationException("Could not open file " + file.getAbsolutePath());
            }
         }
      };
   }
}
