/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.control;

import ikr.simlib.control.simulationcontrol.callbacks.SimulationControlManager;
import ikr.simlib.exceptions.SimulationException;
import ikr.simlib.model.Model;
import ikr.simlib.model.SimNodeManager;
import ikr.simlib.result.PrintResultWriter;
import ikr.simlib.result.ResultType;
import ikr.simlib.result.ResultWriter;
import ikr.simlib.result.exceptions.PrintException;

/**
 * Standard implementation for normal sequential simulations with one global
 * model.
 */
public class StdSimulation extends Simulation {

   protected final Model model;

   protected final SimulationEnvironment simEnv;

   public StdSimulation(Model model, SimulationEnvironment simEnv) {
      super();

      if (model == null) {
         throw new SimulationException("no model");
      }

      this.model = model;
      this.simEnv = simEnv;
   }

   protected void printBatchResults(int batchNumber) {
      if (batchNumber == 0) {
         throw new SimulationException("Invalid batch number: " + batchNumber);
      }

      if (simEnv.getBatchLogWriters().size() > batchNumber - 1) {
         PrintResultWriter writer = simEnv.getBatchLogWriters().get(batchNumber - 1);
         try {
            writer.writeStartResults();
            SimNodeManager.getInstance().doPrintResults(ResultType.BATCH, writer);
            writer.writeEndResults();
            writer.close();
         } catch (PrintException e) {
            throw new RuntimeException("error while printing batch results of batch no. " + batchNumber, e);
         }
      }

      if (simEnv.getExportWriters().size() > batchNumber - 1) {
         ResultWriter writer = simEnv.getExportWriters().get(batchNumber - 1);
         try {
            writer.writeStartResults();
            SimNodeManager.getInstance().doExportBatch(writer);
            writer.writeEndResults();
            writer.close();
         } catch (PrintException e) {
            throw new RuntimeException("error while exporting batch results of batch no. " + batchNumber, e);
         }
      }
   }

   @Override
   public void handleEndOfCurrentBatch() {
      // we have processed all events for this phase
      model.stopProcessingEvents();
      // processPendingEvents() will return after the current event has been
      // processed
   }

   @Override
   public void handleEndOfTransientPhase() {
      // we have processed all events for this phase
      model.stopProcessingEvents();
      // processPendingEvents() will return after the current event has been
      // processed
   }

   /**
    * prints the simulation results
    */
   @Override
   public void printResults() {
      if (simEnv.getLogWriter() != null) {
         PrintResultWriter writer = simEnv.getLogWriter();
         try {
            writer.writeStartResults();
            SimNodeManager.getInstance().doPrintResults(ResultType.FINAL, simEnv.getLogWriter());
            writer.writeEndResults();
            writer.close();
         } catch (PrintException e) {
            throw new RuntimeException("error while printing final results", e);
         }
      }
   }

   /**
    * executes the actual simulation with a subdivision in simulation phases<br>
    * <ul>
    * <li>a warm-up (transient) phase at the beginning of simulation</li>
    * <li>several batches during which statistical data is collected</li>
    * </ul>
    * In each phase event generation and processing is started by calling the
    * method <code>processPendingEvents()</code> of the model object given as
    * parameter on construction of {@link ikr.simlib.control.StdSimulation}
    * <ul>
    * <li>notifies the
    * {@link ikr.simlib.control.simulationcontrol.callbacks.SimulationControlManager}
    * at the beginning and end of each simulation phase as well as at the end of
    * the simulation</li>
    * <li>prints the partial results after each batch (
    * <code>printBatchResults()</code>)</li>
    * </ul>
    */
   @Override
   public void runSimulation() {

      // transient phase
      SimulationControlManager.getInstance().startTransientPhase();
      model.processPendingEvents();
      SimulationControlManager.getInstance().stopTransientPhase();

      // do the batches
      for (int batchNumber = 1; batchNumber <= simEnv.getNumberOfBatches(); batchNumber++) {
         SimulationControlManager.getInstance().startBatch(batchNumber);
         model.processPendingEvents();
         SimulationControlManager.getInstance().stopBatch(batchNumber);

         // print results
         printBatchResults(batchNumber);
      }

      // end simulation
      SimulationControlManager.getInstance().stopSimulation();
   }
}
