/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions.continuous;

import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.random.RandomNumberGenerator;

/**
 * Deterministic Value<br>
 * 
 * <table border="0">
 * <tr>
 * <th align="left" valign="top">Meaning:</th>
 * <td>A constant value \(d\) is returned</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parameters:</th>
 * <td>constant ("mean") value \(d\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">PDF:</th>
 * <td>\(P(T=t) = f(t) = \delta(t-d) \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">DF:</th>
 * <td>\(P(T \le t) = F(t) = \sigma(t-d) = \begin{cases} 0 &\mbox{for } t < d \\
 * 1 & \mbox{else } \end{cases} \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Expected value:</th>
 * <td>\(E[T]= d \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Variance:</th>
 * <td>\(VAR[T]= 0\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Coefficient of variation:</th>
 * <td>\(c_T= 0\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">LST:</th>
 * <td>\(\phi(s) = exp(-sd)\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parser example:</th>
 * <td><code>
         [...].Distribution = ikr.simlib.distributions.continuous.ConstantDistribution
         <br>
         [...].Distribution.Mean = 1.7</code></td>
 * </tr>
 * </table>
 * 
 */
public class ConstantDistribution extends ContinuousDistribution {

   private static final String MEAN = "Mean";

   /** as required by {@link ReflectionConstructable} */
   public static ConstantDistribution createInstance(SimNode ownNode, Parameters pars, @SuppressWarnings("unused") RandomNumberGenerator rng) {
      return new ConstantDistribution(pars.get(ownNode, MEAN).asDouble());
   }

   public final double mean;

   public ConstantDistribution(double mean) {
      this.mean = mean;
   }

   public ConstantDistribution(SimNode ownNode, Parameters pars) {
      this.mean = pars.get(ownNode, "Mean").asDouble();
   }

   @Override
   public double next() {
      return mean;
   }
}
