/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions.continuous;

import ikr.simlib.distributions.DistributionException;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.random.RandomNumberGenerator;

/**
 * Erlang k Distribution<br>
 * 
 * <center>
 * <img src="{@docRoot}/ikr/simlib/doc-files/erlangKP25.svg" width="600" alt="Erlang k distribution"> <br>
 * Erlang k distribution </center>
 * 
 * <table border="0">
 * <tr>
 * <th align="left" valign="top">Meaning:</th>
 * <td>Distribution for the sum of \(k\) random variables that are each
 * negative-exponentially distributed with the parameter \(\lambda\) (serial
 * switch in the phase model).
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parameters:</th>
 * <td>
 * <ul>
 * <li>
 * order \(k>0\)
 * <li>
 * rate \(\lambda > 0\) of the individual phases or total mean value \(m =
 * \frac{k}{\lambda}\)
 * </ul>
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">PDF:</th>
 * <td>\(P(T=t) = f(t) = \lambda \cdot \frac{(\lambda t)^{k-1}}{(k-1)!} \cdot
 * exp(-\lambda t)\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">DF:</th>
 * <td>\(P(T \le t) = F(t) = 1 - exp(-\lambda t) \cdot \sum\limits_{i=0}^{k-1}
 * \frac{(\lambda t)^i}{i!} \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Expected value:</th>
 * <td>\(E[T]= \frac{k}{\lambda} = m\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Variance:</th>
 * <td>\(VAR[T]= \frac{k}{\lambda^2} = \frac{m^2}{k}\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Coefficient of variation:</th>
 * <td>\(c_T= \frac{1}{\sqrt{k}} \le 1\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">LST:</th>
 * <td>\(\phi(s) = (\frac{\lambda}{\lambda +s})^k = (\frac{1}{1+ms})^k\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parser example:</th>
 * <td>
 * <code>
         [...].Distribution = ikr.simlib.distributions.continuous.ErlangDistribution
         <br>
         [...].Distribution.Mean = 4.5 # k/lambda
         <br>
         [...].Distribution.Order = 3 # number of phases (k)
      </code></td>
 * </tr>
 * </table>
 * 
 */
public class ErlangDistribution extends ContinuousDistribution {

   private static final String MEAN = "Mean";
   private static final String ORDER = "Order";

   /** as required by {@link ReflectionConstructable} */
   public static ErlangDistribution createInstance(SimNode ownNode, Parameters pars, RandomNumberGenerator rng) {
      final double mean = pars.get(ownNode, MEAN).asDouble();
      final int order = pars.get(ownNode, ORDER).asInteger();
      return new ErlangDistribution(mean, order, rng);
   }

   public final int order;
   public final NegExpDistribution phase;

   public ErlangDistribution(double mean, int order, RandomNumberGenerator rng) {
      super(rng);

      if (order <= 0 || order > 100) {
         throw new DistributionException("order must be > 0 and <= 100");
      }

      this.order = order;
      phase = new NegExpDistribution(mean / order);
   }

   public ErlangDistribution(double mean, int order) {
      this(mean, order, getDefaultRNG());
   }

   @Override
   public double next() {
      double sum = 0;

      for (int i = 0; i < order; i++) {
         sum += phase.next();
      }

      return sum;
   }
}
