/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions.continuous;

import ikr.simlib.distributions.DistributionException;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.random.RandomNumberGenerator;

/**
 * Jakes Distribution<br>
 * 
 * 
 * <table border="0">
 * <tr>
 * <th align="left" valign="top">Meaning:</th>
 * <td>The class <code>
         JakesDistribution
      </code> implementes a continuous distribution which delivers
 * random variables in accordance with Jake's Doppler power density spectrum.
 * Jake's power density spectrum is often used to model the distribution of
 * Doppler shifts in a fading channel. The distribution delivers random
 * variables according to Jake's spectrum between - Fdmax and +Fdmax. For an
 * introduction to this topic and the inverse of the cdf see in the references.</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parameters:</th>
 * <td>maximum Doppler shift \(f_{D,max}\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">PDF:</th>
 * <td>\(P(T=t) = f(t) = \frac{1}{\pi \cdot f_{D,max} \cdot \sqrt{1-
 * \frac{t}{f_{D,max}}}} \mbox{ for } |t| < f_{D,max} \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">DF:</th>
 * <td>\(P(T \le t) = F(t) = \frac{1}{2} + \frac{1}{\pi} \cdot
 * \arcsin{\frac{t}{f_{D,max}}} \mbox{ for } |t| < f_{D,max}\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Expected value:</th>
 * <td>\(E[T]= 0\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Variance:</th>
 * <td>\(VAR[T]= \frac{1}{2}f_{D,max}^2\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Coefficient of variation:</th>
 * <td>not applicable</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parser example:</th>
 * <td>
 * <code>
         [...].Distribution = ikr.simlib.distributions.continuous.JakesDistribution
         <br>
         [...].Distribution.Fdmax = 100.0 # Hz
      </code></td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">References:</th>
 * <td>
 * <ul>
 * <li>
 * P. Hoeher: Kohaerenter Empfang trelliscodierter PSK-Signale auf
 * frequenzselektiven Mobilfunkkanaelen - Entzerrung, Decodierung und
 * Kanalparameterschaetzung, VDI Fortschrittsberichte, Reihe 10, Nr. 147.
 * <li>
 * P. Hoeher: A Statistical Discrete-Time Model for the WSSUS Multipath Channel,
 * IEEE Transactions on Vehicular Technology, Vol. 41, No. 4, pp. 461-468,
 * November 1992.
 * </ul>
 * </td>
 * </tr>
 * </table>
 * 
 */
public class JakesDistribution extends ContinuousDistribution {

   public final double fdmax;

   private static final String FDMAX = "Fdmax";

   /** as required by {@link ReflectionConstructable} */
   public static JakesDistribution createInstance(SimNode ownNode, Parameters pars, RandomNumberGenerator rng) {
      final double fdmax = pars.get(ownNode, FDMAX).asDouble();
      return new JakesDistribution(fdmax, rng);
   }

   public JakesDistribution(double fdmax, RandomNumberGenerator rng) {
      super(rng);

      if (fdmax <= 0) {
         throw new DistributionException("Fdmax must be > 0");
      }

      this.fdmax = fdmax;
   }

   public JakesDistribution(double fdmax) {
      this(fdmax, getDefaultRNG());
   }

   @Override
   public double next() {
      return fdmax * Math.cos(2.0 * Math.PI * rng.next());
   }
}
