/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions.continuous;

import ikr.simlib.distributions.DistributionException;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.random.RandomNumberGenerator;

/**
 * Normal Distribution<br>
 * 
 * <table border="0">
 * <tr>
 * <th align="left" valign="top">Meaning:</th>
 * <td>Limit distribution of the sum of many independent random variables with
 * arbitrary statistical characteristics, if the contribution of a single random
 * variable remains neglectably small.</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parameters:</th>
 * <td>
 * <ul>
 * <li>
 * mean value \(\mu\)
 * <li>
 * standart deviation \(\sigma > 0\)
 * </ul>
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">PDF:</th>
 * <td>\(P(T=t) = f(t) = \frac{1}{\sqrt{2 \pi} \sigma} \cdot exp( -\frac{(t-
 * \mu)^2}{2\sigma^2} )\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">DF:</th>
 * <td>\(P(T \le t) = F(t) = \frac{1}{2} \cdot erf(\frac{t-\mu}{\sqrt{2}
 * \sigma}) \mbox{ with } erf(x) = \frac{2}{\sqrt{\pi}} \cdot \int_0^x \!
 * exp(-y^2) \, dy \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Expected value:</th>
 * <td>\(E[T]= \mu \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Variance:</th>
 * <td>\(VAR[T]= \sigma^2\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Coefficient of variation:</th>
 * <td>\(c_T= \frac{\sigma}{\mu}\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">LST:</th>
 * <td>\(\phi(s) = exp(-\mu s + \frac{(s \sigma)^2}{2})\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parser example:</th>
 * <td>
 * <code>
         [...].Distribution = ikr.simlib.distributions.continuous.NormalDistribution
         <br>
         [...].Distribution.Mean = 6.3
         <br>
         [...].Distribution.StandardDeviation = 1.5
      </code></td>
 * </tr>
 * </table>
 * 
 */
public class NormalDistribution extends ContinuousDistribution {

   public final double mean;
   public final double standardDeviation;

   private static final String MEAN = "Mean";
   private static final String STANDARD_DEVIATION = "StandardDeviation";

   /** as required by {@link ReflectionConstructable} */
   public static NormalDistribution createInstance(SimNode ownNode, Parameters pars, RandomNumberGenerator rng) {
      final double mean = pars.get(ownNode, MEAN).asDouble();
      final double standardDeviation = pars.get(ownNode, STANDARD_DEVIATION).asDouble();
      return new NormalDistribution(mean, standardDeviation, rng);
   }

   public NormalDistribution(double mean, double standardDeviation, RandomNumberGenerator rng) {
      super(rng);

      if (standardDeviation < 0) {
         throw new DistributionException("Standard deviation < 0!");
      }

      this.mean = mean;
      this.standardDeviation = standardDeviation;
   }

   public NormalDistribution(double mean, double standardDeviation) {
      this(mean, standardDeviation, getDefaultRNG());
   }

   @Override
   public double next() {
      double aux_term_1, aux_term_2;

      aux_term_1 = Math.sqrt(-2.0 * Math.log(rng.next()));
      aux_term_2 = Math.cos(Math.PI * rng.next());
      return mean + standardDeviation * aux_term_1 * aux_term_2;
   }
}
