/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions.continuous.statedep;

import ikr.simlib.distributions.continuous.ContinuousDistribution;
import ikr.simlib.distributions.continuous.NegExpDistribution;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.random.RandomNumberGenerator;

/**
 * Continuous-State GMPP<br>
 * 
 * <table border="0">
 * <tr>
 * <th align="left" valign="top">Meaning:</th>
 * <td>Modeling of sources where the state space is continuous.</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Description:</th>
 * <td>State machine with continuous (infinite) state space: <br>
 * <ul>
 * <li>
 * length of a state according to a continuous distribution
 * <li>
 * rate \(r\) in a state according to a continuous distribution
 * <li>
 * Poisson process with rate \(r\) within a state
 * </ul>
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parameters:</th>
 * <td>
 * <ul>
 * <li>
 * continuous distribution of the length of a phase
 * <li>
 * continuous distribution of the rate in a phase
 * </ul>
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Characteristic values:</th>
 * <td>Moments and distribution of the arrival intervals</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parser example:</th>
 * <td><code>
         [...].distribution = ikr.simlib.distributions.continuous.statedep.ContStateGMPPDistribution
         <br>
         [...].distribution.PhaseLengthDist = NegExp
         <br>
         [...].distribution.PhaseLengthDist.Mean = 2.3
         <br>
         [...].distribution.RateDist = Uniform
         <br>
         [...].distribution.RateDist.LowerBound = 1.5
         <br>
         [...].distribution.RateDist.UpperBound = 3.5
      </code></td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">References:</th>
 * <td>T. THEIMER: How to compute the moments of a GMDP, RACE 1022, Document
 * UST_123_0023_CD_CC, 1989.</td>
 * </tr>
 * </table>
 * 
 */
public class ContStateGMPPDistribution extends ContinuousDistribution {

   public final ContinuousDistribution phaseLengthDist;
   public final ContinuousDistribution rateDist;

   private static final String PHASE_LENGTH_DISTRIBUTION = "PhaseLengthDist";
   private static final String RATE_DISTRIBUTION = "RateDist";

   /** as required by {@link ReflectionConstructable} */
   public static ContStateGMPPDistribution createInstance(SimNode ownNode, Parameters pars, RandomNumberGenerator rng) {
      final ContinuousDistribution phaseLengthDist = pars.get(ownNode, PHASE_LENGTH_DISTRIBUTION).asInstance(ContinuousDistribution.class, ownNode.createChildNode(PHASE_LENGTH_DISTRIBUTION), pars, rng);
      final ContinuousDistribution rateDist = pars.get(ownNode, RATE_DISTRIBUTION).asInstance(ContinuousDistribution.class, ownNode.createChildNode(RATE_DISTRIBUTION), pars, rng);
      return new ContStateGMPPDistribution(phaseLengthDist, rateDist, rng);
   }

   private double residualPhaseTime;
   private NegExpDistribution currentIATDist;

   public ContStateGMPPDistribution(ContinuousDistribution phaseLengthDist, ContinuousDistribution rateDist, RandomNumberGenerator rng) {
      super(rng);
      this.phaseLengthDist = phaseLengthDist;
      this.rateDist = rateDist;
   }

   public ContStateGMPPDistribution(ContinuousDistribution phaseLengthDist, ContinuousDistribution rateDist) {
      this(phaseLengthDist, rateDist, getDefaultRNG());
   }

   @Override
   public double next() {
      double result = 0;
      double interArrivalTime;

      do {
         if (currentIATDist == null) {
            double currentRate = 0;
            do {
               currentRate = rateDist.next();
            } while (currentRate <= 0);
            currentIATDist = new NegExpDistribution(1 / currentRate);
            residualPhaseTime = phaseLengthDist.next();
         }

         interArrivalTime = currentIATDist.next();
         if (interArrivalTime < residualPhaseTime) {
            residualPhaseTime -= interArrivalTime;
            result += interArrivalTime;
         } else {
            result += residualPhaseTime;
            currentIATDist = null;
         }
      } while (currentIATDist == null);

      return result;
   }

   public double getResidualPhaseTime() {
      return residualPhaseTime;
   }

   public NegExpDistribution getCurrentIATDist() {
      return currentIATDist;
   }
}
