/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions.continuous.statedep;

import ikr.simlib.distributions.ProbabilityVector;
import ikr.simlib.distributions.continuous.ContinuousDistribution;
import ikr.simlib.distributions.discrete.DiscreteDistribution;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.random.RandomNumberGenerator;

/**
 * GMDP (Generally Modulated Determistic Process)<br>
 * 
 * <table border="0">
 * <tr>
 * <th align="left" valign="top">Meaning:</th>
 * <td>Modeling of sources with multiple states (usually 2 - 5) in ATM networks</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Description:</th>
 * <td>A GMDP comprises a state machine with \(m\) states: <br>
 * <br>
 * <ul>
 * <li>
 * In state \(i\) arrivals occur in constant intervals \(d_i\). If \(d_i<0\) it
 * is a silence state without arrivals. In that case \(-d_i\) describes the
 * virtual slot duration in this state.
 * <li>
 * \(X_i\) representing the number of arrivals in state \(i\) is arbitrarily
 * (discretely) distributed (often shifted geometrically -> MMDP); in the
 * silence state \(X_i\) means the number of time slots of the silence phase.
 * <li>
 * After \(X_i\) arrivals incl. an additional interval of the length \(d_i\) (or
 * after \(X_i\) slots of the length \(-d_i\) in the silence state), a
 * transition occurs to the state \(j\) with the probability \(p_{ij}\).
 * <li>
 * According to definition: \(p_{ii} = 0\) and \(\sum\limits_{i=1}^m p_{ij} =1\)
 * </ul>
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parameters:</th>
 * <td>
 * <ul>
 * <li>
 * number of states \(m\)
 * <li>
 * transition probabilities \(p_{ij}\)
 * <li>
 * distribution of the number of arrivals \(X_i\) in the individual states
 * <li>
 * arrival intervals \(d_i\) in the individual states
 * </ul>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Characteristic values:</th>
 * <td>Moments and distribution of the arrival intervals</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parser example:</th>
 * <td><code>
         [...].Dist = ikr.simlib.distributions.continuous.statedep.GMDPDistribution
         [...].Dist.States = 3
         <br>
         [...].Dist.PMAP = [ [0 0.5 0.5] [0.5 0 0.5] [0.5 0.5 0] ]
         <br>
         [...].Dist.PhaselengthDistribution_0 = ShiftedGeometric
         <br>
         [...].Dist.PhaselengthDistribution_0.Mean = 10
         <br>
         [...].Dist.PhaselengthDistribution_1 = ShiftedGeometric
         <br>
         [...].Dist.PhaselengthDistribution_1.Mean = 10
         <br>
         [...].Dist.PhaselengthDistribution_2 = ShiftedGeometric
         <br>
         [...].Dist.PhaselengthDistribution_2.Mean = 5
         <br>
         # negative value denotes silence state
         <br>
         [...].Dist.InterCellTimes = [ -1 2 4 ]
      </code></td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">References:</th>
 * <td>
 * <ul>
 * <li>
 * P. J. KHN: Reminder on queueing theory for ATM networks. Telecommunication
 * Systems, No. 5, 1996, pp. 1-24.
 * <li>
 * G. D. STAMOULIS, M. E. ANAGNOSTOU, A. D. GEORGANTAS: Traffic source models
 * for ATM networks: a survey. Computer Communications, Vol. 17, No. 6, Juni,
 * 1994.
 * <li>
 * T. THEIMER: How to compute the moments of a GMDP, RACE 1022, Document
 * UST_123_0023_CD_CC, 1989.
 * </ul>
 * </td>
 * </tr>
 * </table>
 * 
 */
public class GMDPDistribution extends ContinuousDistribution {

   private static final String STATES = "States";
   private static final String INTER_CELL_TIMES = "InterCellTimes";
   private static final String PMAP = "PMAP";
   private static final String PHASE_LENGHT_DISTRIBUTION = "PhaselengthDistribution_";

   /** as required by {@link ReflectionConstructable} */
   public static GMDPDistribution createInstance(SimNode ownNode, Parameters pars, RandomNumberGenerator rng) {
      final int noOfStates = pars.get(ownNode, STATES).asInteger();
      final double[] interCellTimes = pars.get(ownNode, INTER_CELL_TIMES).asDoubleVector();
      final double[][] transitionProbs = pars.get(ownNode, PMAP).asDoubleMatrix();

      assert interCellTimes.length == noOfStates;
      assert transitionProbs.length == noOfStates;
      for (int i = 0; i < transitionProbs.length; i++) {
         assert transitionProbs[i].length == noOfStates;
      }

      final DiscreteDistribution[] phaseLengthDists = new DiscreteDistribution[noOfStates];
      for (int i = 0; i < noOfStates; i++) {
         phaseLengthDists[i] = pars.get(ownNode, PHASE_LENGHT_DISTRIBUTION + i).asInstance(DiscreteDistribution.class, ownNode.createChildNode(PHASE_LENGHT_DISTRIBUTION + i), pars, rng);
      }

      return new GMDPDistribution(noOfStates, interCellTimes, phaseLengthDists, transitionProbs, rng);
   }

   public static class GMDPStateParameters {

      public final double interCellTime;
      public final DiscreteDistribution phaseLengthDistribution;
      public final ProbabilityVector cumulatedTransitionProbs;

      public GMDPStateParameters(double interCellTime, DiscreteDistribution phaseLengthDist, double[] transitionProbs) {
         this.interCellTime = interCellTime;
         phaseLengthDistribution = phaseLengthDist;
         cumulatedTransitionProbs = new ProbabilityVector(transitionProbs);
      }
   }

   public final int noOfStates;
   public final GMDPStateParameters[] stateParameters;

   private int currentState;
   private double currentInterCellTime;
   private int cellCounter;

   protected GMDPDistribution(int noOfStates, RandomNumberGenerator rng) {
      super(rng);
      this.noOfStates = noOfStates;
      stateParameters = new GMDPStateParameters[noOfStates];
   }

   protected GMDPDistribution(int noOfStates) {
      this(noOfStates, getDefaultRNG());
   }

   public GMDPDistribution(int noOfStates, double[] interCellTimes, DiscreteDistribution[] phaseLengthDists, double[][] transitionProbs, RandomNumberGenerator rng) {
      super(rng);
      this.noOfStates = noOfStates;
      stateParameters = new GMDPStateParameters[noOfStates];

      for (int i = 0; i < noOfStates; i++) {
         stateParameters[i] = new GMDPStateParameters(interCellTimes[i], phaseLengthDists[i], transitionProbs[i]);
      }
   }

   public GMDPDistribution(int noOfStates, double[] interCellTimes, DiscreteDistribution[] phaseLengthDists, double[][] transitionProbs) {
      this(noOfStates, interCellTimes, phaseLengthDists, transitionProbs, getDefaultRNG());
   }

   @Override
   public void reset() {
      cellCounter = 0;
      currentState = 0;
   }

   @Override
   public double next() {
      double silenceTime = 0.0;

      if (cellCounter <= 0) {
         do {
            currentInterCellTime = stateParameters[currentState].interCellTime;
            cellCounter = stateParameters[currentState].phaseLengthDistribution.next();
            if (currentInterCellTime < 0) {
               silenceTime -= currentInterCellTime * cellCounter;
            }
            int nextState = stateParameters[currentState].cumulatedTransitionProbs.next(rng.next());
            currentState = nextState;
         } while ((currentInterCellTime < 0) || (cellCounter <= 0));
      }

      cellCounter--;

      return silenceTime + currentInterCellTime;
   }

   public int getCurrentState() {
      return currentState;
   }

   public double getCurrentInterCellTime() {
      return currentInterCellTime;
   }

   public int getCellCounter() {
      return cellCounter;
   }
}
