/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions.continuous.statedep;

import ikr.simlib.distributions.DistributionException;
import ikr.simlib.distributions.ProbabilityVector;
import ikr.simlib.distributions.continuous.ContinuousDistribution;
import ikr.simlib.distributions.continuous.NegExpDistribution;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.random.RandomNumberGenerator;
import ikr.simlib.support.Vectors;

/**
 * MAP (Markovian Arrival Process)<br>
 * 
 * <table border="0">
 * <tr>
 * <th align="left" valign="top">Meaning:</th>
 * <td>Modeling of sources with multiple states e.g., at the call level</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Description:</th>
 * <td>
 * <ul>
 * <li>
 * State machine with \(m\) states. <br>
 * <li>
 * Continuous Markovian process with the transition rate \(q_{ij}\) from state
 * \(i\) to state \(j\), which is composed of two components: \(q_{ij} = c_{ij}
 * + d_{ij}\)
 * <li>
 * Transition rate from state \(i\) to state \(j \mbox{ } (j \neq i)\) without
 * an arrival event upon transition: \(c_{ij}\)
 * <li>
 * Transition rate from state \(i\) to state \(j\) with an arrival event upon
 * transition: \(d_{ij}\)
 * <li>
 * According to definition: \(c_{ii} = -\sum\limits_{i \neq i}
 * c_{ij}-\sum\limits_{i} d_{ij} \Rightarrow q_{ii} = -\sum\limits_{i \neq i}
 * q_{ij}\)
 * </ul>
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parameters:</th>
 * <td>- number of states \(m\) <br>
 * <br>
 * - transition rates without arrival \(c_{ij}\) <br>
 * <br>
 * - transition rates with arrival \(d_{ij}\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Characteristic values:</th>
 * <td>See Reference LUCANTONI</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parser example:</th>
 * <td>
 * <code>
         [...].distribution = ikr.simlib.distributions.continuous.statedep.MAPDistribution
         <br>
         [...].distribution.States = 2
         <br>
         [...].distribution.CMAP = [[-0.26 0.05] [0.05 -0.47]]
         <br>
         [...].distribution.DMAP = [[0.2 0.01] [0.02 0.4]]
      </code></td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">References:</th>
 * <td>
 * <ul>
 * <li>
 * P. J. KHN: Reminder on queueing theory for ATM networks. Telecommunication
 * Systems, No. 5, 1996, pp. 1-24.
 * <li>
 * D. M. LUCANTONI, K. S. MEIER-HELLSTERN, M. F. NEUTS: A Single-Server Queue
 * with Server Vacations and a Class of Non-Renewal Arrival Processes. Advances
 * in Applied Probability, Vol. 22, No. 1, 1989, pp. 676-705.
 * <li>
 * G. D. STAMOULIS, M. E. ANAGNOSTOU, A. D. GEORGANTAS: Traffic source models
 * for ATM networks: a survey. Computer Communications, Vol. 17, No. 6, Juni,
 * 1994.
 * <li>
 * D. M. LUCANTONI, K. S. MEIER-HELLSTERN, M. F. NEUTS: A Single-Server Queue
 * with Server Vacations and a Class of Non-Renewal Arrival Processes. Advances
 * in Applied Probability, Vol. 22, No. 1, 1989, pp. 676-705.
 * </ul>
 * </td>
 * </tr>
 * </table>
 * 
 */
public class MAPDistribution extends ContinuousDistribution {

   private static final String STATES = "States";
   private static final String CMAP = "CMAP";
   private static final String DMAP = "DMAP";

   /** as required by {@link ReflectionConstructable} */
   public static MAPDistribution createInstance(SimNode ownNode, Parameters pars, RandomNumberGenerator rng) {
      final int noOfStates = pars.get(ownNode, STATES).asInteger();
      final double[][] cMatrix = pars.get(ownNode, CMAP).asDoubleMatrix();
      final double[][] dMatrix = pars.get(ownNode, DMAP).asDoubleMatrix();
      return new MAPDistribution(noOfStates, cMatrix, dMatrix, rng);
   }

   public static class MAPStateParameters {
      /**
       * Distribution for time spent in state before transition or event occurs
       */
      public final NegExpDistribution idleTimeDist;
      /**
       * Probabilities that a event without a transition occurs under condition
       * that transition or event occurs
       */
      public final double eventNoTransProb;
      /**
       * Transition probabilities under condition that a transition happens
       */
      public final ProbabilityVector cumulatedTransitionProbs;

      /**
       * Probabilities that event occurs on transition to a certain state
       */
      public final double[] eventProbs;

      public MAPStateParameters(double idleTimeMean, double eventNoTransProb, double[] tranitionProbs, double[] eventProbs) {
         idleTimeDist = new NegExpDistribution(idleTimeMean);
         this.eventNoTransProb = eventNoTransProb;
         cumulatedTransitionProbs = new ProbabilityVector(tranitionProbs);
         this.eventProbs = new double[eventProbs.length];

         System.arraycopy(eventProbs, 0, this.eventProbs, 0, eventProbs.length);
      }
   }

   public final int noOfStates;
   public final MAPStateParameters[] stateParameters;

   private int currentState;

   public MAPDistribution(int noOfStates, double[][] c, double[][] d, RandomNumberGenerator rng) {
      super(rng);
      this.noOfStates = noOfStates;
      stateParameters = new MAPStateParameters[noOfStates];

      final double cEpsilon = 1e-10;

      for (int i = 0; i < noOfStates; i++) {
         double[] transitionProbs = Vectors.add(c[i], d[i]);
         double[] eventProbs = new double[noOfStates];
         double eventNoTransProb;
         int k;

         double sumOfRates = Vectors.sum(c[i]) - c[i][i] + Vectors.sum(d[i]);

         /** check whether parameters are correct */
         if (Math.abs(sumOfRates + c[i][i]) > cEpsilon) {
            throw new DistributionException("c_ii must be sum of (-c_ij (j#i) - d_ij)");
         }

         if (sumOfRates <= cEpsilon) {
            throw new DistributionException("sum of rates must be > 0 for each state");
         }

         for (k = 0; k < noOfStates; k++) {
            if (d[i][k] < 0) {
               throw new DistributionException("transition rate must be >= 0");
            }

            if (k != i) {
               if (d[i][k] < 0) {
                  throw new DistributionException("transition rate must be >= 0");
               }
            }
         }

         if (d[i][i] < sumOfRates) {
            transitionProbs[i] = 0;
            Vectors.normalize(transitionProbs);
            eventNoTransProb = d[i][i] / sumOfRates;
         } else {
            transitionProbs[i] = 1;
            eventNoTransProb = 1;
         }

         for (k = 0; k < noOfStates; k++) {
            double sum = c[i][k] + d[i][k];
            if (sum > 0) {
               eventProbs[k] = d[i][k] / sum;
            } else {
               eventProbs[k] = 0;
            }
         }

         stateParameters[i] = new MAPStateParameters(1 / sumOfRates, eventNoTransProb, transitionProbs, eventProbs);
      }

      reset();
   }

   public MAPDistribution(int noOfStates, double[][] c, double[][] d) {
      this(noOfStates, c, d, getDefaultRNG());
   }

   @Override
   public double next() {
      double silenceSum = 0.0;
      boolean eventFlag;

      do {
         silenceSum += stateParameters[currentState].idleTimeDist.next();
         if (rng.next() <= stateParameters[currentState].eventNoTransProb) {
            eventFlag = true;
         } else {
            int nextState = stateParameters[currentState].cumulatedTransitionProbs.next(rng.next());
            eventFlag = (rng.next() <= stateParameters[currentState].eventProbs[nextState]);
            currentState = nextState;
         }
      } while (!eventFlag);

      return silenceSum;

   }

   @Override
   public void reset() {
      currentState = 0;
   }

   public int getCurrentState() {
      return currentState;
   }
}
