/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions.continuous.statedep;

import ikr.simlib.distributions.DistributionException;
import ikr.simlib.distributions.continuous.ContinuousDistribution;
import ikr.simlib.distributions.discrete.DiscreteDistribution;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.random.RandomNumberGenerator;

import java.util.ArrayDeque;
import java.util.Queue;

/**
 * On-Off Infinity Model<br>
 * 
 * <table border="0">
 * <tr>
 * <th align="left" valign="top">Meaning:</th>
 * <td>Superposition of infinitely many on-off sources which allows the
 * modelling of aggregated traffic. Bursts of traffic are generated according to
 * an interarrival time distribution. The burst length distribution represents
 * the length in number of packets which are sent within a burst with the
 * specified intercell time. Usually a negative-exponential distribution is used
 * as burst interarrival time distribution leading to an \(M/G/\infty\) source
 * model.</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Description:</th>
 * <td>Traffic is characterized by
 * <ul>
 * <li>
 * burst length distribution
 * <li>
 * burst interarrival time
 * <li>
 * inter cell time \(d\)
 * </ul>
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parameters:</th>
 * <td>
 * <ul>
 * <li>
 * (discrete) distribution of the burst length \(X\) in number of packets
 * <li>
 * (continuous) distribution of burst interarrival time \(T_A\)
 * <li>
 * inter-cell time \(d\)
 * </ul>
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Characteristic values:</th>
 * <td>mean (packet interarrival time): \(\mu = \frac{E[T_A]}{E[X]}\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parser example:</th>
 * <td>
 * <code>
         [...].Dist = ikr.simlib.distributions.continuous.statedep.OnOffInfinityDistribution
         <br>
         [...].Dist.BurstLengthDist = Geometric
         <br>
         [...].Dist.BurstLengthDist.Mean = 2.3
         <br>
         [...].Dist.BurstIATDist = NegExp
         <br>
         [...].Dist.BurstIATDist.Mean = 2.88
         <br>
         [...].Dist.BurstIATDist.InterCellTime = 1.0
      </code></td>
 * </tr>
 * </table>
 * 
 */
public class OnOffInfinityDistribution extends ContinuousDistribution {

   private static final String BURST_LENGTH_DISTRIBUTION = "BurstLengthDist";
   private static final String BURST_IAT_DISTRIBUTION = "BurstIATDist";
   private static final String INTER_CELL_TIME = "InterCellTime";

   /** as required by {@link ReflectionConstructable} */
   public static OnOffInfinityDistribution createInstance(SimNode ownNode, Parameters pars, RandomNumberGenerator rng) {
      final DiscreteDistribution burstLengthDist = pars.get(ownNode, BURST_LENGTH_DISTRIBUTION).asInstance(DiscreteDistribution.class, ownNode.createChildNode(BURST_LENGTH_DISTRIBUTION), pars, rng);
      final ContinuousDistribution burstIATDist = pars.get(ownNode, BURST_IAT_DISTRIBUTION).asInstance(ContinuousDistribution.class, ownNode.createChildNode(BURST_IAT_DISTRIBUTION), pars, rng);
      final double interCellTime = pars.get(ownNode, INTER_CELL_TIME).asDouble();
      return new OnOffInfinityDistribution(burstLengthDist, burstIATDist, interCellTime, rng);
   }

   public static class BurstData {

      public double nextArrivalTime;
      public int residualLength;

      public BurstData(double burstArrivalTime, int length) {
         nextArrivalTime = burstArrivalTime;
         residualLength = length;
      }
   }

   public final DiscreteDistribution burstLengthDist;
   public final ContinuousDistribution burstIATDist;
   public final double interCellTime;
   public final Queue<BurstData> burstList;

   private double currentTime;
   private BurstData nextBurst;

   public OnOffInfinityDistribution(DiscreteDistribution burstLengthDist, ContinuousDistribution burstIATDist, double interCellTime, RandomNumberGenerator rng) {
      super(rng);

      if (interCellTime < 0) {
         throw new DistributionException("intercell time < 0");
      }

      this.burstLengthDist = burstLengthDist;
      this.burstIATDist = burstIATDist;
      this.interCellTime = interCellTime;
      currentTime = 0;
      burstList = new ArrayDeque<BurstData>();

      nextBurst = new BurstData(burstIATDist.next(), burstLengthDist.next());
   }

   public OnOffInfinityDistribution(DiscreteDistribution burstLengthDist, ContinuousDistribution burstIATDist, double interCellTime) {
      this(burstLengthDist, burstIATDist, interCellTime, getDefaultRNG());
   }

   @Override
   public double next() {
      BurstData currentBurst;

      if (!burstList.isEmpty() && burstList.element().nextArrivalTime <= nextBurst.nextArrivalTime) {
         currentBurst = burstList.poll();
      } else {
         currentBurst = nextBurst;
         nextBurst = new BurstData(currentBurst.nextArrivalTime + burstIATDist.next(), burstLengthDist.next());
      }

      double result = currentBurst.nextArrivalTime - currentTime;
      currentTime = currentBurst.nextArrivalTime;
      if (currentBurst.residualLength > 1) {
         currentBurst.residualLength--;
         currentBurst.nextArrivalTime += interCellTime;
         burstList.add(currentBurst);
      }

      return result;
   }

   public double getCurrentTime() {
      return currentTime;
   }

   public BurstData getNextBurst() {
      return nextBurst;
   }
}
