/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions.continuous.statedep;

import ikr.simlib.distributions.continuous.ContinuousDistribution;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.random.RandomNumberGenerator;

/**
 * Video Source <br>
 * 
 * <table border="0">
 * <tr>
 * <th align="left" valign="top">Meaning:</th>
 * <td>Modeling of video sources in ATM/packet networks</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Description:</th>
 * <td>
 * <ul>
 * <li>
 * Parameters: frame duration \(d_F\), frame size \(s_F\), packet size \(s_P\)
 * <li>
 * Autoregressive process with the bit rate \(\lambda_i\) in frame \(i\) in
 * bits/pixel: \(\lambda_i = (a \cdot \lambda_{i-1} + b \cdot \varepsilon) \cdot
 * r(0,\lambda_{max})\) with the constants \(a\) and \(b\) as well as the
 * function \(r\) to limit the rate to the interval \([0,\lambda_max]\).
 * <li>
 * Usually normal distribution for \(\varepsilon\)
 * </ul>
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parameters:</th>
 * <td>
 * <ul>
 * <li>
 * frame size \(s_F\)
 * <li>
 * frame duration \(d_F\)
 * <li>
 * start rate \(\lambda_0\)
 * <li>
 * constants \(a\) and \(b\)
 * <li>
 * maximum rate \(\lambda_{max}\)
 * <li>
 * distribution \(\varepsilon\)
 * <li>
 * packet size \(s_P\)
 * </ul>
 * </td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Characteristic values:</th>
 * <td>Average packet intervals within the frame \(i: E[T] = \frac{s_P \cdot
 * d_F}{s_F \cdot \lambda_i}\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parser example:</th>
 * <td>
 * <code>
         [...].dist = ikr.simlib.distributions.continuous.statedep.VideoSourceDistribution
         <br>
         [...].dist.StartRate = 0.5333333 #
         bits/pixel
         <br>
         [...].dist.UpperlimitRate = 1.0666667 # max.
         bits/pixel
         <br>
         [...].dist.RateDistribution = Normal
         <br>
         [...].dist.RateDistribution.Mean = 0.587 # bits/pixel
         <br>
         [...].dist.RateDistribution.StandardDeviation = 1
         <br>
         [...].dist.FrameDuration = 33333 # inter-picture time
         <br>
         [...].dist.FrameSize = 250000 # 250 000 pixel/picture
         <br>
         [...].dist.PacketSize = 384 # ATM cell 48 * 8 bits
         <br>
         [...].dist.ConstantA = 0.8781 # factor for last rate
         <br>
         [...].dist.ConstantB = 0.1108 # factor for new rate
      </code></td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">References:</th>
 * <td>
 * <ul>
 * <li>
 * W. Held, 2. Semesterarbeit (950)
 * <li>
 * B. Maglaris et al., Performance Models of Statistical Multiplexing in Packet
 * Video Communications, IEEE Transactions on Communications, Vol.36, No. 7,
 * 07/88
 * </ul>
 * </td>
 * </table>
 */
public class VideoSourceDistribution extends ContinuousDistribution {

   private static final String FRAME_SIZE = "FrameSize";
   private static final String FRAME_DURATION = "FrameDuration";
   private static final String START_RATE = "StartRate";
   private static final String CONSTANT_A = "ConstantA";
   private static final String CONSTANT_B = "ConstantB";
   private static final String UPPER_LIMIT_RATE = "UpperLimitRate";
   private static final String RATE_DISTRIBUTION = "RateDistribution";
   private static final String PACKET_SIZE = "PacketSize";

   /** as required by {@link ReflectionConstructable} */
   public static VideoSourceDistribution createInstance(SimNode ownNode, Parameters pars, RandomNumberGenerator rng) {
      final int frameSize = pars.get(ownNode, FRAME_SIZE).asInteger();
      final double frameDuration = pars.get(ownNode, FRAME_DURATION).asDouble();
      final double startRate = pars.get(ownNode, START_RATE).asDouble();
      final double constantA = pars.get(ownNode, CONSTANT_A).asDouble();
      final double constantB = pars.get(ownNode, CONSTANT_B).asDouble();
      final double upperLimitRate = pars.get(ownNode, UPPER_LIMIT_RATE).asDouble();
      final ContinuousDistribution rateDistribution = pars.get(ownNode, RATE_DISTRIBUTION).asInstance(ContinuousDistribution.class, ownNode.createChildNode(RATE_DISTRIBUTION), pars, rng);
      final int packetSize = pars.get(ownNode, PACKET_SIZE).asInteger();
      return new VideoSourceDistribution(frameSize, frameDuration, startRate, constantA, constantB, upperLimitRate, rateDistribution, packetSize, rng);
   }

   /**
    * e.g. 250000 pixels/picture
    */
   public final int frameSize;

   /**
    * e.g. 33333 us -> 30 pictures/sec
    */
   public final double frameDuration;

   /**
    * e.g. 0.52 bits/pixel
    */
   public final double startRate;

   /**
    * e.g. 0.8781
    */
   public final double constantA;

   /**
    * e.g. 0.1108
    */
   public final double constantB;

   /**
    * e.g. 1.41 bits/pixel
    */
   public final double upperLimitRate;

   /**
    * e.g. normal distribution with mean 0.572 bits/pixel
    */
   public final ContinuousDistribution rateDistribution;

   /**
    * e.g. 384 bits/cell
    */
   public final int packetSize;

   private double actRate;
   private int frameCounter;
   private int packetCounter;
   private double videoInterCellTime;

   public int generatePackets() {
      double lastrate;

      lastrate = actRate;
      actRate = Math.max(constantA * lastrate + constantB * rateDistribution.next(), 0);
      if ((upperLimitRate > -1) && (actRate > upperLimitRate))
         actRate = upperLimitRate;
      frameCounter++;
      return ((int) Math.rint(frameSize * actRate) + ((int) Math.rint(frameSize * lastrate) % packetSize)) / packetSize;
   }

   public VideoSourceDistribution(int framesize, double frameduration, double startrate, double constanta, double constantb, double upperlimitrate, ContinuousDistribution ratedistribution, int packetsize, RandomNumberGenerator rng) {
      super(rng);

      frameSize = framesize;
      frameDuration = frameduration;
      startRate = startrate;
      constantA = constanta;
      constantB = constantb;
      upperLimitRate = upperlimitrate;
      rateDistribution = ratedistribution;
      packetSize = packetsize;
      actRate = startrate;
      frameCounter = 0;
      packetCounter = 0;
      videoInterCellTime = 0;
   }

   public VideoSourceDistribution(int framesize, double frameduration, double startrate, double constanta, double constantb, double upperlimitrate, ContinuousDistribution ratedistribution, int packetsize) {
      this(framesize, frameduration, startrate, constanta, constantb, upperlimitrate, ratedistribution, packetsize, getDefaultRNG());
   }

   @Override
   public double next() {
      double aux;

      if (packetCounter > 0) {
         packetCounter--;
         return videoInterCellTime;
      } else {
         aux = 0;
         do {
            packetCounter = generatePackets();
            if (packetCounter > 0)
               videoInterCellTime = frameDuration / packetCounter;
            else
               aux += frameDuration;
         } while (packetCounter <= 0);
         return aux + videoInterCellTime;
      }
   }

   @Override
   public void reset() {
      actRate = startRate;
      frameCounter = 0;
      packetCounter = 0;
      videoInterCellTime = 0;
   }

   public double getActRate() {
      return actRate;
   }

   public int getFrameCounter() {
      return frameCounter;
   }

   public int getPacketCounter() {
      return packetCounter;
   }

   public double getVideoInterCellTime() {
      return videoInterCellTime;
   }
}
