/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions.discrete;

import ikr.simlib.distributions.DistributionException;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.random.RandomNumberGenerator;

/**
 * Bernoulli Distribution<br>
 * 
 * <table border="0">
 * <tr>
 * <th align="left" valign="top">Meaning:</th>
 * <td>Single random experiment with the success probability \(q (0 \le q \le
 * 1)\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parameters:</th>
 * <td>Success probability = mean value \(q\)</td>
 * </tr>
 * <tr>
 * <td><i>Distribution:</i></td>
 * <td>\(P(X=i) = p_i = \begin{cases} 1-q &\mbox{for } i=0 \\ q &\mbox{for } i=1
 * \\ 0 & \mbox{else } \end{cases} \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Expected value:</th>
 * <td>\(E[X]=q\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Variance:</th>
 * <td>\(VAR[X]=q(1-q)\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Coefficient of variation:</th>
 * <td>\(c_T=\sqrt{ \frac{ 1-q }{q} }\)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Generating func.:</th>
 * <td>\(G(z)=1-q+qz \)</td>
 * </tr>
 * <tr>
 * <th align="left" valign="top">Parser example:</th>
 * <td><code>
         [...].Distribution = ikr.simlib.distributions.discrete.BernoulliDistribution
         <br>
         [...].Distribution.mean = 0.6</code></td>
 * </tr>
 * </table>
 * 
 */
public class BernoulliDistribution extends DiscreteDistribution {

   private static final String MEAN = "Mean";

   /** as required by {@link ReflectionConstructable} */
   public static BernoulliDistribution createInstance(SimNode ownNode, Parameters pars, RandomNumberGenerator rng) {
      final double mean = pars.get(ownNode, MEAN).asDouble();
      return new BernoulliDistribution(mean, rng);
   }

   public final double mean;

   public BernoulliDistribution(double mean, RandomNumberGenerator rng) {
      super(rng);

      if (mean < 0 || mean > 1) {
         throw new DistributionException("Mean < 0 or > 1");
      }

      this.mean = mean;
   }

   public BernoulliDistribution(double mean) {
      this(mean, getDefaultRNG());
   }

   @Override
   public int next() {
      return (rng.next() < mean) ? 1 : 0;
   }
}
