/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions.discrete;

import ikr.simlib.distributions.DistributionException;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.random.RandomNumberGenerator;

/**
 * Zipf-Mandelbrot Distribution
 * 
 * <pre>
 * p_k=C*(v+k)&circ;(-q), k=0..N-1, q&gt;1, v&gt;0
 * C: normalization constant, C=1/sum((v+k)&circ;(-q),k=0..N-1)
 * </pre>
 * 
 * @see "Hoermann, W. and Derflinger, G.: &quot;Rejection-inversion to generate variates from monotone discrete distributions&quot;, ACM Transactions on Modeling and Computer Simulation (TOMACS), July 1996, Vol. 6, Iss. 3, pp. 169-184"
 * 
 */
public class ZipfMandelbrotDistribution extends DiscreteDistribution {

   public final double Q;
   public final double V;
   public final double IMax;

   public final double Hx0;
   public final double S;

   private static final String Q_VALUE = "Q";
   private static final String N_VALUE = "N";
   private static final String V_VALUE = "V";

   /** as required by {@link ReflectionConstructable} */
   public static ZipfMandelbrotDistribution createInstance(SimNode ownNode, Parameters pars, RandomNumberGenerator rng) {
      final double q = pars.get(ownNode, Q_VALUE).asDouble();
      final int n = pars.get(ownNode, N_VALUE).asInteger();
      final double v = pars.get(ownNode, V_VALUE).asDouble();
      return new ZipfMandelbrotDistribution(q, n, v, rng);
   }

   public ZipfMandelbrotDistribution(double pQ, int pN, double pV, RandomNumberGenerator rng) {
      super(rng);

      if (pQ <= 1.0) {
         throw new DistributionException("q must be greater 1");
      }
      if (pV <= 0) {
         throw new DistributionException("v must be greater 0");
      }
      if (pN == 0) {
         throw new DistributionException("N must be greater 0");
      }

      Q = pQ;
      V = pV;

      IMax = (double) pN - 1;

      Hx0 = H(0.5) - Math.exp(Math.log(V) * (-Q));
      S = 1.0 - HInv(H(1.5) - Math.exp(Math.log(V + 1.0) * (-Q)));
   }

   public ZipfMandelbrotDistribution(double pQ, int pN, double pV) {
      this(pQ, pN, pV, getDefaultRNG());
   }

   public ZipfMandelbrotDistribution(double pQ, int pN, RandomNumberGenerator rng) {
      this(pQ, pN, 1, rng);
   }

   public ZipfMandelbrotDistribution(double pQ, int pN) {
      this(pQ, pN, 1, getDefaultRNG());
   }

   private double H(double pX) {
      return Math.exp((1.0 - Q) * Math.log(V + pX)) * (1.0 / (1.0 - Q));
   }

   private double HInv(double pX) {
      return -V + Math.exp((1.0 / (1.0 - Q)) * Math.log((1.0 - Q) * pX));
   }

   @Override
   public int next() {
      while (true) {
         double U = rng.next();

         U = H(IMax + 0.5) + U * (Hx0 - H(IMax + 0.5));
         double X = HInv(U);
         double K = Math.floor(X + 0.5);
         if ((K - X) <= S) {
            return (int) K;
         } else if (U >= (H(K + 0.5) - Math.exp(-Math.log(V + K) * Q))) {
            return (int) K;
         }
      }
   }
}
