/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.distributions.mapped;

import ikr.simlib.distributions.DistributionException;
import ikr.simlib.random.RandomNumberGenerator;

import java.util.LinkedHashMap;
import java.util.Map;

/**
 * This distribution contains items to which a probability can be assigned.
 * <code>next</code> returns one of this items randomly chosen due to the
 * probabilities.
 */
public class MappedGeneralDistribution<I> extends MappedDistribution<I> {
   private final static double EPSILON = 1e-9;

   private final Map<I, Double> probabilityMap;

   public MappedGeneralDistribution(RandomNumberGenerator rng) {
      super(rng);
      probabilityMap = new LinkedHashMap<I, Double>();
   }

   public MappedGeneralDistribution() {
      this(getDefaultRNG());
   }

   private void checkRange() {
      final double sum = getSum();
      if (sum < 1 - EPSILON || sum > 1 + EPSILON) {
         throw new DistributionException("sum " + sum + " is not 1.0");
      }
   }

   @Override
   public I next() {
      checkRange();

      final double randomNumber = rng.next();
      double sum = 0.0;
      I lastItem = null;

      for (Map.Entry<I, Double> e : probabilityMap.entrySet()) {
         sum += e.getValue();
         lastItem = e.getKey();
         if (randomNumber <= sum) {
            return lastItem;
         }
      }
      // handle numeric inaccuracies: sum in [1-EPSILON, 1)
      assert sum >= 1 - EPSILON;
      assert lastItem != null;
      return lastItem;
   }

   public void bind(I item, double prob) {
      assert item != null;
      if (prob < 0 || prob > 1) {
         throw new DistributionException("probability " + prob + " is not in [0,1]");
      }

      if (probabilityMap.containsKey(item)) {
         throw new DistributionException("Item already in list");
      }

      probabilityMap.put(item, prob);

      assert getSum() < 1 + EPSILON;
   }

   public void rebind(I item, double prob) {
      assert item != null;
      if (prob < 0 || prob > 1) {
         throw new DistributionException("probability " + prob + " is not in [0,1]");
      }

      if (!probabilityMap.containsKey(item)) {
         throw new DistributionException("Item not available");
      }

      probabilityMap.put(item, prob);

      assert getSum() < 1 + EPSILON;
   }

   public void unbind(I item) {
      if (!probabilityMap.containsKey(item)) {
         throw new DistributionException("Item not available");
      }

      probabilityMap.remove(item);
   }

   public void normalize() {
      if (probabilityMap.isEmpty()) {
         throw new DistributionException("cannot normalize because the map is empty");
      }

      final double sum = getSum();

      if (sum < 0 || sum > 1) {
         throw new DistributionException("sum " + sum + " is not in [0,1]");
      }

      for (Map.Entry<I, Double> e : probabilityMap.entrySet()) {
         e.setValue(e.getValue() / sum);
      }
      checkRange();
   }

   public void equalize() {
      if (probabilityMap.isEmpty()) {
         return;
      }

      final double prob = 1 / (double) probabilityMap.size();

      for (Map.Entry<I, Double> e : probabilityMap.entrySet()) {
         e.setValue(prob);
      }
      checkRange();
   }

   public double getSum() {
      double sum = 0.0;

      for (Double v : probabilityMap.values()) {
         sum += v;
      }

      return sum;
   }

   public int getSize() {
      return probabilityMap.size();
   }
}
