/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.entities.branches;

import ikr.simlib.entities.Entity;
import ikr.simlib.exceptions.SimulationException;
import ikr.simlib.messages.Message;
import ikr.simlib.model.SimNode;
import ikr.simlib.ports.input.InputPort;
import ikr.simlib.ports.output.SynchronousOutputPort;
import ikr.simlib.random.RandomNumberGenerator;

/**
 * A branch element routes an incoming message randomly to one of its output
 * ports. A probability can be assigned to each output port.
 * 
 * <p>
 * <code>BinaryBranch</code> is a special branch element with two ports. The
 * probability of the first port ("output 1") is submitted to the constructor.
 * </p>
 * 
 * @see Branch
 */
public class BinaryBranch extends Entity {

   protected InputPort inputPort;
   protected SynchronousOutputPort firstOutputPort;
   protected SynchronousOutputPort secondOutputPort;
   protected double firstProbability;
   protected RandomNumberGenerator rng;

   protected SynchronousOutputPort getRandomPort() {
      if (rng.next() <= firstProbability)
         return firstOutputPort;
      else
         return secondOutputPort;
   }

   public BinaryBranch(double firstProb, SimNode ownNode, RandomNumberGenerator rng) {
      super(ownNode);

      inputPort = new InputPort(this) {
         @Override
         protected void handleMessageIndication(Message msg) {
            getRandomPort().sendMessage(inputPort.getMessage());
         }
      };

      firstOutputPort = new SynchronousOutputPort(this, "output 1");
      secondOutputPort = new SynchronousOutputPort(this, "output 2");
      firstProbability = firstProb;
      this.rng = rng;
   }

   public BinaryBranch(double firstProb, SimNode ownNode) {
      this(firstProb, ownNode, RandomNumberGenerator.getSystemRNG());
   }

   public void setFirstProbability(double prob) {
      if (prob >= 0 && prob <= 1)
         firstProbability = prob;
      else
         throw new SimulationException("Probability = " + prob);
   }

   public void equalize() {
      firstProbability = 0.5;
   }

   public SynchronousOutputPort getFirstOutputPort() {
      return firstOutputPort;
   }

   public SynchronousOutputPort getSecondOutputPort() {
      return secondOutputPort;
   }

   public InputPort getInputPort() {
      return inputPort;
   }
}
