/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.entities.branches;

import ikr.simlib.distributions.DistributionException;
import ikr.simlib.exceptions.SimulationException;
import ikr.simlib.messages.Message;
import ikr.simlib.model.SimNode;
import ikr.simlib.ports.output.SynchronousOutputPort;
import ikr.simlib.random.RandomNumberGenerator;

/**
 * A branch element routes an incoming message randomly to one of its output
 * ports. A probability can be assigned to each output port.
 * 
 * <p>
 * <code>Branch</code> has a fixed initial number of ports that can be defined
 * on construction. The port probability can be defined by
 * <code>SetProbability</code>. The ports are called "output 1", "output 2", ...
 * . Moreover, it is possible to add and remove ports dynamically.
 * </p>
 * 
 * @see Expander
 * @see BinaryBranch
 */
public class Branch extends Expander<Double> {

   protected boolean isNormalized;
   protected RandomNumberGenerator rng;

   public Branch(int nrOfPorts, SimNode ownNode) {
      super(nrOfPorts, 0.0, ownNode);

      rng = RandomNumberGenerator.getSystemRNG();
   }

   public Branch(SimNode ownNode) {
      super(ownNode);

      rng = RandomNumberGenerator.getSystemRNG();
   }

   protected SynchronousOutputPort getRandomPort() {
      double randomNumber = rng.next();
      double sum = 0.0;
      int index = -1;

      if (!isNormalized) {
         throw new SimulationException("Sum of prababilities is not 1.");
      }
      for (int i = 0; i < outputPorts.size(); i++) {
         sum += portParameters.get(i);
         if (randomNumber <= sum) {
            index = i;
            break;
         }
      }

      if (index == -1) {
         throw new DistributionException("No Item available");
      }

      return outputPorts.get(index);
   }

   @Override
   public void handleMessageIndication(Message msg) {
      getRandomPort().sendMessage(inputPort.getMessage());
   }

   @Override
   public void addPort(Double prob) {
      if (prob >= 0 && prob <= 1) {
         super.addPort(prob);
         isNormalized = getProbabilitySum() == 1;
      } else {
         throw new SimulationException("Probability = " + prob);
      }
   }

   @Override
   public void addPort(String portName, Double prob) {
      if (prob >= 0 && prob <= 1) {
         super.addPort(portName, prob);
         isNormalized = getProbabilitySum() == 1;
      } else {
         throw new SimulationException("Probability = " + prob);
      }
   }

   @Override
   public void removePort(String portName) {
      super.removePort(portName);
      isNormalized = (getProbabilitySum() == 1);

   }

   public void setProbability(int portNr, double prob) {
      if (prob >= 0 && prob <= 1) {
         super.setPortParameter(portNr, prob);
         isNormalized = (getProbabilitySum() == 1);
      } else {
         throw new SimulationException("Probability = " + (prob));
      }
   }

   public double getProbabilitySum() {
      double sum = 0.0;

      for (double d : portParameters) {
         sum += d;
      }

      return sum;
   }

   public void normalize() {
      double sum = getProbabilitySum();

      for (int i = 0; i < outputPorts.size(); i++) {
         portParameters.set(i, portParameters.get(i) / sum);
      }

      isNormalized = true;
   }

   public void equalize() {
      for (int i = 0; i < outputPorts.size(); i++) {
         portParameters.set(i, 1.0 / outputPorts.size());
      }
   }

}
