/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.entities.demultiplexers;

import ikr.simlib.entities.Entity;
import ikr.simlib.messages.Message;
import ikr.simlib.model.SimNode;
import ikr.simlib.ports.Port;
import ikr.simlib.ports.input.InputPort;
import ikr.simlib.ports.output.OutputPort;

/**
 * The class <code>Demultiplexer</code> is the counter part to {@link ikr.simlib.entities.multiplexers.Multiplexer} and is the base class for
{@link ikr.simlib.entities.demultiplexers.StdDemultiplexer}, {@link ikr.simlib.entities.demultiplexers.LabelDemultiplexer}, and {@link ikr.simlib.entities.demultiplexers.IPDemultiplexer}. An arriving
message at the input port ({@link ikr.simlib.ports.input.InputPort}) is passed on to one of the output ports
({@link ikr.simlib.ports.output.OutputPort}) which is determined by calling the method <code>getOutputPort()</code>.
Whereas the class {@link ikr.simlib.entities.demultiplexers.StdDemultiplexer} does not provide an implementation of the method
<code>getOutputPort()</code>, the classes {@link ikr.simlib.entities.demultiplexers.LabelDemultiplexer}, and {@link ikr.simlib.entities.demultiplexers.IPDemultiplexer} are
ready-to-use. In {@link ikr.simlib.entities.demultiplexers.StdDemultiplexer} the initial number of output ports is given on construction.
Outputs are named: "output 1" ... "output N". Further ports may be added via
<code>addPort()</code> and removed via <code>removePort()</code>.<br>
<br>
- The decision which message will reach which port is made with the method
<code>getOutputPort()</code>, that must be defined in the derived classes. In order for this decision
to be consistent throughout the lifetime of the demultiplexer, {@link ikr.simlib.entities.demultiplexers.LabelDemultiplexer}
and {@link ikr.simlib.entities.demultiplexers.IPDemultiplexer} are not allowed to use <code>removePort()</code>.<br>
<br>
- {@link ikr.simlib.entities.demultiplexers.LabelDemultiplexer} implements the method <code>getOutputPort()</code> by returning the
port with the index label of a <code>LabelMessage</code> (contained in the data member label of
<code>LabelMessage</code>) minus the start label (contained in startLabel). The start label can
be given as parameter on construction and defaults to 0.<br>
<br>
- {@link ikr.simlib.entities.demultiplexers.IPDemultiplexer} is a ready-to use static demultiplexer for packets belonging to the
class <code>IPPacket</code> and destinguishes between the modes "destination-based", "group-based"
and "type-based". The functionallity is the same as described in {@link ikr.simlib.entities.demultiplexers.LabelDemultiplexer}
except that not the label is taken into consideration but destination, group or type, depending
on the mode that is set. Therefore, additional to the parameters described for <code>StaticLabelDemux</code>,
a parameter indicating the mode has to be provided upon construction.<br>
<br>
- {@link ikr.simlib.entities.branches.Expander} can also be used to define a demultiplexer. The output ports of
this demultiplexer are not allowed to be blocked.
 */
public abstract class Demultiplexer extends Entity {

   protected final InputPort inputPort;

   protected abstract Port determineOutputPort(Message msg);

   public Demultiplexer(SimNode ownNode) {
      super(ownNode);

      inputPort = new InputPort(this) {
         @Override
         protected void handleMessageIndication(Message msg) {
            determineOutputPort(msg).messageIndication(msg);
         }
      };
   }

   public InputPort getInput() {
      return inputPort;
   }

   public abstract OutputPort getOutput(int index);

   protected Message handleGetMessage(Port port) {
      return inputPort.getMessage();
   }

   protected boolean handleIsMessageAvailable(Port port) {
      return inputPort.isMessageAvailable();
   }
}
