/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.entities.demultiplexers;

import ikr.simlib.exceptions.SimulationException;
import ikr.simlib.messages.LabelMessage;
import ikr.simlib.messages.Message;
import ikr.simlib.model.SimNode;
import ikr.simlib.ports.Port;

/**
 * <code>LabelDemultiplexer</code> provides the method
 * <code>GetOutputPort</code> which returns an output port dependent on the
 * label contained in a label message. An offset is considered if a start label
 * is given. If a modulo &gt; 1 is given the lower bits are not significant,
 * i.e. the port index returned is given by (label - start label) div modulus.
 */
public class LabelDemultiplexer extends StdDemultiplexer {

   protected final int startId;
   protected final int modulus;

   @Override
   protected Port determineOutputPort(Message msg) {
      LabelMessage lMsg = (LabelMessage) msg;
      int label = lMsg.getLabel();

      if (label < startId || label >= (startId + outputPorts.size() * modulus)) {
         throw new SimulationException("label out of range:" + label);
      }

      return outputPorts.get((label - startId) / modulus);
   }

   public LabelDemultiplexer(int noOfPorts, SimNode ownNode, int startLabel, int labelModulus) {
      super(noOfPorts, ownNode);

      startId = startLabel;
      modulus = labelModulus;
   }

   public LabelDemultiplexer(int noOfPorts, SimNode ownNode, int startLabel) {
      this(noOfPorts, ownNode, startLabel, 1);
   }

   public LabelDemultiplexer(int noOfPorts, SimNode ownNode) {
      this(noOfPorts, ownNode, 0, 1);
   }

   public LabelDemultiplexer(SimNode ownNode, int startLabel, int labelModulus) {
      this(0, ownNode, startLabel, labelModulus);
   }

   public LabelDemultiplexer(SimNode ownNode, int startLabel) {
      this(0, ownNode, startLabel, 1);
   }

   public LabelDemultiplexer(SimNode ownNode) {
      this(0, ownNode, 0, 1);
   }

   /**
    * Not supported by this implementation.
    */
   @Override
   public void removePort(String portName) {
      throw new SimulationException("Method not supported to ensure consistency");
   }
}
