/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.entities.phases;

import ikr.simlib.entities.Entity;
import ikr.simlib.events.CalendarCallback;
import ikr.simlib.events.calendar.Calendar;
import ikr.simlib.events.time.Duration;
import ikr.simlib.events.time.PointInTime;
import ikr.simlib.messages.Message;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable2;
import ikr.simlib.ports.input.InputPort;
import ikr.simlib.ports.output.OutputPort;
import ikr.simlib.ports.output.SynchronousOutputPort;
import ikr.simlib.statistics.integral.IntegralStatistic;
import ikr.simlib.statistics.integral.StdIntegralStatistic;

/**
 * The abstract base class <code>Phase</code> derived from
 * {@link ikr.simlib.entities.Entity}. The class
 * {@link ikr.simlib.entities.phases.StdPhase} is used to model service units.
 * The service time of {@link ikr.simlib.entities.phases.UnitPhase} (unit refers
 * here to information unit) is proportional to the length specified in
 * {@link ikr.simlib.messages.Message} and therefore models a link with a fixed
 * bandwidth. The service time of {@link ikr.simlib.entities.phases.StdPhase} is
 * obtained independently of the length of a packet by a distribution.<br>
 * In the following, the detailed characteristics are described:<br>
 * <br>
 * - The time intervals of the events, that represent the end of service are
 * determined by the derived classes.<br>
 * <br>
 * - State of the phase (free or busy) decides if arriving messages at the input
 * port can be accepted or not (class {@link ikr.simlib.ports.input.InputPort}).<br>
 * <br>
 * - Upon end of service the serviced message is sent to the succeeding service
 * unit via the output port (class
 * {@link ikr.simlib.ports.output.SynchronousOutputPort}), whereby blocking must
 * be avoided.
 */
public abstract class Phase extends Entity implements ReflectionConstructable2<SimNode, Parameters> {

   protected final SynchronousOutputPort outputPort;
   protected final InputPort inputPort;

   protected final IntegralStatistic occupancy;

   private Message message;

   public Phase(SimNode ownNode) {
      super(ownNode);

      outputPort = new SynchronousOutputPort(this);
      inputPort = new InputPort(this) {
         @Override
         protected void handleMessageIndication(Message msg) {
            Phase.this.handleMessageIndication();
         }
      };
      occupancy = new StdIntegralStatistic(ownNode.createChildNode("Occupancy"));
      message = null;
   }

   public InputPort getInput() {
      return inputPort;
   }

   public OutputPort getOutput() {
      return outputPort;
   }

   /**
    * 
    * @param msg
    *           the message which is currently processed
    * @return the duration between now (start of message processing) and when
    *         the message should be forwarded
    */
   protected abstract Duration getMessageProcessingDuration(Message msg);

   /**
    * Transfers message to output
    * 
    * Override for any message dependent processing
    */
   protected void forwardProcessedMessage() {
      outputPort.sendMessage(message);
   }

   private void startProcessingNextMessage() {
      message = inputPort.getMessage();
      final PointInTime now = Calendar.getInstance().getSystemTime();
      Calendar.getInstance().postEvent(new CalendarCallback() {
         @Override
         public void execute() {
            handleServiceTimeOverEvent();
         }
      }, null, now.plus(getMessageProcessingDuration(message)));
   }

   private void handleMessageIndication() {
      if (message == null) {
         startProcessingNextMessage();
         occupancy.update(1);
      }
   }

   private void handleServiceTimeOverEvent() {
      forwardProcessedMessage();
      message = null;

      if (inputPort.isMessageAvailable()) {
         startProcessingNextMessage();
      } else {
         occupancy.update(0);
      }
   }
}
