/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.entities.phases;

import ikr.simlib.entities.Entity;
import ikr.simlib.events.CalendarCallback;
import ikr.simlib.events.EventToken;
import ikr.simlib.events.calendar.Calendar;
import ikr.simlib.events.time.Duration;
import ikr.simlib.events.time.PointInTime;
import ikr.simlib.messages.Message;
import ikr.simlib.model.SimNode;
import ikr.simlib.ports.input.InputPort;
import ikr.simlib.ports.output.SynchronousOutputPort;
import ikr.simlib.statistics.integral.IntegralStatistic;
import ikr.simlib.statistics.integral.StdIntegralStatistic;

import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

/**
 * Phase which models ideal processor sharing. The serviceTimePerUnit (the
 * reciprocal of the rate) is shared equally between all messages. As more
 * messages arrive in the system, the rate per message is reduced and the total
 * finish time increases. If the load is above 1 (more messages arrive than can
 * be processed), the {@link ProcessorSharingPhase} becomes unstable.
 * <p>
 * For analytical evaluation, see e.g. Yashkov (1986): Processor-sharing Queues:
 * Some Progress in Analysis
 */
public class ProcessorSharingPhase extends Entity {

   private final Map<Message, Double> remainingLength = new HashMap<Message, Double>();
   private final Duration serviceTimePerUnit;

   private final IntegralStatistic occupancyStat, activeMessagesStat;

   private final SynchronousOutputPort outputPort = new SynchronousOutputPort(this);
   private final InputPort inputPort = new InputPort(this) {
      @Override
      public void handleMessageIndication(Message msg) {
         ProcessorSharingPhase.this.handleMessageIndication();
      }
   };

   private EventToken currentEventToken;
   private PointInTime lastUpdateTime;

   /**
    * 
    */
   public ProcessorSharingPhase(SimNode ownNode, Duration serviceTimePerUnit) {
      super(ownNode);

      this.serviceTimePerUnit = serviceTimePerUnit;
      this.occupancyStat = new StdIntegralStatistic(ownNode.createChildNode("Occupancy"));
      this.activeMessagesStat = new StdIntegralStatistic(ownNode.createChildNode("ActiveMessages"));
      this.lastUpdateTime = Calendar.getInstance().getSystemTime();
   }

   public InputPort getInput() {
      return inputPort;
   }

   public SynchronousOutputPort getOutput() {
      return outputPort;
   }

   /**
    * called when one message is finished, see also {@link #rescheduleEvent()}
    */
   public void handleProcessEvent() {
      updateRemainingLength();
      final Message finishedMessage = getNextFinishedMessage();
      final double remaining = remainingLength.remove(finishedMessage);
      assert Math.abs(remaining) < 1e-6; // should be zero, but could be
                                         // affected by rounding errors
      outputPort.sendMessage(finishedMessage);
      rescheduleEvent();
      updateStats();
   }

   /**
    * called when a new message arrives at the input port
    */
   public void handleMessageIndication() {
      updateRemainingLength();
      final Message msg = inputPort.getMessage();
      final Double oldVal = remainingLength.put(msg, (double) msg.getLength());
      assert oldVal == null;
      rescheduleEvent();
      updateStats();
   }

   /**
    * calculates how many length units have been processed since the last call
    * of this method and reduces the remainingt lengthes stored in the map by
    * that amount.
    * <p>
    * should be called before changes are made to the map of messages
    */
   private void updateRemainingLength() {
      final PointInTime now = Calendar.getInstance().getSystemTime();
      final Duration timeDifference = lastUpdateTime.until(now);
      lastUpdateTime = now;

      if (remainingLength.size() > 0) {
         final Duration serviceTimePerUnitPerMessage = serviceTimePerUnit.multiplyBy(remainingLength.size());
         final double transmittedPerMessage = timeDifference.toSeconds() / serviceTimePerUnitPerMessage.toSeconds();
         for (Entry<Message, Double> e : remainingLength.entrySet())
            e.setValue(e.getValue() - transmittedPerMessage);
      }
   }

   /**
    * calculates the time at which the next message is finished and schedules
    * the event for that time.
    * <p>
    * should be called after changes have been made to the map of messages
    */
   private void rescheduleEvent() {
      assert lastUpdateTime.equals(Calendar.getInstance().getSystemTime());
      if (currentEventToken != null && currentEventToken.isPending())
         currentEventToken.cancel();
      if (remainingLength.size() > 0) {
         final Duration serviceTimePerUnitPerMsg = serviceTimePerUnit.multiplyBy(remainingLength.size());
         final Message nextMessage = getNextFinishedMessage();
         final PointInTime finishTime = Calendar.getInstance().getSystemTime().plus(serviceTimePerUnitPerMsg.multiplyBy(remainingLength.get(nextMessage)));
         currentEventToken = Calendar.getInstance().postEvent(new CalendarCallback() {
            @Override
            public void execute() {
               handleProcessEvent();
            }
         }, null, finishTime);
      }
   }

   /**
    * should be called after changes have been made to the map of messages
    */
   private void updateStats() {
      occupancyStat.update(remainingLength.size() > 0 ? 1 : 0);
      activeMessagesStat.update(remainingLength.size());
   }

   /**
    * finds the message with the smallest remaining length. Not done with a
    * comparator to avoid excessive map lookups.
    * 
    * @return message with the smallest remaining length
    */
   private Message getNextFinishedMessage() {
      assert remainingLength.size() > 0;
      Entry<Message, Double> smallest = null;
      for (Entry<Message, Double> e : remainingLength.entrySet())
         if (smallest == null || e.getValue() < smallest.getValue())
            smallest = e;
      assert smallest != null;
      return smallest.getKey();
   }
}