/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.entities.queues;

import ikr.simlib.algorithms.queuingDisciplines.BoundedFIFOQDisc;
import ikr.simlib.algorithms.queuingDisciplines.BoundedLIFOQDisc;
import ikr.simlib.algorithms.queuingDisciplines.QueuingDiscipline;
import ikr.simlib.algorithms.queuingDisciplines.StackableQueuingDiscipline;
import ikr.simlib.algorithms.queuingDisciplines.UnboundedFIFOQDisc;
import ikr.simlib.algorithms.queuingDisciplines.UnboundedLIFOQDisc;
import ikr.simlib.entities.Entity;
import ikr.simlib.messages.Message;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.parameters.reflection.ReflectionConstructable2;
import ikr.simlib.ports.Port;
import ikr.simlib.ports.input.InputPort;
import ikr.simlib.ports.output.OutputPort;

/**
 * {@link QueuingEntity} is the entity that integrates a queuing discipline
 * managing some buffer into the simulation. The {@link QueuingEntity} has ports
 * and only calls few public methods of QueuingDiscipline:
 * <ul>
 * <li><code>isEmpty()</code>
 * <li><code>dequeue()</code>
 * <li><code>handleMessage()</code>
 * <li>potentially <code>dropScheduledNext()</code> The handleMessage() may just
 * add the message to the {@link StackableQueuingDiscipline}'s data buffer or
 * result in packet discards, maybe of the provided packet or other packet(s)
 * already in the buffer.
 */
public class QueuingEntity extends Entity implements ReflectionConstructable2<SimNode, Parameters> {

   /** as required by {@link ReflectionConstructable} */
   public static QueuingEntity createInstance(SimNode ownNode, Parameters pars) {
      return new QueuingEntity(ownNode, pars);
   }

   public static QueuingEntity createBoundedFIFOQueue(SimNode ownNode, int size) {
      final QueuingEntity qe = new QueuingEntity(ownNode);
      qe.setQDisc(new BoundedFIFOQDisc(ownNode.createChildNode("BoundedFIFOQDisc"), size));
      return qe;
   }

   public static QueuingEntity createBoundedFIFOQueue(SimNode ownNode, Parameters pars) {
      final QueuingEntity qe = new QueuingEntity(ownNode);
      qe.setQDisc(new BoundedFIFOQDisc(ownNode.createChildNode("BoundedFIFOQDisc"), pars));
      return qe;
   }

   public static QueuingEntity createUnboundedFIFOQueue(SimNode ownNode) {
      final QueuingEntity qe = new QueuingEntity(ownNode);
      qe.setQDisc(new UnboundedFIFOQDisc(ownNode.createChildNode("UnboundedFIFOQDisc")));
      return qe;
   }

   public static QueuingEntity createUnboundedFIFOQueue(SimNode ownNode, Parameters pars) {
      final QueuingEntity qe = new QueuingEntity(ownNode);
      qe.setQDisc(new UnboundedFIFOQDisc(ownNode.createChildNode("UnboundedFIFOQDisc"), pars));
      return qe;
   }

   public static QueuingEntity createBoundedLIFOQueue(SimNode ownNode, int size) {
      final QueuingEntity qe = new QueuingEntity(ownNode);
      qe.setQDisc(new BoundedLIFOQDisc(ownNode.createChildNode("BoundedLIFOQDisc"), size));
      return qe;
   }

   public static QueuingEntity createBoundedLIFOQueue(SimNode ownNode, Parameters pars) {
      final QueuingEntity qe = new QueuingEntity(ownNode);
      qe.setQDisc(new BoundedLIFOQDisc(ownNode.createChildNode("BoundedLIFOQDisc"), pars));
      return qe;
   }

   public static QueuingEntity createUnboundedLIFOQueue(SimNode ownNode) {
      final QueuingEntity qe = new QueuingEntity(ownNode);
      qe.setQDisc(new UnboundedLIFOQDisc(ownNode.createChildNode("UnboundedLIFOQDisc")));
      return qe;
   }

   public static QueuingEntity createUnboundedLIFOQueue(SimNode ownNode, Parameters pars) {
      final QueuingEntity qe = new QueuingEntity(ownNode);
      qe.setQDisc(new UnboundedLIFOQDisc(ownNode.createChildNode("UnboundedLIFOQDisc"), pars));
      return qe;
   }

   protected final InputPort inputPort = new InputPort(this) {
      @Override
      protected void handleMessageIndication(Message msg) {
         QueuingEntity.this.handleMessageIndication(msg);
      }
   };
   protected final OutputPort outputPort = new OutputPort(this) {
      @Override
      protected Message handleGetMessage() {
         return QueuingEntity.this.handleGetMessage(this);
      }

      @Override
      protected boolean handleIsMessageAvailable() {
         return QueuingEntity.this.handleIsMessageAvailable(this);
      }
   };
   protected QueuingDiscipline qdisc;

   public QueuingEntity(SimNode ownNode, Parameters pars) {
      super(ownNode);
      qdisc = pars.get(ownNode, "QueuingDiscipline").asInstance(QueuingDiscipline.class, ownNode.createChildNode("QueuingDiscipline"), pars);
   }

   /**
    * When using this constructor, you MUST use setQDisc to set the
    * {@link QueuingDiscipline} before using this {@link QueuingEntity}. It is
    * expected, but not enforced, that the {@link QueuingEntity} instance is the
    * parent {@link SimNode} of the {@link QueuingDiscipline}.
    */
   public QueuingEntity(SimNode ownNode) {
      super(ownNode);
   }

   public void setQDisc(QueuingDiscipline queuingDisc) {
      this.qdisc = queuingDisc;
   }

   public InputPort getInput() {
      return inputPort;
   }

   public OutputPort getOutput() {
      return outputPort;
   }

   public QueuingDiscipline getQDisc() {
      return qdisc;
   }

   protected Message handleGetMessage(Port port) {
      return qdisc.dequeue().getMessage();
   }

   protected boolean handleIsMessageAvailable(Port port) {
      return !qdisc.isEmpty();
   }

   protected void handleMessageIndication(Message msg) {
      Message m = inputPort.getMessage();
      qdisc.handleMessage(m);
      if (qdisc.getCurrentNumberOfMessages() == 1) {
         // first message in Queue
         outputPort.messageIndication(m);
      }
   }
}
