/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.events.calendar;

import ikr.simlib.events.CalendarCallback;
import ikr.simlib.events.EventToken;
import ikr.simlib.events.time.PointInTime;
import ikr.simlib.exceptions.SimulationException;
import ikr.simlib.statistics.UnweightedSampleCollector;

/**
 * Calendars are special event handlers derived from the abstract base class
 * {@link Calendar}. The classes store references to callback methods (
 * {@link CalendarCallback}) and associated points in time ({@link PointInTime})
 * when the callbacks are to be executed.
 * <p>
 * In addition to the event time, a priority of the event can be provided. These
 * priorities come into effect if events are scheduled to be executed at the
 * same time. In that case, they are processed in the order of decreasing
 * priority, i.e., the event with the highest priority is processed first.
 * <p>
 * When posting an event, the calendar returns an instance of {@link EventToken}
 * , which can be used to query the state of the event and to cancel the event.
 * <p>
 * With the method {@link #processNextEvent()} the events can be executed in the
 * scheduled order. Calendars are used by models, see
 * {@link ikr.simlib.model.Model } for more details.
 * 
 * <p>
 * Event priorities are integer numbers and may be negative. The default
 * priority is 0. The processing order of events with the same time <em>and</em>
 * the same priority is <em>not</em> defined and may vary in-between calendar
 * implementations.
 * </p>
 */
public abstract class Calendar {

   /**
    * reference to calendar
    */
   protected static Calendar CALENDAR = null;

   /**
    * calendar access method
    */
   public static Calendar getInstance() {
      assert (CALENDAR != null);
      return CALENDAR;
   }

   public static void setInstance(Calendar calendar) {
      CALENDAR = calendar;
   }

   public static boolean hasInstance() {
      return CALENDAR != null;
   }

   /**
    * the current simulation time
    */
   protected PointInTime systemTime = PointInTime.ZERO;

   protected UnweightedSampleCollector<Double> eventsWithSameTimeStat;

   /**
    * counter to count the number of events with exact same time stamp
    */
   protected int concurrentEvents = 0;

   protected Calendar() {
   }

   /**
    * Creates a new entry in the calendar.
    * 
    * @param onExecute
    *           Function to call when the event is processed. May be null.
    * @param onCancel
    *           Function to call when the event is cancelled. May be null.
    * @param eventTime
    *           Time at which to call the <code>onExecute</code> function. Must
    *           be later or equal than the current system time returned by
    *           {@link #getSystemTime()}.
    * @param eventPriority
    *           Key used for sorting to break ties for two entries with the same
    *           <code>eventTime</code>.
    * @return An object which allows event-specific interaction with the
    *         calendar.
    */
   public abstract EventToken postEvent(CalendarCallback onExecute, CalendarCallback onCancel, PointInTime eventTime, int eventPriority);

   /**
    * As
    * {@link #postEvent(CalendarCallback, CalendarCallback, PointInTime, int)},
    * but sets the <code>eventPriority</code> to 0.
    */
   public EventToken postEvent(CalendarCallback onExecute, CalendarCallback onCancel, PointInTime eventTime) {
      return postEvent(onExecute, onCancel, eventTime, 0);
   }

   /**
    * Cancel all events which are currently posted.
    */
   public abstract void cancelAllEvents();

   /**
    * Return true if there are events in the calendar.
    */
   public abstract boolean isEventAvailable();

   /**
    * Returns the time stamp of the next event in the calendar without removing
    * it.
    * 
    * @return preview onto the time stamp of the next event in the calendar
    */
   public abstract PointInTime peekNextEventTime();

   /**
    * Returns the current simulation time
    * 
    * @return the simulation time
    */
   public PointInTime getSystemTime() {
      return systemTime;
   }

   /**
    * Gets the next event on the schedule and calls its onExecute method.
    */
   public abstract void processNextEvent();

   /**
    * If set, the number of events with the exact same time stamp are collected
    * in the given {@link UnweightedSampleCollector}. The number of events with
    * same time stamp
    */
   public void setSameTimeEventStatistic(UnweightedSampleCollector<Double> collector) {
      if (this.eventsWithSameTimeStat != null) {
         throw new SimulationException("Collector already set");
      }
      this.eventsWithSameTimeStat = collector;
   }

}
