/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.events.calendar;

import ikr.simlib.events.CalendarCallback;
import ikr.simlib.events.EventToken;
import ikr.simlib.events.time.PointInTime;
import ikr.simlib.exceptions.SimulationException;

/**
 * The simplest and slowest implementation of a calendar. It consists solely of
 * a sorted linked list of event entries, the current system time and the
 * current calendar size.
 */
public class StdOrderedCalendar extends Calendar {

   /**
    * Current number of events.
    */
   private transient int size;

   /**
    * Head of the event list.
    */
   private Entry head;

   public StdOrderedCalendar() {
      super();
   }

   private void remove(Entry f, Entry e) {
      if (e == head) {
         head = e.next;
      } else {
         f.next = e.next;
      }

      size--;
   }

   private void removeFromTimeline(Entry entry) {
      if (head == null)
         throw new SimulationException("no entries in timeline");

      int removed = 0;
      for (Entry e = head, f = null; e != null; f = e, e = e.next) {
         if (e == entry) {
            remove(f, e);
            removed++;
         }
      }

      if (removed == 0)
         throw new SimulationException("entry not found in timeline");
      if (removed > 1)
         throw new SimulationException("entry found multiple times in timeline");
   }

   @Override
   public void cancelAllEvents() {
      while (head != null) {
         head.callCancel();
         remove(null, head);
      }
   }

   public Entry popNextEvent() {
      if (head == null)
         throw new SimulationException("there is no event in the calendar");

      assert head.eventTime.isAfterOrEqual(systemTime);
      // update number of events with same time, in case we advance in time
      // value is updated in the statistic
      if (eventsWithSameTimeStat != null && head.eventTime.isAfter(systemTime)) {
         eventsWithSameTimeStat.update((double) concurrentEvents);
         concurrentEvents = 1;
      } else if (head.eventTime.equals(systemTime)) {
         concurrentEvents++;
      }
      systemTime = head.eventTime;

      Entry oldHead = head;

      remove(null, head);

      return oldHead;
   }

   @Override
   public EventToken postEvent(CalendarCallback onExecute, CalendarCallback onCancel, PointInTime eventTime, int eventPriority) {
      if (systemTime.isAfter(eventTime)) {
         throw new SimulationException("Event Time < System Time");
      }

      final Entry entry = new Entry(eventTime, onExecute, onCancel, eventPriority);

      if (head != null) {
         for (Entry e = head, f = null; e != null; f = e, e = e.next) {
            if (e.eventTime.isAfter(eventTime) || (e.eventTime.equals(eventTime) && e.priority < eventPriority)) {

               // Insert before
               entry.next = e;

               // Are we at the head?
               if (e == head) {
                  assert f == null;
                  head = entry;
               } else {
                  assert f != null;
                  f.next = entry;
               }

               break;
            } else if (e.next == null) {
               // Last entry
               e.next = entry;

               break;
            }
         }
      } else {
         head = entry;
      }
      size++;
      return entry;
   }

   @Override
   public void processNextEvent() {
      final Entry entry = popNextEvent();
      entry.callProcess(systemTime);
   }

   /**
    * Get the current number of events in the calendar.
    * 
    * @return number of events in the calendar
    */
   public int getSize() {
      return size;
   }

   @Override
   public boolean isEventAvailable() {
      return head != null;
   }

   @Override
   public PointInTime peekNextEventTime() {
      if (head == null) {
         return PointInTime.ZERO;
      }

      return head.eventTime;
   }

   /**
    * Single linked calendar entry.
    */
   private class Entry implements EventToken {

      private final PointInTime eventTime;
      private final int priority;
      private final CalendarCallback process;
      private final CalendarCallback cancel;

      private boolean executed;
      private boolean cancelled;

      public Entry next;

      public Entry(PointInTime eventTime, CalendarCallback process, CalendarCallback cancel, int priority) {
         this.eventTime = eventTime;
         this.process = process;
         this.cancel = cancel;
         this.priority = priority;
         this.executed = false;
         this.cancelled = false;
      }

      void callProcess(PointInTime t) {
         executed = true;
         assert t.equals(eventTime);
         if (process != null)
            process.execute();
      }

      void callCancel() {
         cancelled = true;
         if (cancel != null)
            cancel.execute();
      }

      @Override
      public void cancel() {
         callCancel();
         removeFromTimeline(this);
      }

      @Override
      public boolean isPending() {
         return !(hasBeenExecuted() || hasBeenCancelled());
      }

      @Override
      public boolean hasBeenExecuted() {
         return executed;
      }

      @Override
      public boolean hasBeenCancelled() {
         return cancelled;
      }
   }

}
