/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.events.time;

import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.result.ResultReader;
import ikr.simlib.result.exceptions.ImportException;

/**
 * Represents an absolute point in time.
 * <p>
 * Can be created by adding a {@link Duration} to another {@link PointInTime}.
 * For the generation of new {@link PointInTime} objects, double values are
 * parsed to the internal representation using long values. The method
 * {@link Math#round(double)} is used to get the best fitting long value.
 * 
 */
public class PointInTime extends TimeBase implements Comparable<PointInTime> {
   private static final long serialVersionUID = 3334547011423341628L;

   public static PointInTime createInstance(SimNode ownNode, Parameters pars) {
      TimeUnits unitToParse = null;
      for (TimeUnits unit : TimeUnits.values()) {
         if (pars.hasParameter(ownNode, unit.getName()) || pars.hasParameter(ownNode, unit.getShortName())) {
            if (unitToParse != null) {
               throw new RuntimeException("multiple unit values defined for " + ownNode.getFullName());
            }
            unitToParse = unit;
         }
      }
      if (unitToParse == null) {
         throw new RuntimeException("no valid unit value defined for " + ownNode.getFullName());
      }

      final double value = pars.get(ownNode, unitToParse.getName()).asDouble();
      return from(value, unitToParse);
   }

   public static PointInTime ZERO = new PointInTime(0);
   public static PointInTime MAX = new PointInTime(Long.MAX_VALUE);

   public static PointInTime fromDays(double days) {
      return from(days, TimeUnits.d);
   }

   public static PointInTime fromHours(double hours) {
      return from(hours, TimeUnits.h);
   }

   public static PointInTime fromMinutes(double minutes) {
      return from(minutes, TimeUnits.m);
   }

   /**
    * create a new time object from the given value (SI-unit)
    */
   public static PointInTime fromSeconds(double seconds) {
      return from(seconds, TimeUnits.s);
   }

   public static PointInTime fromMilliSeconds(double milliSeconds) {
      return from(milliSeconds, TimeUnits.ms);
   }

   public static PointInTime fromMicroSeconds(double microSeconds) {
      return from(microSeconds, TimeUnits.us);
   }

   public static PointInTime fromNanoSeconds(double nanoSeconds) {
      return from(nanoSeconds, TimeUnits.ns);
   }

   public static PointInTime fromPicoSeconds(double picoSeconds) {
      return from(picoSeconds, TimeUnits.ps);
   }

   public static PointInTime from(double value, TimeUnits unit) {
      return new PointInTime(TimeBase.secondsToInternal(value / unit.getTimeBase()));
   }

   public static PointInTime importBatch(ResultReader reader, String name) throws ImportException {
      reader.readStartNode(PointInTime.class.getSimpleName(), name);
      final String timeUnit = reader.readString(EXPORT_UNIT);
      final long value = reader.readLong(EXPORT_VALUE);
      reader.readEndNode();

      if (!getTimeUnit().equals(timeUnit))
         throw new ImportException("unknown", "unknown", "invalid time unit: expected " + getTimeUnit() + " but read " + timeUnit);

      return new PointInTime(reader.readLong(EXPORT_VALUE));
   }

   protected PointInTime(long value) {
      super(value);
   }

   /**
    * @return time span from this time to the other time, negative if other lies
    *         in the past
    */
   public Duration until(PointInTime other) {
      return new Duration(this, other);
   }

   public PointInTime plus(Duration duration) {
      return new PointInTime(this.getValue() + duration.getValue());
   }

   /**
    * {@inheritDoc}
    */
   @Override
   public int compareTo(PointInTime other) {
      return Long.compare(getValue(), other.getValue());
   }

   /**
    * Compare two Time objects
    * 
    * @param other
    *           the Time to compare with the current object
    * @return Whether the current object points to a later point in time than
    *         the parameter other.
    */
   public boolean isAfter(PointInTime other) {
      return this.getValue() > other.getValue();
   }

   public boolean isAfterOrEqual(PointInTime other) {
      return this.getValue() >= other.getValue();
   }

   public boolean isBefore(PointInTime other) {
      return this.getValue() < other.getValue();
   }

   public boolean isBeforeOrEqual(PointInTime other) {
      return this.getValue() <= other.getValue();
   }

   /**
    * Return the later of two points in time
    * 
    * @param other
    *           the other value
    * @return the later of the two points in time
    */
   public PointInTime selectLatest(PointInTime other) {
      final PointInTime ret;
      if (this.isAfter(other)) {
         ret = this;
      } else {
         ret = other;
      }
      return ret;
   }

   /**
    * Return the earlier of two points in time
    * 
    * @param other
    *           the other value
    * @return the earlier of the two points in time
    */
   public PointInTime selectEarliest(PointInTime other) {
      final PointInTime ret;
      if (this.isAfter(other)) {
         ret = other;
      } else {
         ret = this;
      }
      return ret;
   }
}
