/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.events.time;

import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable2;
import ikr.simlib.result.ResultKeyword;
import ikr.simlib.result.ResultWriter;
import ikr.simlib.result.exceptions.PrintException;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;

/**
 * Common base class for @{link PointInTime} and @{link Duration}. Both have a
 * single value in a time unit that is configurable system-wide. The difference
 * is that {@link PointInTime} refers to a single system-wide reference (
 * {@link PointInTime#ZERO}), while {@link Duration} is a difference of two
 * times.
 */
public abstract class TimeBase implements ReflectionConstructable2<SimNode, Parameters>, Serializable {
   private static final long serialVersionUID = -3966851620508126850L;
   protected static final String EXPORT_VALUE = ResultKeyword.asExport("value");
   protected static final String EXPORT_UNIT = ResultKeyword.asExport("unit");

   private static TimeUnits DEFAULT_TIME_UNIT = TimeUnits.ns;

   private static boolean initialized = false;

   /** string describing the time unit, used for {@link #toString()} */
   private static String timeUnit;

   /**
    * TimeBase.getValue() / timeBase = [seconds]
    * <p>
    * timeBase = second * timeUnit
    */
   private static double timeBase;

   /**
    * TimeBase.getValue() * invTimeBase = [seconds]
    * <p>
    * invTimeBase = second / timeUnit
    */
   private static double invTimeBase;

   /**
    * converts internal units to seconds
    * 
    * @param internal
    *           time in internal units
    * @return time in seconds
    */
   public static double internalToSeconds(long internal) {
      assertInitialized();
      return internal * invTimeBase;
   }

   /**
    * converts seconds to internal units (incl. rounding)
    * 
    * @param seconds
    *           time in seconds
    * @return time in internal units
    */
   public static long secondsToInternal(double seconds) {
      assertInitialized();
      return Math.round(seconds * timeBase);
   }

   /**
    * This method allows to change the time-base before the first time object
    * has been created.
    * 
    * @param unit
    *           the new time unit
    */
   public static void changeTimeBase(TimeUnits unit) {
      changeTimeBase(unit.getTimeBase(), unit.getShortName());
   }

   /**
    * This method allows to change the time-base before the first time object
    * has been created.
    * 
    * @param newTimeBase
    *           number of time units per second
    * @param newTimeUnit
    *           string describing the new time unit
    */
   public static void changeTimeBase(double newTimeBase, String newTimeUnit) {
      if (initialized) {
         throw new RuntimeException("Time-base must not be changed after time objects have been created.");
      }
      initialized = true;
      timeBase = newTimeBase;
      timeUnit = newTimeUnit;
      invTimeBase = 1.0 / newTimeBase;
   }

   private static void assertInitialized() {
      if (!initialized) {
         changeTimeBase(DEFAULT_TIME_UNIT);
      }
   }

   /**
    * represents the actual time in the smallest model-wide time-units
    */
   final private long timeValue;

   protected TimeBase(long value) {
      assertInitialized();
      timeValue = value;
   }

   protected long getValue() {
      return timeValue;
   }

   public static String getTimeUnit() {
      return timeUnit;
   }

   @Override
   public String toString() {
      return getValue() + timeUnit;
   }

   public String toString(double outputTimeBase, String outputTimeUnit) {
      return (toSeconds() * outputTimeBase) + outputTimeUnit;
   }

   public String toString(TimeUnits unit, int decimalPlaces) {
      return String.format("%." + decimalPlaces + "f", toSeconds() * unit.getTimeBase());
   }

   /**
    * @return double time value in days
    */
   public double toDays() {
      return toSeconds() * TimeUnits.d.getTimeBase();
   }

   /**
    * @return double time value in hours
    */
   public double toHours() {
      return toSeconds() * TimeUnits.h.getTimeBase();
   }

   /**
    * @return double time value in minutes
    */
   public double toMinutes() {
      return toSeconds() * TimeUnits.m.getTimeBase();
   }

   /**
    * @return double time value in seconds
    */
   public double toSeconds() {
      return internalToSeconds(getValue());
   }

   /**
    * @return double time value in milliseconds
    */
   public double toMilliSeconds() {
      return toSeconds() * TimeUnits.ms.getTimeBase();
   }

   /**
    * @return double time value in microseconds
    */
   public double toMicroSeconds() {
      return toSeconds() * TimeUnits.us.getTimeBase();
   }

   /**
    * @return double time value in nanoseconds
    */
   public double toNanoSeconds() {
      return toSeconds() * TimeUnits.ns.getTimeBase();
   }

   /**
    * @return double time value in picoseconds
    */
   public double toPicoSeconds() {
      return toSeconds() * TimeUnits.ps.getTimeBase();
   }

   @Override
   public int hashCode() {
      return Long.valueOf(timeValue).hashCode();
   }

   @Override
   public boolean equals(Object obj) {
      if (this == obj)
         return true;
      if (obj == null)
         return false;
      if (obj.getClass() != this.getClass())
         return false;
      TimeBase other = (TimeBase) obj;
      if (timeValue != other.timeValue)
         return false;
      return true;
   }

   public void exportBatch(ResultWriter writer, String name) throws PrintException {
      writer.writeStartNode(getClass().getSimpleName(), name);
      writer.writeString(EXPORT_UNIT, timeUnit);
      writer.writeLong(EXPORT_VALUE, timeValue);
      writer.writeEndNode();
   }

   @SuppressWarnings("static-method")
   private void writeObject(ObjectOutputStream out) throws IOException {
      out.defaultWriteObject();
      out.writeDouble(timeBase);
      out.writeObject(timeUnit);
   }

   @SuppressWarnings("static-method")
   private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {
      in.defaultReadObject();
      final double serializedTimeBase = in.readDouble();
      final String serializedTimeUnit = (String) in.readObject();
      if (!initialized) {
         changeTimeBase(serializedTimeBase, serializedTimeUnit);
      } else if (serializedTimeBase != timeBase || !serializedTimeUnit.equals(timeUnit)) {
         throw new InvalidTimeBaseException("different time base and / or time unit in serialized object");
      }
   }

   public static class InvalidTimeBaseException extends RuntimeException {
      private static final long serialVersionUID = 1L;

      public InvalidTimeBaseException(String message) {
         super(message);
      }
   }

}
