/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.messages.factories;

import ikr.simlib.distributions.Distribution;
import ikr.simlib.distributions.discrete.DiscreteDistribution;
import ikr.simlib.messages.IPPacket;
import ikr.simlib.messages.Message;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.random.RandomNumberGenerator;

/**
 * FlexIPMessageFactory is more complex and less flexible regarding the messages
 * compared to its base class CloneMessageFactory.
 * 
 * <ul>
 * <li>Just messages of type {@link IPPacket IPPacket} are supported.</li>
 * <li>Four distributions are required. These distributions are adapting the
 * header entries <code>SourceId</code>, <code>DestinationId</code>,
 * <code>GroupId</code> and <code>TypeId</code>.</li>
 * <li>Parsing is also supported, the keywords are: <code>IPPacket</code>,
 * <code>SourceIdDist</code>, <code>DestinationIdDist</code>,
 * <code>GroupIdDist</code>, <code>TypeIdDist</code> and <code>Tag</code>.</li>
 * </ul>
 * 
 * @see UniqueIdIPMessageFactory
 */
public class FlexIPMessageFactory extends IPPacketFactory {

   private static final String LENGTH = "Length";
   private static final String SOURCE_ID = "SourceId";
   private static final String DESTINATION_ID = "DestinationId";
   private static final String GROUP_ID = "GroupId";
   private static final String TYPE_ID = "TypeId";
   private static final String TAG = "Tag";
   private static final String SOURCE_ID_DIST = "SourceIdDist";
   private static final String DESTINATION_ID_DIST = "DestinationIdDist";
   private static final String GROUP_ID_DIST = "GroupIdDist";
   private static final String TYPE_ID_DIST = "TypeIdDist";

   /** as required by {@link ReflectionConstructable} */
   public static FlexIPMessageFactory createInstance(SimNode ownNode, Parameters pars) {
      long length = Message.DEFAULT_MESSAGE_LENGTH;
      if (pars.hasParameter(ownNode, LENGTH)) {
         length = pars.get(ownNode, LENGTH).asLong();
      }
      final int sourceId = pars.get(ownNode, SOURCE_ID).asInteger();
      final int destinationId = pars.get(ownNode, DESTINATION_ID).asInteger();
      final int groupId = pars.get(ownNode, GROUP_ID).asInteger();
      final int typeId = pars.get(ownNode, TYPE_ID).asInteger();
      final double tag = pars.get(ownNode, TAG).asDouble();

      final RandomNumberGenerator rng = Distribution.getDefaultRNG();
      DiscreteDistribution sourceIdDist = null;
      if (pars.hasParameter(ownNode, SOURCE_ID_DIST))
         sourceIdDist = pars.get(ownNode, SOURCE_ID_DIST).asInstance(DiscreteDistribution.class, ownNode.createChildNode(SOURCE_ID_DIST), pars, rng);

      DiscreteDistribution destinationIdDist = null;
      if (pars.hasParameter(ownNode, DESTINATION_ID_DIST))
         destinationIdDist = pars.get(ownNode, DESTINATION_ID_DIST).asInstance(DiscreteDistribution.class, ownNode.createChildNode(DESTINATION_ID_DIST), pars, rng);

      DiscreteDistribution groupIdDist = null;
      if (pars.hasParameter(ownNode, GROUP_ID_DIST))
         groupIdDist = pars.get(ownNode, GROUP_ID_DIST).asInstance(DiscreteDistribution.class, ownNode.createChildNode(GROUP_ID_DIST), pars, rng);

      DiscreteDistribution typeIdDist = null;
      if (pars.hasParameter(ownNode, TYPE_ID_DIST))
         typeIdDist = pars.get(ownNode, TYPE_ID_DIST).asInstance(DiscreteDistribution.class, ownNode.createChildNode(TYPE_ID_DIST), pars, rng);

      return new FlexIPMessageFactory(sourceIdDist, destinationIdDist, groupIdDist, typeIdDist, length, sourceId, destinationId, groupId, typeId, tag);
   }

   public final DiscreteDistribution sourceIdDist;
   public final DiscreteDistribution destinationIdDist;
   public final DiscreteDistribution groupIdDist;
   public final DiscreteDistribution typeIdDist;

   public FlexIPMessageFactory(DiscreteDistribution sourceIdDist, DiscreteDistribution destIdDist, DiscreteDistribution groupIdDist, DiscreteDistribution typeIdDist, long length, int sourceId, int destinationId, int groupId, int typeId, double tag) {
      super(length, sourceId, destinationId, groupId, typeId, tag);

      this.sourceIdDist = sourceIdDist;
      this.destinationIdDist = destIdDist;
      this.groupIdDist = groupIdDist;
      this.typeIdDist = typeIdDist;
   }

   @Override
   public IPPacket create() {
      IPPacket temppackage = super.create();

      if (sourceIdDist != null)
         temppackage.setSourceId(sourceIdDist.next());

      if (destinationIdDist != null)
         temppackage.setDestinationId(destinationIdDist.next());

      if (groupIdDist != null)
         temppackage.setGroupId(groupIdDist.next());

      if (typeIdDist != null)
         temppackage.setTypeId(typeIdDist.next());

      return temppackage;
   }
}
