/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.messages.factories;

import ikr.simlib.messages.IPPacket;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;

/**
 * UniqueIdIPMessageFactory assigns each MessageFactory a unique value of the
 * message field stored in field. This is realized by a static member
 * FACTORY_ID.
 */
public class UniqueIdIPMessageFactory extends IPPacketFactory {

   public enum Field {
      SourceId, DestinationId, GroupId, TypeId
   }

   private static final String LENGTH = "Length";
   private static final String SOURCE_ID = "SourceId";
   private static final String DESTINATION_ID = "DestinationId";
   private static final String GROUP_ID = "GroupId";
   private static final String TYPE_ID = "TypeId";
   private static final String TAG = "Tag";
   private static final String ID_FIELD_NAME = "IdFieldName";

   /** as required by {@link ReflectionConstructable} */
   public static UniqueIdIPMessageFactory createInstance(SimNode ownNode, Parameters pars) {
      final long length = pars.get(ownNode, LENGTH).asLong();
      final int sourceId = pars.get(ownNode, SOURCE_ID).asInteger();
      final int destinationId = pars.get(ownNode, DESTINATION_ID).asInteger();
      final int groupId = pars.get(ownNode, GROUP_ID).asInteger();
      final int typeId = pars.get(ownNode, TYPE_ID).asInteger();
      final double tag = pars.get(ownNode, TAG).asDouble();

      final Field field = pars.get(ownNode, ID_FIELD_NAME).asEnum(Field.class);

      return new UniqueIdIPMessageFactory(length, sourceId, destinationId, groupId, typeId, tag, field);
   }

   protected static int FACTORY_ID = 0;

   protected int factoryID;
   protected Field field;

   public UniqueIdIPMessageFactory(long length, int sourceId, int destinationId, int groupId, int typeId, double tag, Field field) {
      super(length, sourceId, destinationId, groupId, typeId, tag);

      factoryID = FACTORY_ID++;
      this.field = field;
   }

   @Override
   public IPPacket create() {
      IPPacket tempPacket = super.create();
      assignId(tempPacket);
      return tempPacket;
   }

   protected void assignId(IPPacket packet) {
      switch (field) {
      case SourceId:
         packet.setSourceId(factoryID);
         break;
      case DestinationId:
         packet.setDestinationId(factoryID);
         break;
      case GroupId:
         packet.setGroupId(factoryID);
         break;
      case TypeId:
         packet.setTypeId(factoryID);
         break;
      default:
         throw new RuntimeException("unsupported field type " + field);
      }
   }
}
