/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.meters.cdv;

import ikr.simlib.events.calendar.Calendar;
import ikr.simlib.events.time.Duration;
import ikr.simlib.events.time.PointInTime;
import ikr.simlib.messages.Message;
import ikr.simlib.meters.OnePointMeter;
import ikr.simlib.ports.Port;
import ikr.simlib.statistics.UnweightedSampleCollector;

/**
 * An CDV meter can be attached to a port to measure one point CDV (cell delay
 * variation). The {@link #useMessage(Message)} method can be overridden to use
 * only those messages for measurement that satisfy certain criteria, e.g.,
 * based on message content, or message type.
 * <p>
 * Upon construction, {@link CDVMeter} requires a (fixed) rate which is taken to
 * calculate the cell delay variation (CDV) of the measured values.
 */
public class CDVMeter extends OnePointMeter {

   /**
    * creates a new {@link CDVMeter} with a statistic that accepts Duration
    * objects
    * 
    * @param statistic
    *           statistic to fill with the values
    * @param rate
    *           expected rate
    * @return the new {@link CDVMeter}
    */
   public static CDVMeter createWithDurationStatistic(UnweightedSampleCollector<Duration> statistic, double rate) {
      return new CDVMeter(statistic, rate);
   }

   /**
    * creates a new {@link CDVMeter} with a statistic that accepts Double
    * values. The measured durations are converted to seconds and then filled
    * into the statistic.
    * 
    * @param statistic
    *           statistic to fill with the values
    * @param rate
    *           expected rate
    * @return the new {@link CDVMeter}
    */
   public static CDVMeter createWithDoubleStatistic(final UnweightedSampleCollector<? super Double> statistic, double rate) {
      return new CDVMeter(new UnweightedSampleCollector<Duration>() {
         @Override
         public void update(Duration value) {
            statistic.update(value.toSeconds());
         }
      }, rate);
   }

   private final double rate;
   /** statistic to update with the durations */
   private final UnweightedSampleCollector<Duration> statistic;
   /** last reference time */
   private PointInTime referenceTime;

   protected CDVMeter(UnweightedSampleCollector<Duration> statistic, double rate) {
      this.statistic = statistic;
      this.rate = rate;
      referenceTime = PointInTime.ZERO;
   }

   /**
    * called when a message was seen
    * 
    * @param sample
    *           the current sample
    */
   protected void evaluateSample(Duration sample) {
      statistic.update(sample);
   }

   /**
    * filters messages that should be used for IAT measurement
    * 
    * @param msg
    *           the observed message
    */
   protected boolean useMessage(Message msg) {
      return true;
   }

   @Override
   protected void evaluateMessage(Message msg, Port port) {
      if (useMessage(msg)) {
         PointInTime currentTime = Calendar.getInstance().getSystemTime();
         evaluateSample(referenceTime.until(currentTime));
         referenceTime = referenceTime.selectLatest(currentTime).plus(Duration.fromSeconds(1.0 / rate));
      }
   }
}
