/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.meters.iat;

import ikr.simlib.events.calendar.Calendar;
import ikr.simlib.events.time.Duration;
import ikr.simlib.events.time.PointInTime;
import ikr.simlib.messages.Message;
import ikr.simlib.meters.OnePointMeter;
import ikr.simlib.ports.Port;
import ikr.simlib.statistics.UnweightedSampleCollector;

/**
 * An IAT meter can be attached to a port to measure message interarrival times.
 * The {@link #useMessage(Message)} method can be overridden to use only those
 * messages for measurement that satisfy certain criteria, e.g., based on
 * message content, or message type.
 */
public class IATMeter extends OnePointMeter {

   /**
    * creates a new {@link IATMeter} with a statistic that accepts Duration
    * objects
    * 
    * @param statistic
    *           statistic to fill with the values
    * @return the new {@link IATMeter}
    */
   public static IATMeter createWithDurationStatistic(UnweightedSampleCollector<Duration> statistic) {
      return new IATMeter(statistic);
   }

   /**
    * creates a new {@link IATMeter} with a statistic that accepts Double
    * values. The measured durations are converted to seconds and then filled
    * into the statistic.
    * 
    * @param statistic
    *           statistic to fill with the values
    * @return the new {@link IATMeter}
    */
   public static IATMeter createWithDoubleStatistic(final UnweightedSampleCollector<? super Double> statistic) {
      return new IATMeter(new UnweightedSampleCollector<Duration>() {
         @Override
         public void update(Duration value) {
            statistic.update(value.toSeconds());
         }
      });
   }

   /** statistic to update with the durations */
   private final UnweightedSampleCollector<Duration> statistic;
   /** arrival time of the previous message */
   private PointInTime lastTime;

   protected IATMeter(UnweightedSampleCollector<Duration> statistic) {
      this.statistic = statistic;
      lastTime = PointInTime.ZERO;
   }

   /**
    * called when a message was seen
    * 
    * @param prevMessageTime
    *           the time the previous message was seen
    * @param thisMessageTime
    *           the time this message was seen (now)
    */
   protected void evaluateSample(PointInTime prevMessageTime, PointInTime thisMessageTime) {
      statistic.update(prevMessageTime.until(thisMessageTime));
   }

   /**
    * filters messages that should be used for IAT measurement
    * 
    * @param msg
    *           the observed message
    */
   protected boolean useMessage(Message msg) {
      return true;
   }

   @Override
   protected void evaluateMessage(Message msg, Port port) {
      if (useMessage(msg)) {
         PointInTime currentTime = Calendar.getInstance().getSystemTime();
         evaluateSample(lastTime, currentTime);
         lastTime = currentTime;
      }
   }

}
