/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.model;

import ikr.simlib.events.calendar.Calendar;

/**
 * Model is the standard model for normal simulation. After ProcessPendingEvents
 * is called by the simulation control class (Simulation), it retrieves events
 * from the calendar and processes them. This goes on until no more events are
 * available, or StopProcessingEvents has been called.
 * 
 * The class Model has the following characteristics:<br>
<br>
- The function {@link #processPendingEvents()}, which is generally called by sequence control
upon begin of each batch, processes events in a loop by calling the calendar method
<code>processNextEvent()</code>.<br>
<br>
- The function {@link #stopProcessingEvents()}, which is generally called by sequence control
at the end of each batch, indirectly causes the termination of the loop in
{@link #processPendingEvents()} by setting the flag <code>done</code>, which is checked after each loop
run, to true.
 */
public class Model {

   protected boolean done;
   protected final SimNode simNode;

   public Model(String name) {
      simNode = StdSimNode.createRootNode(name);
      new SimNodeManager(simNode);
      done = false;
   }

   public void stopProcessingEvents() {
      // no more events
      done = true;
   }

   public void processPendingEvents() {
      // moved behind processNextEvent()-loop to allow transient phases with
      // zero events (i.e., empty transient phase)
      boolean lastEvent = false;
      while (!done && !lastEvent) {
         if (Calendar.getInstance().isEventAvailable())
            Calendar.getInstance().processNextEvent();
         else
            lastEvent = true;
      }
      done = false;
   }

   public SimNode getSimNode() {
      return simNode;
   }
}
