/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.parameters;

import ikr.simlib.control.simulationcontrol.callbacks.InitSimulationCallbackI;
import ikr.simlib.control.simulationcontrol.callbacks.SimulationControlManager;
import ikr.simlib.control.simulationcontrol.callbacks.StopSimulationCallbackI;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.keys.Key;
import ikr.simlib.parameters.keys.KeyQueryBase;
import ikr.simlib.parameters.keys.Query;
import ikr.simlib.parameters.values.Value;

import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.logging.Logger;

public class Parameters {

   private static final Logger LOG = Logger.getLogger(Parameters.class.getName());

   private final Map<Key, Value> database;
   private final Map<Key, Integer> queryCounter;
   private final Map<Query, Set<String>> defaultValues;
   private final PrintWriter parserLogStream;

   public Parameters(PrintWriter parserLogStream) {
      this.parserLogStream = parserLogStream;
      this.database = new HashMap<Key, Value>();
      this.queryCounter = new HashMap<Key, Integer>();
      this.defaultValues = new HashMap<Query, Set<String>>();

      /*
       * Pare log should be written twice to ease debugging: Static models can
       * be handled directly after the initialization of the simulation. Dynamic
       * models may change during the simulation run and must be handled at the
       * end of the simulation.
       */
      SimulationControlManager.getInstance().registerInitSimulationCallback(new InitSimulationCallbackI() {
         @Override
         public void handleInitSimulation() {
            logKeyCountersAndDefaults();
         }
      });

      SimulationControlManager.getInstance().registerStopSimulationCallback(new StopSimulationCallbackI() {
         @Override
         public void handleStopSimulation() {
            logKeyCountersAndDefaults();
         }
      });
   }

   public void add(String name, Value value) {
      if (database.containsKey(new Key(name)))
         throw new ParameterException("parameter " + name + " does exist already");

      database.put(new Key(name), value);
   }

   public boolean hasParameter(SimNode simNode) {
      return hasParameter(simNode.getFullName());
   }

   public boolean hasParameter(SimNode simNode, String name) {
      return hasParameter(simNode.getFullName() + NamingConstants.DELIMITER + name);
   }

   public boolean hasParameter(String name) {
      // create query from parameter name
      Query query = new Query(name);

      // get all matching keys and values
      Set<Entry<Key, Value>> matchingKeyValuePairs = getMatchingKeyValuePairs(query);

      // select the key and value with longest matching prefix
      Entry<Key, Value> longestPrefixMatch = getLongestPrefixMatch(matchingKeyValuePairs, query);

      return longestPrefixMatch != null;
   }

   public boolean hasSubTree(SimNode simNode) {
      return hasSubTree(simNode.getFullName());
   }

   public boolean hasSubTree(String name) {
      // create query from parameter name
      Query query = new Query(name);

      // get all entries of the database
      Set<Entry<Key, Value>> matches = new HashSet<Entry<Key, Value>>();

      // search database for partial matches
      for (Entry<Key, Value> entry : database.entrySet()) {
         if (entry.getKey().matchesPartially(query)) {
            matches.add(entry);
         }
      }

      return !matches.isEmpty();
   }

   public Value get(String name) {
      Value value = query(new Query(name));
      if (value == null)
         throw new ParameterException("parameter " + name + " doesn't exist");

      return value;
   }

   public Value get(SimNode simNode) {
      return get(simNode.getFullName());
   }

   public Value get(SimNode simNode, String name) {
      return get(simNode.getFullName() + NamingConstants.DELIMITER + name);
   }

   /**
    * caution - handle with care
    */
   public Value getOrUseDefault(String name, String defaultValue) {
      final Value ret;
      if (hasParameter(name)) {
         ret = get(name);
      } else {
         rememberDefaultUsage(new Query(name), defaultValue);
         LOG.config("query: " + new Query(name).toString() + "  ->  no matching key, using default value: " + defaultValue);
         ret = Value.fromString(defaultValue);
      }
      return ret;
   }

   public Value getOrUseDefault(SimNode simNode, String defaultValue) {
      return getOrUseDefault(simNode.getFullName(), defaultValue);
   }

   public Value getOrUseDefault(SimNode simNode, String name, String defaultValue) {
      return getOrUseDefault(simNode.getFullName() + NamingConstants.DELIMITER + name, defaultValue);
   }

   private Value query(Query query) {
      // get all matching keys and values
      Set<Entry<Key, Value>> matchingKeyValuePairs = getMatchingKeyValuePairs(query);

      // select the key and value with longest matching prefix
      Entry<Key, Value> longestPrefixMatch = getLongestPrefixMatch(matchingKeyValuePairs, query);

      if (longestPrefixMatch == null) {
         // catch case of no hit
         throw new ParameterException("query " + query.toString() + " returned nothing");
      }

      LOG.config("query: " + query.toString() + "  ->  matching key: " + longestPrefixMatch.getKey().toString() + "  matching value: " + longestPrefixMatch.getValue().toString());
      incrementQueryCount(longestPrefixMatch.getKey());
      return longestPrefixMatch.getValue();
   }

   private Set<Entry<Key, Value>> getMatchingKeyValuePairs(Query query) {
      Set<Entry<Key, Value>> matches = new HashSet<Entry<Key, Value>>();

      // search database for matches
      for (Entry<Key, Value> entry : database.entrySet()) {
         if (entry.getKey().matches(query)) {
            matches.add(entry);
         }
      }

      return matches;
   }

   private static Entry<Key, Value> getLongestPrefixMatch(Set<Entry<Key, Value>> matches, Query query) {
      int matchingPrefixLength = Integer.MIN_VALUE;
      Entry<Key, Value> bestMatch = null;
      boolean multipleFound = false;

      // pick the match with the longest matching prefix
      for (Entry<Key, Value> entry : matches) {
         int length = entry.getKey().getMatchingPrefixLength(query);
         if (length > matchingPrefixLength) {
            matchingPrefixLength = entry.getKey().getMatchingPrefixLength(query);
            bestMatch = entry;
            multipleFound = false;
         } else if (length == matchingPrefixLength) {
            multipleFound = true;
         }
      }

      if (multipleFound) {
         throw new ParameterException("multiple matches with the same prefix length found");
      }

      return bestMatch;
   }

   private void incrementQueryCount(Key key) {
      if (!queryCounter.containsKey(key))
         queryCounter.put(key, 0);
      queryCounter.put(key, queryCounter.get(key) + 1);
   }

   private void rememberDefaultUsage(Query query, String defVal) {
      if (!defaultValues.containsKey(query))
         defaultValues.put(query, new HashSet<String>());
      defaultValues.get(query).add(defVal);
   }

   private void logKeyCountersAndDefaults() {
      if (parserLogStream != null) {
         parserLogStream.println("query counters");
         final ArrayList<Key> keys = new ArrayList<Key>(database.keySet());
         Collections.sort(keys, new KeyQueryBaseComp());
         for (Key key : keys) {
            final int counter = queryCounter.containsKey(key) ? queryCounter.get(key) : 0;
            parserLogStream.println("\t" + key.toString() + "\t" + counter);
         }
         parserLogStream.println("default values");
         final ArrayList<Query> queries = new ArrayList<Query>(defaultValues.keySet());
         Collections.sort(queries, new KeyQueryBaseComp());
         for (Query query : queries) {
            parserLogStream.println("\t" + query.toString() + "\t" + defaultValues.get(query).toString());
         }
         parserLogStream.close();
      }
   }

   private class KeyQueryBaseComp implements Comparator<KeyQueryBase> {
      @Override
      public int compare(KeyQueryBase o1, KeyQueryBase o2) {
         return o1.toString().compareTo(o2.toString());
      }
   }
}
