/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.result;

import ikr.simlib.result.exceptions.PrintException;

/**
 * This interface allows serial output of result trees. This is the most basic
 * way to write results and no formatting or keyword handling is applied.
 * 
 * <p>
 * Comments can be written anywhere to the result tree but might not be
 * supported by implementations and are always skipped by result readers. They
 * can however increase readability in certain cases.
 * </p>
 * 
 * <p>
 * Result names are optional.
 * </p>
 * 
 */
public interface ResultWriter {

   /**
    * Closes the writer and underlying streams.
    */
   void close() throws PrintException;

   /**
    * Writes the beginning of the results. In order to complete the results the
    * user <strong>must</strong> call {@link #writeEndResults() writeEndResults}
    * .
    * 
    * @throws PrintException
    *            if any error in writing the results occurs
    */
   void writeStartResults() throws PrintException;

   /**
    * Writes the end of the results. It <strong>must</strong> be preceded by a
    * call to {@link #writeStartResults() writeStartResults} and all other
    * <code>writeStartXXX</code> must have been matched by their
    * <code>writeEndXXX</code> counterparts.
    * 
    * @throws PrintException
    *            if any error in writing the results occurs
    */
   void writeEndResults() throws PrintException;

   /**
    * Writes the beginning of a result node. In order to complete the node the
    * user <strong>must</strong> call {@link #writeEndNode() writeEndNode}.
    * 
    * @param type
    *           of the result node
    * 
    * @throws PrintException
    *            if any error in writing the results occurs
    */
   void writeStartNode(String type, String name) throws PrintException;

   /**
    * Writes the end of a result node. It <strong>must</strong> be preceded by a
    * call to {@link #writeStartNode(String, String) writeStartNode}.
    * 
    * @throws PrintException
    *            if any error in writing the results occurs
    */
   void writeEndNode() throws PrintException;

   /**
    * Writes the beginning of a result array. In order to complete the array you
    * <strong>must</strong> call {@link #writeEndArray() writeEndArray}.
    * 
    * @throws PrintException
    *            if any error in writing the results occurs
    */
   void writeStartArray(String name) throws PrintException;

   /**
    * Writes the end of a result array. It <strong>must</strong> be preceded by
    * a call to {@link #writeStartArray(String) writeStartArray}.
    * 
    * @throws PrintException
    *            if any error in writing the results occurs
    */
   void writeEndArray() throws PrintException;

   /**
    * Writes the beginning of a result array bucket. In order to complete the
    * bucket you <strong>must</strong> call {@link #writeEndArrayItem()
    * writeEndBucket}. It is only legal to call this method in an array.
    * 
    * @throws PrintException
    *            if any error in writing the results occurs
    */
   void writeStartArrayItem(long index) throws PrintException;

   /**
    * Writes the end of a result array bucket. It <strong>must</strong> be
    * preceded by a call to {@link #writeStartArrayItem(long) writeStartBucket}.
    * 
    * @throws PrintException
    *            if any error in writing the results occurs
    */
   void writeEndArrayItem() throws PrintException;

   /**
    * A shorthand method for
    * 
    * <pre>
    * &lt;code&gt;
    * writeLong(value);
    * writeName(name);
    * &lt;/code&gt;
    * </pre>
    * 
    * @param name
    *           name of the result
    * @param value
    *           value of the result
    * @throws PrintException
    *            if any error in writing the results occurs
    */
   void writeLong(String name, long value) throws PrintException;

   /**
    * A shorthand method for
    * 
    * <pre>
    * &lt;code&gt;
    * writeDouble(value);
    * writeName(name);
    * &lt;/code&gt;
    * </pre>
    * 
    * @param name
    *           name of the result
    * @param value
    *           value of the result
    * @throws PrintException
    *            if any error in writing the results occurs
    */
   void writeDouble(String name, double value) throws PrintException;

   /**
    * A shorthand method for
    * 
    * <pre>
    * &lt;code&gt;
    * writeString(value);
    * writeName(name);
    * &lt;/code&gt;
    * </pre>
    * 
    * @param name
    *           name of the result
    * @param value
    *           value of the result
    * @throws PrintException
    *            if any error in writing the results occurs
    */
   void writeString(String name, String value) throws PrintException;

}
