/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.statistics;

import ikr.simlib.result.ResultKeyword;
import ikr.simlib.result.ResultKeywords;
import ikr.simlib.result.ResultReader;
import ikr.simlib.result.ResultWriter;
import ikr.simlib.result.exceptions.ImportException;
import ikr.simlib.result.exceptions.PrintException;

import java.io.Serializable;

/**
 * Range of samples.
 * 
 */
public final class Range implements Serializable, Cloneable {

   private static final long serialVersionUID = 6137432605414075839L;

   // -- Results --
   private static final String RANGE_TYPE = "Range";

   private static final String EXPORT_MIN = ResultKeyword.asExport(ResultKeywords.MIN.toString());
   private static final String EXPORT_MAX = ResultKeyword.asExport(ResultKeywords.MAX.toString());

   /**
    * Upper bound.
    */
   private double maximum;

   /**
    * Lower bound.
    */
   private double minimum;

   /**
    * Whether this range has been updated since last reset.
    */
   private boolean isDefined;

   /**
    * Constructs a new range with undefined bounds
    */
   public Range() {
      reset();
   }

   /**
    * Constructs a new range with the given bounds.
    */
   public Range(double min, double max) {
      if (min > max) {
         throw new IllegalArgumentException();
      }

      isDefined = true;

      minimum = min;
      maximum = max;
   }

   @Override
   public boolean equals(Object obj) {
      if (obj == null) {
         return false;
      }

      if (obj.getClass() != getClass()) {
         return false;
      }

      Range range = (Range) obj;

      return Double.valueOf(maximum).equals(range.maximum) && Double.valueOf(minimum).equals(range.minimum);
   }

   @Override
   public int hashCode() {
      return Double.valueOf(maximum).hashCode() + Double.valueOf(minimum).hashCode();
   }

   @Override
   public String toString() {
      if (!isDefined) {
         return "[?, ?]";
      }

      StringBuilder str = new StringBuilder();

      str.append('[');
      str.append(getMinimum());
      str.append(", ");
      str.append(getMaximum());
      str.append(']');

      return str.toString();
   }

   public void exportState(String name, ResultWriter writer) throws PrintException {
      writer.writeStartNode(RANGE_TYPE, name);

      writer.writeDouble(EXPORT_MIN, minimum);
      writer.writeDouble(EXPORT_MAX, maximum);

      writer.writeEndNode();
   }

   public void importState(String name, ResultReader reader) throws ImportException {
      reader.readStartNode(RANGE_TYPE, name);

      minimum = reader.readDouble(EXPORT_MIN);
      maximum = reader.readDouble(EXPORT_MAX);

      reader.readEndNode();
   }

   public void joinStateWithImportState(String name, ResultReader reader) throws ImportException {
      reader.readStartNode(RANGE_TYPE, name);

      update(reader.readDouble(EXPORT_MIN));
      update(reader.readDouble(EXPORT_MAX));

      reader.readEndNode();
   }

   /**
    * Updates the bounds of this range.
    */
   public void update(double value) {
      if (!isDefined) {
         isDefined = true;

         minimum = value;
         maximum = value;

      } else {
         if (Double.isNaN(value)) {
            minimum = Double.NaN;
            maximum = Double.NaN;

         } else {
            if (value > maximum) {
               maximum = value;
            }

            if (value < minimum) {
               minimum = value;
            }
         }
      }
   }

   /**
    * Resets this range to undefined bounds.
    */
   public void reset() {
      isDefined = false;

      minimum = Double.NaN;
      maximum = Double.NaN;
   }

   /**
    * Returns the upper bound of this range.
    * 
    * @return the upper bound or NaN
    */
   public double getMaximum() {
      return maximum;
   }

   /**
    * Returns the lower bound of this range.
    * 
    * @return the lower bound or NaN
    */
   public double getMinimum() {
      return minimum;
   }

   /**
    * Returns whether this is a defined range.
    */
   public boolean isDefined() {
      return isDefined;
   }
}
