/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.statistics;

import ikr.simlib.control.simulationcontrol.callbacks.InitSimulationCallbackI;
import ikr.simlib.control.simulationcontrol.callbacks.SimulationControlManager;
import ikr.simlib.control.simulationcontrol.callbacks.StartBatchCallbackI;
import ikr.simlib.control.simulationcontrol.callbacks.StartTransientPhaseCallbackI;
import ikr.simlib.control.simulationcontrol.callbacks.StopBatchCallbackI;
import ikr.simlib.control.simulationcontrol.callbacks.StopTransientPhaseCallbackI;
import ikr.simlib.model.SimNode;
import ikr.simlib.result.PrintResultWriter;
import ikr.simlib.result.ResultKeyword;
import ikr.simlib.result.ResultTaggable;
import ikr.simlib.result.ResultWriter;
import ikr.simlib.result.exceptions.PrintException;

import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Map.Entry;

/**
 * {@link ikr.simlib.statistics.Statistic} is an abstract base class that
 * controls measurement and printing. <br>
 * <br>
 * All statistics classes are derived from the abstract base class
 * {@link ikr.simlib.statistics.Statistic} that is characterized by the
 * following:
 * <ul>
 * <li>
 * reacts to the signals <i>init simulation</i>, <i>start simulation</i>,
 * <i>stop simulation</i>, <i>start transient phase</i>, and <i>stop transient
 * phase</i>.
 * <li>
 * defines an interface of purely abstract functions
 * <ul>
 * <li>
 * {@link #resetStatistic()} causes a complete reset of the statistics at the
 * beginning of a simulation as well as after a warm-up phase
 * <li>
 * {@link #resetBatchStatistic()} causes a reset of counters, etc. at the
 * beginning of a batch
 * <li>
 * {@link #computeMeasures(int)} calculates a partial spot test from the
 * measured values of a batch
 * </ul>
 * <li>
 * has a {@link ikr.simlib.model.SimNode} to support the exporting, importing,
 * and printing of results.
 * </ul>
 * Generally, measured values are acquired with the help of statistics during a
 * batch (e.g. with the function <code>update()</code> which often exists in
 * derived classes). These results are used to calculate an intermediate result
 * with {@link #computeMeasures(int)} at the end of a batch depending on the
 * statistics type. At the end of all batches the individual intermediate
 * results are dealt with as samples. Normally, a mean value is composed from
 * these samples and the confidence interval is determined. The confidence
 * interval is an indication of the statistical significance of the determined
 * value. It is defined as an interval, in which the actual value lies with the
 * probability \(q\) (default value \(q = 0.95\)). In order to determine the
 * confidence interval the student-t distribution is applied *. <br>
 * <br>
 * Depending on the type of measured values, different procedures for
 * registration and evaluation are available. Several classes have been derived
 * from the base class {@link ikr.simlib.statistics.Statistic}, which define
 * these measurement procedures. There is one class for each of these special
 * statistics, that defines the interface for evaluation (e.g.,
 * {@link ikr.simlib.statistics.sample.SampleStatistic}) and another that
 * implements the evaluation functions (e.g.,
 * {@link ikr.simlib.statistics.sample.StdSampleStatistic}). <br>
 * The following auxiliary classes are available for the individual statistics:
 * <ul>
 * <li><code>Summation</code> for the registration and evaluation of measured
 * values or samples from the batches e.g., mean value, variance, standard
 * deviation and confidence interval of the mean values.
 * <li><code>Range</code> to determine the smallest and largest mean value
 * during a batch or the complete simulation.
 * <li><code>EstimationManager</code> , <code>StatisticEstimation</code> ,
 * <code>Student</code> , <code>StudentSearch</code> , <code>StudentCalc</code>
 * , and <code>StudentMixed</code> to determine the confidence interval (calls
 * occur from <code>Summation</code> automatically)
 * </ul>
 * Note: In earlier versions, a global statistics manager (class
 * <code>StatisticManager</code> ), which contained the appropriate
 * <code>create<*>Statistic()</code> and <code>delete<*>Statistic()</code>
 * methods, was used to create an object of a special statistics class. The
 * interface of the statistics manager was used to assure that the code of the
 * model components remained unchanged, should a statistics type (e.g.,
 * {@link ikr.simlib.statistics.sample.SampleStatistic}) be realized by a class
 * not belonging to the pre-defined standard statistics class (e.g.,
 * {@link ikr.simlib.statistics.sample.StdSampleStatistic}). <br>
 * 
 * @see "C. GRG: Verkehrstheoretische Modelle und stochastische
 *      Simulationstechniken zur Leistungsanalyse von Kommunikationsnetzen,
 *      Habilitation, RWTH Aachen, 1997."
 * @see "A. M. LAW, W. D. KELTON: Simulation Modeling & Analysis, 2nd edition,
 *      McGraw- Hill, 1991."
 */
public abstract class Statistic implements ResultTaggable {

   protected final SimNode simNode;
   private final Map<ResultKeyword, String> tags;

   public Statistic(SimNode ownNode) {
      // give statistic a name
      simNode = ownNode;
      tags = new LinkedHashMap<ResultKeyword, String>();

      SimulationControlManager.getInstance().registerInitSimulationCallback(new InitSimulationCallbackI() {
         @Override
         public void handleInitSimulation() {
            Statistic.this.handleInitSimulation();
         }
      });

      SimulationControlManager.getInstance().registerStartTransientPhaseCallback(new StartTransientPhaseCallbackI() {
         @Override
         public void handleStartTransientPhase() {
            Statistic.this.handleStartTransientPhase();
         }
      });

      SimulationControlManager.getInstance().registerStopTransientPhaseCallback(new StopTransientPhaseCallbackI() {
         @Override
         public void handleStopTransientPhase() {
            Statistic.this.handleStopTransientPhase();
         }
      });

      SimulationControlManager.getInstance().registerStartBatchCallback(new StartBatchCallbackI() {
         @Override
         public void handleStartBatch(int batchNumber) {
            Statistic.this.handleStartBatch(batchNumber);
         }
      });

      SimulationControlManager.getInstance().registerStopBatchCallback(new StopBatchCallbackI() {
         @Override
         public void handleStopBatch(int batchNumber) {
            Statistic.this.handleStopBatch(batchNumber);
         }
      });
   }

   /**
    * reset complete statistic
    */
   public abstract void resetStatistic();

   /**
    * only between batches
    */
   public abstract void resetBatchStatistic();

   public abstract void computeMeasures(int batchNumber);

   @Override
   public void addResultTag(ResultKeyword keyword, String value) {
      tags.put(keyword, value);
   }

   @Override
   public void addResultTag(String keyword, String value) {
      addResultTag(new ResultKeyword(keyword), value);
   }

   protected void printComments(PrintResultWriter writer) throws PrintException {
      for (Entry<ResultKeyword, String> tag : tags.entrySet()) {
         writer.writeString(tag.getKey(), tag.getValue());
      }
   }

   public void handleInitSimulation() {
      resetStatistic();
   }

   public void handleStartBatch(int batchNumber) {
      resetBatchStatistic();
   }

   public void handleStartTransientPhase() {
      resetStatistic();
   }

   public void handleStopBatch(int batchNumber) {
      computeMeasures(batchNumber);
   }

   public void handleStopTransientPhase() {
      resetStatistic();
   }

   /**
    * Writer object to write trace entries
    */
   protected ResultWriter traceWriter = null;

   protected int sampleIndex = 0;

   /**
    * Basic update method, that triggers the output of trace entries
    * 
    * @throws PrintException
    */
   protected void update() throws PrintException {
      if (traceWriter != null) {
         traceWriter.writeStartArrayItem(sampleIndex);
         sampleIndex++;
         writeTraceEntry();
         traceWriter.writeEndArrayItem();
      }
   }

   /**
    * Sets the writer object and enables tracing
    * 
    * @param writer
    *           <code>ResultWriter</code> connected to trace file
    * @throws PrintException
    */
   public void enableTracing(ResultWriter writer) throws PrintException {
      this.traceWriter = writer;
      traceWriter.writeStartResults();
      traceWriter.writeStartArray("trace");
   }

   /**
    * Un-sets the writer object and disables tracing
    * 
    * @throws PrintException
    */
   public void disableTracing() throws PrintException {
      traceWriter.writeEndArray();
      traceWriter.writeEndResults();
      traceWriter = null;
   }

   /**
    * Abstract method to be overwritten by derived statistics to output trace
    * entries
    */
   protected void writeTraceEntry() {
   }
}
