/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.statistics.aggregating;

import ikr.simlib.events.CalendarCallback;
import ikr.simlib.events.calendar.Calendar;
import ikr.simlib.events.time.Duration;
import ikr.simlib.statistics.UnweightedSampleCollector;

/**
 * {@link IntervalAggregatingStatistic} aggregates values over a configured
 * interval. Each time the interval is over, the aggregated value is written to
 * the backing statistic and the aggregator is reset. *
 * <p>
 * Implements the interface {@link UnweightedSampleCollector}, so that it can be used
 * as a replacement for a normal statistic in any meter.
 * <p>
 * The type of input and output can be different, as aggregation may convert the
 * type. Example: the average of long values can be returned as double.
 * 
 * @param <T>
 *           type of the input values
 * @param <V>
 *           type of the output values
 */
public class IntervalAggregatingStatistic<T, V> implements UnweightedSampleCollector<T> {
   private final UnweightedSampleCollector<V> baseStat;
   private final Duration aggregationInterval;
   private final Aggregator<? super T, ? extends V> aggregator;

   public IntervalAggregatingStatistic(UnweightedSampleCollector<V> baseStat, Duration aggregationInterval, Aggregator<? super T, ? extends V> aggregator) {
      this.baseStat = baseStat;
      this.aggregationInterval = aggregationInterval;
      this.aggregator = aggregator;

      Calendar.getInstance().postEvent(new CalendarCallback() {
         @Override
         public void execute() {
            putAggregatedValueToStatistic();
         }
      }, null, Calendar.getInstance().getSystemTime().plus(aggregationInterval));
   }

   @Override
   public void update(T value) {
      aggregator.putValue(value);
   }

   private void putAggregatedValueToStatistic() {
      final V aggregate = aggregator.getAndReset();
      baseStat.update(aggregate);
      Calendar.getInstance().postEvent(new CalendarCallback() {
         @Override
         public void execute() {
            putAggregatedValueToStatistic();
         }
      }, null, Calendar.getInstance().getSystemTime().plus(aggregationInterval));
   }
}
