/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.statistics.batchsum;

import ikr.simlib.model.SimNode;
import ikr.simlib.result.PrintResultWriter;
import ikr.simlib.result.Printable;
import ikr.simlib.result.ResultKeywords;
import ikr.simlib.result.ResultType;
import ikr.simlib.result.exceptions.PrintException;
import ikr.simlib.statistics.Statistic;
import ikr.simlib.statistics.UnweightedSampleCollector;

/**
 * The classes <code>BatchSumStatistic</code> or
 * {@link ikr.simlib.statistics.batchsum.StdBatchSumStatistic} add up measured
 * values (e.g., number of bytes that have passed a certain port) of the type
 * <code>long</code> during a batch with the help of the method
 * {@link #update()}.
 * <p>
 * The following functions are available for evaluation:
 * <ul>
 * <li>Mean value of batch sums ({@link #getMean()}) including a confidence
 * interval ({@link #getConfidenceInterval()})</li>
 * <li>Value range ({@link #getMinimum()} and {@link #getMaximum()}) of the
 * batch sums</li>
 * <li>Number of calls to the update() method over all batches (
 * {@link #getNumberOfEvents()})</li>
 * <li>Number of batches ({@link #getNumberOfBatches()})</li>
 * </ul>
 * 
 * Results can be printed using the following keywords within a print format: <br>
 * <dl>
 * <dt>
 * mean</dt>
 * <dd>
 * mean value of the registered values per batch</dd>
 * <dt>
 * cintmean</dt>
 * <dd>
 * confidence interval of the mean values in +/- notation</dd>
 * <dt>
 * numberofevents</dt>
 * <dd>
 * number of calls to the update() method over all batches</dd>
 * <dt>
 * numberofbatches</dt>
 * <dd>
 * number of all batches</dd>
 * </dl>
 * 
 */
public abstract class BatchSumStatistic extends Statistic implements Printable, UnweightedSampleCollector<Long> {

   public BatchSumStatistic(SimNode ownNode) {
      super(ownNode);

      simNode.setPrintHandler(this);
   }

   @Override
   public void printResults(ResultType type, PrintResultWriter writer) throws PrintException {
      switch (type) {
      case FINAL:
         printComments(writer);
         writer.writeDouble(ResultKeywords.MEAN, getMean());
         writer.writeDouble(ResultKeywords.CINT, getConfidenceInterval());
         writer.writeDouble(ResultKeywords.DEVIATION, getStandardDeviation());
         writer.writeDouble(ResultKeywords.MAX, getMaximum());
         writer.writeDouble(ResultKeywords.MIN, getMinimum());
         writer.writeLong(ResultKeywords.NUMBER_OF_EVENTS, getNumberOfEvents());
         writer.writeLong(ResultKeywords.NUMBER_OF_BATCHES, getNumberOfBatches());
         break;

      case BATCH:
         writer.writeLong(ResultKeywords.BATCH_SUM, getBatchSum());
         break;
      }
   }

   // -- Results --

   public abstract void update(long sample);

   @Override
   public void update(Long value) {
      update(value.longValue());
   }

   public abstract double getMaximum();

   public abstract double getMinimum();

   public abstract double getMean();

   public abstract long getNumberOfEvents();

   public abstract long getBatchSum();

   public abstract long getNumberOfBatchEvents();

   public abstract long getNumberOfBatches();

   public abstract double getConfidenceInterval();

   public abstract double getStandardDeviation();
}
