/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.statistics.batchsum;

import ikr.simlib.model.SimNode;
import ikr.simlib.result.BatchExportable;
import ikr.simlib.result.ResultKeyword;
import ikr.simlib.result.ResultReader;
import ikr.simlib.result.ResultWriter;
import ikr.simlib.result.exceptions.ImportException;
import ikr.simlib.result.exceptions.PrintException;
import ikr.simlib.statistics.Range;
import ikr.simlib.statistics.Summation;

/**
 * <code>StdBatchSumStatistic</code> extends
 * {@link ikr.simlib.statistics.batchsum.BatchSumStatistic} and provides an
 * implementation for a batch sum statistic.<br>
 * 
 */
public class StdBatchSumStatistic extends BatchSumStatistic implements BatchExportable {

   private static final String EXPORT_SUM = ResultKeyword.asExport("sum");
   private static final String EXPORT_SAMPLES = ResultKeyword.asExport("samples");

   protected long sum;
   protected long batchSamples;
   protected long totalSamples;
   protected final Summation measuresSummation;
   protected final Range measuresRange;

   public StdBatchSumStatistic(SimNode ownNode) {
      super(ownNode);

      simNode.setExportHandler(this);

      totalSamples = 0;

      measuresSummation = new Summation();
      measuresRange = new Range();
   }

   @Override
   public void exportBatch(ResultWriter writer) throws PrintException {
      writer.writeLong(EXPORT_SUM, sum);
      writer.writeLong(EXPORT_SAMPLES, batchSamples);
   }

   @Override
   public void importBatch(ResultReader reader) throws ImportException {
      sum = reader.readLong(EXPORT_SUM);
      batchSamples = reader.readLong(EXPORT_SAMPLES);
      computeMeasures(0);
   }

   @Override
   public double getConfidenceInterval() {
      return measuresSummation.getConfidenceInterval();
   }

   @Override
   public double getStandardDeviation() {
      return measuresSummation.getStandardDeviation();
   }

   @Override
   public double getMaximum() {
      return measuresRange.getMaximum();
   }

   @Override
   public double getMean() {
      return measuresSummation.getMean();
   }

   @Override
   public double getMinimum() {
      return measuresRange.getMinimum();
   }

   @Override
   public long getNumberOfBatchEvents() {
      return batchSamples;
   }

   @Override
   public long getBatchSum() {
      return sum;
   }

   @Override
   public long getNumberOfBatches() {
      return measuresSummation.getNumberOfSamples();
   }

   @Override
   public long getNumberOfEvents() {
      return totalSamples;
   }

   @Override
   public void update(long sample) {
      sum += sample; // Counting samples
      batchSamples++;
   }

   @Override
   public void computeMeasures(int batchNumber) {
      measuresSummation.update(sum);
      measuresRange.update(sum);
      totalSamples += batchSamples;
   }

   @Override
   public void resetBatchStatistic() {
      sum = 0;
      batchSamples = 0;
   }

   @Override
   public void resetStatistic() {
      totalSamples = 0;
      measuresSummation.reset();
      measuresRange.reset();
      resetBatchStatistic();
   }
}
