/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.statistics.boundary;

import ikr.simlib.exceptions.SimulationException;
import ikr.simlib.model.SimNode;
import ikr.simlib.result.BatchExportable;
import ikr.simlib.result.ResultKeyword;
import ikr.simlib.result.ResultReader;
import ikr.simlib.result.ResultWriter;
import ikr.simlib.result.exceptions.ImportException;
import ikr.simlib.result.exceptions.PrintException;

import java.util.ArrayList;
import java.util.Collections;

/**
 * <code>StdBoundaryStatistic</code> provides an implementation of
 * <code>BoundaryStatistic</code> to calculate quantiles for a given number of
 * samples and percentage. Note that the percentage-parameter in the constructor
 * must be between 0 and 1.
 */
public class StdBoundaryStatistic extends BoundaryStatistic implements BatchExportable {

   private final double percentage;
   private final int listSize;
   private int noOfSamples;
   private int noOfSamplesSinceBatchBegin;

   private static final String EXPORT_NUMBER_OF_OBSERVED_SAMPLES_SINCE_BATCH_BEGIN = ResultKeyword.asExport("noOfObservedSamples");
   private static final String EXPORT_NUMBER_OF_STORED_SAMPLES_SINCE_BATCH_BEGIN = ResultKeyword.asExport("noOfStoredSamples");
   private static final String EXPORT_SAMPLES = ResultKeyword.asExport("samples");
   private static final String EXPORT_SAMPLE = ResultKeyword.asExport("sample");

   /**
    * Sorted list of samples
    */
   private final ArrayList<Double> queue;
   private final ArrayList<Double> queueSinceBatchBegin;

   public StdBoundaryStatistic(int expectedNoOfSamples, double percentage, SimNode ownNode) {
      super(ownNode);

      simNode.setExportHandler(this);

      if (percentage < 0 || percentage > 1) {
         throw new SimulationException("Percentage out of range");
      }

      this.percentage = percentage;
      queue = new ArrayList<Double>();
      queueSinceBatchBegin = new ArrayList<Double>();

      if (percentage < 0.5) {
         listSize = (int) Math.ceil(expectedNoOfSamples * percentage) + 1;
      } else {
         listSize = (int) Math.ceil(expectedNoOfSamples * (1.0 - percentage)) + 1;
      }

      resetStatistic();
   }

   @Override
   public int getNoOfSamples() {
      return noOfSamples;
   }

   @Override
   public double getPercentage() {
      return percentage;
   }

   @Override
   public double getQuantile() {
      final double realPos;

      if (noOfSamples == 0) {
         // no data => quantile undefined
         return Double.NaN;
      }

      if (percentage > 0.5) {
         realPos = listSize - noOfSamples * (1.0 - percentage);
         if (realPos < 0) {
            throw new SimulationException("Boundary list too short");
         }
      } else {
         realPos = noOfSamples * percentage;
         if (realPos > listSize - 1) {
            throw new SimulationException("Boundary list too short");
         }
      }

      final int lowerIndex = (int) Math.floor(realPos);
      final int upperIndex = (int) Math.ceil(realPos);
      final double lowerValue = queue.get(lowerIndex);
      final double upperValue = queue.get(upperIndex);

      final double quantile;
      if (lowerIndex == upperIndex) {
         // no interpolation in case of identical indices
         quantile = lowerValue;
      } else {
         // interpolation
         quantile = lowerValue * (realPos - lowerIndex) + upperValue * (upperIndex - realPos);
      }

      return quantile;
   }

   @Override
   public void update(double sample) {
      noOfSamples++;
      insertIntoQueue(queue, sample);

      noOfSamplesSinceBatchBegin++;
      insertIntoQueue(queueSinceBatchBegin, sample);
   }

   @Override
   public void computeMeasures(int batchNumber) {
      // Not handled
   }

   @Override
   public void resetBatchStatistic() {
      queueSinceBatchBegin.clear();
      noOfSamplesSinceBatchBegin = 0;
   }

   @Override
   public void resetStatistic() {
      queue.clear();
      noOfSamples = 0;
      queueSinceBatchBegin.clear();
      noOfSamplesSinceBatchBegin = 0;
   }

   @Override
   public void exportBatch(ResultWriter writer) throws PrintException {

      // export number of samples
      writer.writeLong(EXPORT_NUMBER_OF_OBSERVED_SAMPLES_SINCE_BATCH_BEGIN, noOfSamplesSinceBatchBegin);

      // export samples
      writer.writeLong(EXPORT_NUMBER_OF_STORED_SAMPLES_SINCE_BATCH_BEGIN, queueSinceBatchBegin.size());
      writer.writeStartArray(EXPORT_SAMPLES);
      int index = 0;
      for (Double d : queueSinceBatchBegin) {
         writer.writeStartArrayItem(index);
         writer.writeDouble(EXPORT_SAMPLE, d);
         writer.writeEndArrayItem();
         index++;
      }
      writer.writeEndArray();
   }

   @Override
   public void importBatch(ResultReader reader) throws ImportException {

      // add number of observed samples
      noOfSamples += reader.readLong(EXPORT_NUMBER_OF_OBSERVED_SAMPLES_SINCE_BATCH_BEGIN);

      // read samples
      long noOfStoredSamples = reader.readLong(EXPORT_NUMBER_OF_STORED_SAMPLES_SINCE_BATCH_BEGIN);
      reader.readStartArray(EXPORT_SAMPLES);
      for (long i = 0; i < noOfStoredSamples; i++) {
         // read sample
         reader.readStartArrayItem(i);
         double sample = reader.readDouble(EXPORT_SAMPLE);
         reader.readEndArrayItem();

         // update queue
         insertIntoQueue(queue, sample);
      }
      reader.readEndArray();
   }

   private void insertIntoQueue(ArrayList<Double> queueToModify, double sample) {
      if (queueToModify.size() >= listSize) {
         if (percentage > 0.5) {
            if (sample > queueToModify.get(0)) {
               // Override first sample
               queueToModify.set(0, sample);
               // ...and sort
               Collections.sort(queueToModify);
            }
         } else {
            int last = queueToModify.size() - 1;
            if (sample < queueToModify.get(last)) {
               queueToModify.set(last, sample);
               Collections.sort(queueToModify);
            }
         }
      } else {
         queueToModify.add(sample);
         Collections.sort(queueToModify);
      }

   }

}
