/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.statistics.distribution;

import ikr.simlib.model.SimNode;
import ikr.simlib.result.PrintResultWriter;
import ikr.simlib.result.Printable;
import ikr.simlib.result.ResultKeywords;
import ikr.simlib.result.ResultType;
import ikr.simlib.result.exceptions.PrintException;
import ikr.simlib.statistics.Statistic;
import ikr.simlib.statistics.WeightedSampleCollector;

/**
 * The classes <code>DistributionStatistic</code> or
 * {@link ikr.simlib.statistics.distribution.StdDistributionStatistic} serve to
 * record measured values with the goal of determining a discrete distribution
 * (histogram). The measured values recorded with
 * {@link #update(double)} are qualified using a raster, that is
 * composed of an upper limit value and a lower limit value as well as the
 * number of partial sections in between. The parameters for the raster must be
 * declared upon constructor call. Additional to a measured value, a weight
 * (with default 1) that is multiplied with the value can be passed with
 * {@link #update(double, double)}. <br>
 * <br>
 * The following functions are available for evaluation:
 * <ul>
 * <li>
 * Number of partial sections ({@link #getArraySize()}, generally fixed)
 * <li>
 * Over-flow probability in reference to all registered measured values with
 * confidence interval ({@link #getOverflowProbability()} or
 * {@link #getOverflowConfInterval()})
 * <li>
 * Under-flow probability in reference to all registered measured values with
 * confidence interval ({@link #getUnderflowProbability()} or
 * {@link #getUnderflowConfInterval()})
 * <li>
 * Value of discretized distribution density functions at a certain point (range
 * index) in the form of probabilities that a measured value falls into a
 * certain partial range ({@link #getProbability(int)} or
 * {@link #getProbabilityConfInterval(int)} including a confidence interval)
 * <li>
 * Value of discretized distribution functions at a certain point (range index)
 * as a sum of probabilities for the areas below that point (
 * {@link #getDistribution(int)} or {@link #getDistributionConfInterval(int)}
 * with a confidence interval)
 * <li>
 * Over-flow probability in the current batch ({@link #getBatchOverflowProb()})
 * <li>
 * Under-flow probability in the current batch ({@link #getBatchUnderflowProb()})
 * <li>
 * Value of the discretized distribution density function at a certain point in
 * reference to the current batch ({@link #getBatchProbability(int)})
 * <li>
 * Value of the discretized distribution function at a certain point in
 * reference to the current batch ({@link #getBatchDistribution(int)})
 * </ul>
 * 
 * <p>
 * For printing the results it defines the following keywords:
 * </p>
 * 
 * <dl>
 * <dt>
 * list</dt>
 * <dd>
 * prints a list of results; what exactly is printed on each line has to be
 * specified in brackets behind %[list] using the following keywords</dd>
 * <dt>
 * ubound</dt>
 * <dd>
 * upper bound of current bucket</dd>
 * <dt>
 * uprob</dt>
 * <dd>
 * underflow probability (mean of batch results)</dd>
 * <dt>
 * cintuprob</dt>
 * <dd>
 * underflow probability confidence interval</dd>
 * <dt>
 * oprob</dt>
 * <dd>
 * overflow probability (mean of batch results)</dd>
 * <dt>
 * cintoprob</dt>
 * <dd>
 * overflow probability confidence interval</dd>
 * <dt>
 * dist</dt>
 * <dd>
 * (cumulative) distribution (cdf)</dd>
 * <dt>
 * prob</dt>
 * <dd>
 * probability (density) (pdf)</dd>
 * <dt>
 * cintdist</dt>
 * <dd>
 * confidence interval for cdf/ccdf</dd>
 * <dt>
 * cintprob</dt>
 * <dd>
 * confidence interval for pdf</dd>
 * <dt>
 * buprob</dt>
 * <dd>
 * underflow probability for the current batch</dd>
 * <dt>
 * boprob</dt>
 * <dd>
 * overflow probability for the current batch</dd>
 * <dt>
 * bdist</dt>
 * <dd>
 * cdf for the current batch</dd>
 * <dt>
 * bprob</dt>
 * <dd>
 * pdf for the current batch</dd>
 * </dl>
 */
public abstract class DistributionStatistic extends Statistic implements Printable, WeightedSampleCollector<Double, Double> {

   public DistributionStatistic(SimNode ownNode) {
      super(ownNode);

      simNode.setPrintHandler(this);
   }

   @Override
   public void printResults(ResultType type, PrintResultWriter writer) throws PrintException {
      switch (type) {
      case FINAL:
         printComments(writer);
         writer.writeDouble(ResultKeywords.UNDERFLOW_PROB, getUnderflowProbability());

         writer.writeStartArray(ResultKeywords.PDF.toString());

         writer.writeStartArrayItem(0);
         writer.writeDouble(ResultKeywords.UBOUND, getLowerBound(0));
         writer.writeDouble(ResultKeywords.PROBABILITY, getUnderflowProbability());
         writer.writeDouble(ResultKeywords.PROBABILITY_CINT, getUnderflowConfInterval());
         writer.writeDouble(ResultKeywords.DISTRIBUTION, getUnderflowProbability());
         writer.writeDouble(ResultKeywords.DISTRIBUTION_CINT, getUnderflowConfInterval());
         writer.writeDouble(ResultKeywords.CDIST, 1.0 - getUnderflowProbability());
         writer.writeEndArrayItem();

         for (int i = 0; i < getArraySize(); i++) {
            writer.writeStartArrayItem(i + 1);
            writer.writeDouble(ResultKeywords.UBOUND, getUpperBound(i));
            writer.writeDouble(ResultKeywords.PROBABILITY, getProbability(i));
            writer.writeDouble(ResultKeywords.PROBABILITY_CINT, getProbabilityConfInterval(i));
            writer.writeDouble(ResultKeywords.DISTRIBUTION, getDistribution(i));
            writer.writeDouble(ResultKeywords.DISTRIBUTION_CINT, getDistributionConfInterval(i));
            writer.writeDouble(ResultKeywords.CDIST, 1.0 - getDistribution(i));
            writer.writeEndArrayItem();
         }
         writer.writeEndArray();

         break;

      case BATCH:
         writer.writeDouble(ResultKeywords.BATCH_UNDERFLOW_PROB, getBatchUnderflowProb());

         writer.writeStartArray(ResultKeywords.BATCH_PDF.toString());

         writer.writeStartArrayItem(0);
         writer.writeDouble(ResultKeywords.BATCH_UBOUND, getLowerBound(0));
         writer.writeDouble(ResultKeywords.BATCH_PROBABILITY, getBatchUnderflowProb());
         writer.writeEndArrayItem();

         for (int i = 0; i < getArraySize(); i++) {
            writer.writeStartArrayItem(i + 1);
            writer.writeDouble(ResultKeywords.BATCH_UBOUND, getUpperBound(i));
            writer.writeDouble(ResultKeywords.BATCH_PROBABILITY, getBatchProbability(i));
            writer.writeEndArrayItem();
         }
         writer.writeEndArray();

         break;
      }
   }

   public abstract void update(double sample, double weight);

   @Override
   public void update(Double sample, Double weight) {
      update(sample.doubleValue(), weight.doubleValue());
   }

   public final void update(double sample) {
      update(sample, 1);
   }

   @Override
   public void update(Double sample) {
      update(sample.doubleValue());
   }

   // -- Results --

   public abstract int getArraySize();

   public abstract double getLowerBound(int bucketIndex);

   public abstract double getUpperBound(int bucketIndex);

   public abstract double getUnderflowProbability();

   public abstract double getUnderflowConfInterval();

   public abstract double getOverflowProbability();

   public abstract double getOverflowConfInterval();

   public abstract double getProbability(int bucketIndex);

   public abstract double getProbabilityConfInterval(int bucketIndex);

   public abstract double getDistribution(int bucketIndex);

   public abstract double getDistributionConfInterval(int bucketIndex);

   public abstract double getBatchUnderflowProb();

   public abstract double getBatchOverflowProb();

   public abstract double getBatchProbability(int bucketIndex);

   public abstract double getBatchDistribution(int bucketIndex);
}
