/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.statistics.distribution.trace;

import ikr.simlib.model.SimNode;
import ikr.simlib.result.ResultReader;
import ikr.simlib.result.ResultWriter;
import ikr.simlib.result.exceptions.ImportException;
import ikr.simlib.result.exceptions.PrintException;
import ikr.simlib.statistics.distribution.trace.data.DataSource;
import ikr.simlib.statistics.distribution.trace.data.DataStorage;
import ikr.simlib.statistics.distribution.trace.data.ListDataSource;
import ikr.simlib.statistics.distribution.trace.data.ListDataStorage;
import ikr.simlib.statistics.distribution.trace.data.Sample;
import ikr.simlib.statistics.distribution.trace.quantile.QuantileDescription;

import java.util.ArrayList;
import java.util.List;

/**
 * Implementation of {@link DistributionStatisticBase} which stores all values
 * in RAM. When used with Export / Import, the data is written to the xml file
 * as array.
 * <p>
 * Memory requirements: This class stores samples as Objects of class
 * {@link Sample} in an {@link ArrayList}. Each sample contains a value and a
 * weight (both of type double). Including the object header (12 to 16 bytes), a
 * sample consumes 28 to 32 bytes of heap space. In addition, the ArrayList
 * requires 4 to 8 bytes per entry for the reference. In total, each sample
 * consumes 32 bytes (32bit) to 40 bytes (64bit). Additional memory is required
 * in a collect run or at the end of the simulation, because the batch lists are
 * copied together and the <code>Collections.sort()</code> method internally
 * creates another copy.
 */
public class MemoryDistributionStatistic extends DistributionStatisticBase {

   public MemoryDistributionStatistic(SimNode ownNode, int numberOfQuantiles, boolean writeDropDistributions) {
      super(ownNode, numberOfQuantiles, writeDropDistributions);
   }

   public MemoryDistributionStatistic(SimNode ownNode, int numberOfQuantiles) {
      super(ownNode, numberOfQuantiles);
   }

   public MemoryDistributionStatistic(SimNode ownNode, QuantileDescription quantiles, boolean writeDropDistributions) {
      super(ownNode, quantiles, writeDropDistributions);
   }

   @Override
   protected DataStorage createNewDataStorage(int batchNumber) {
      return new ListDataStorage();
   }

   /**************************************************************************
    * import & export
    ***************************************************************************/

   private final static String EXPORT_LIST_ITEMS = "ListItems";
   private final static String EXPORT_VALUES = "Values";
   private final static String EXPORT_SAMPLE = "Sample";
   private final static String EXPORT_WEIGHT = "Weight";

   @Override
   void exportBatchData(ResultWriter writer, DataSource lastBatchData) throws PrintException {
      final List<Sample> list = ((ListDataSource) lastBatchData).getList();
      writer.writeLong(EXPORT_LIST_ITEMS, list.size());
      writer.writeStartArray(EXPORT_VALUES);
      for (int i = 0; i < list.size(); i++) {
         writer.writeStartArrayItem(i);
         writer.writeDouble(EXPORT_SAMPLE, list.get(i).value);
         writer.writeDouble(EXPORT_WEIGHT, list.get(i).weight);
         writer.writeEndArrayItem();
      }
      writer.writeEndArray();
   }

   @Override
   DataSource importBatchData(ResultReader reader) throws ImportException {
      final long listSize = reader.readLong(EXPORT_LIST_ITEMS);
      assert listSize < Integer.MAX_VALUE;
      final List<Sample> values = new ArrayList<Sample>((int) listSize);
      reader.readStartArray(EXPORT_VALUES);
      for (int i = 0; i < listSize; i++) {
         reader.readStartArrayItem(i);
         final double sample = reader.readDouble(EXPORT_SAMPLE);
         final double weight = reader.readDouble(EXPORT_WEIGHT);
         reader.readEndArrayItem();
         values.add(new Sample(sample, weight));
      }
      reader.readEndArray();
      return new ListDataSource(values);
   }

}
