/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.statistics.integral;

import ikr.simlib.model.SimNode;
import ikr.simlib.result.PrintResultWriter;
import ikr.simlib.result.Printable;
import ikr.simlib.result.ResultKeywords;
import ikr.simlib.result.ResultType;
import ikr.simlib.result.exceptions.PrintException;
import ikr.simlib.statistics.Statistic;

/**
 * The classes {@link IntegralStatistic} or
 * {@link ikr.simlib.statistics.integral.StdIntegralStatistic} serve to record
 * measured values, whose absolute value is not the only important factor, but
 * also the duration of the value's appearance is of importance. Determining the
 * utilization factor of a service unit (see Figure) or the average length of a
 * queue are just a few examples. <br>
 * <br>
 * <center><img src="{@docRoot}
 * /ikr/simlib/doc-files/utilizationP58.svg" width="600"
 * alt="Utilization of a service unit"> <br>
 * Utilization of a service unit </center></td> <br>
 * The emphasis of time duration corresponds to integral creation by adding up
 * the integral sections, that are created by each call of
 * {@link #update(double)} as a product of the returned measured value of the
 * last {@link #update(double)} call (with declaration of the system time) and
 * the intermediate time interval. At the end of a batch this integral value is
 * divided by the elapsed time since batch begin in order to obtain the mean
 * value. <br>
 * <br>
 * The following functions are available for evaluation:
 * <ul>
 * <li>
 * Time-emphasized mean value across all batches ({@link #getMean()}) including
 * a confidence interval ({@link #getConfidenceInterval()})
 * <li>
 * Value range ({@link #getMinimum()} and {@link #getMaximum()}) of returned
 * measured values across all batches
 * <li>
 * Time-emphasized mean value in reference to the current batch (
 * {@link #getBatchMean()})
 * </ul>
 * 
 * Results can be printed using the following keywords within a print format: <br>
 * <dl>
 * <dt>
 * mean</dt>
 * <dd>
 * Mean value of all batch integrals each divided by its own batch duration</dd>
 * <dt>
 * cintmean</dt>
 * <dd>
 * Confidence interval of the mean values in +/- notation</dd>
 * <dt>
 * max</dt>
 * <dd>
 * Greatest of all returned measured values during simulation</dd>
 * <dt>
 * min</dt>
 * <dd>
 * Smallest of all returned measured values during simulation</dd>
 * <dt>
 * barea</dt>
 * <dd>
 * Integral sum of the current batch divided by the duration of the batch</dd>
 * </dl>
 */
public abstract class IntegralStatistic extends Statistic implements Printable {

   public IntegralStatistic(SimNode ownNode) {
      super(ownNode);

      simNode.setPrintHandler(this);
   }

   @Override
   public void printResults(ResultType type, PrintResultWriter writer) throws PrintException {
      switch (type) {
      case FINAL:
         printComments(writer);
         writer.writeDouble(ResultKeywords.MEAN, getMean());
         writer.writeDouble(ResultKeywords.CINT, getConfidenceInterval());
         writer.writeDouble(ResultKeywords.MIN, getMinimum());
         writer.writeDouble(ResultKeywords.MAX, getMaximum());
         break;

      case BATCH:
         writer.writeDouble(ResultKeywords.BATCH_AREA, getBatchMean());
         break;
      }
   }

   public abstract void update(double height);

   public abstract double getMean();

   public abstract double getMaximum();

   public abstract double getMinimum();

   public abstract double getBatchMean();

   public abstract double getConfidenceInterval();

   public abstract void setStartValues(double height);
}
