/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.statistics.integral;

import ikr.simlib.events.calendar.Calendar;
import ikr.simlib.events.time.Duration;
import ikr.simlib.events.time.PointInTime;
import ikr.simlib.exceptions.SimulationException;
import ikr.simlib.model.SimNode;
import ikr.simlib.result.BatchExportable;
import ikr.simlib.result.ResultKeyword;
import ikr.simlib.result.ResultReader;
import ikr.simlib.result.ResultWriter;
import ikr.simlib.result.exceptions.ImportException;
import ikr.simlib.result.exceptions.PrintException;
import ikr.simlib.statistics.Range;
import ikr.simlib.statistics.Summation;

/**
 * <code>StdIntegralStatistic</code> extends
 * {@link ikr.simlib.statistics.integral.IntegralStatistic} and provides an
 * implementation for an integral statistic.
 */
public class StdIntegralStatistic extends IntegralStatistic implements BatchExportable {

   private static final String EXPORT_INTEGRAL_RANGE = ResultKeyword.asExport("integralrange");
   private static final String EXPORT_INTEGRAL_SUM = ResultKeyword.asExport("integralsum");
   private static final String EXPORT_INTEGRAL_TIME = ResultKeyword.asExport("integraltime");

   private double integralSum;
   private Duration elapsedTime;
   private PointInTime lastBatchBegin;
   private PointInTime lastChange;
   private double lastHeight;
   private final Range integralRange;
   private final Summation measuresSummation;

   public StdIntegralStatistic(SimNode ownNode) {
      super(ownNode);

      simNode.setExportHandler(this);

      integralSum = 0;
      elapsedTime = null;
      lastBatchBegin = PointInTime.ZERO;
      lastChange = PointInTime.ZERO;
      lastHeight = 0;
      integralRange = new Range();
      measuresSummation = new Summation();

      resetStatistic();
   }

   @Override
   public double getBatchMean() {
      double value;
      if (elapsedTime.isGreaterThan(Duration.ZERO)) {
         value = integralSum / elapsedTime.toSeconds();
      } else {
         value = Double.NaN;
      }
      return value;
   }

   @Override
   public double getConfidenceInterval() {
      return measuresSummation.getConfidenceInterval();
   }

   @Override
   public double getMaximum() {
      return integralRange.getMaximum();
   }

   @Override
   public double getMean() {
      return measuresSummation.getMean();
   }

   @Override
   public double getMinimum() {
      return integralRange.getMinimum();
   }

   @Override
   public void setStartValues(double height) {
      lastChange = Calendar.getInstance().getSystemTime();
      lastHeight = height;
   }

   @Override
   public void update(double height) {
      final PointInTime time = Calendar.getInstance().getSystemTime();
      if (lastChange.isAfter(time)) {
         throw new SimulationException("current time is smaller than last time");
      }
      integralSum += lastHeight * (lastChange.until(time).toSeconds());
      lastChange = time;
      lastHeight = height;
      integralRange.update(height);
   }

   @Override
   public void resetBatchStatistic() {
      lastBatchBegin = Calendar.getInstance().getSystemTime();
      integralSum = 0;
      lastChange = lastBatchBegin;
      integralRange.update(lastHeight);
      elapsedTime = null;
   }

   @Override
   public void resetStatistic() {
      lastBatchBegin = Calendar.getInstance().getSystemTime();
      integralSum = 0;
      lastChange = lastBatchBegin;
      integralRange.reset();
      measuresSummation.reset();
      elapsedTime = null;
   }

   @Override
   public void computeMeasures(int batchNumber) {
      // calculate area since last change
      PointInTime time = Calendar.getInstance().getSystemTime();
      elapsedTime = lastBatchBegin.until(time);
      update(lastHeight);

      if (elapsedTime.isGreaterThan(Duration.ZERO)) {
         measuresSummation.update(integralSum / elapsedTime.toSeconds());
      } else {
         measuresSummation.update(Double.NaN);
      }
   }

   @Override
   public void exportBatch(ResultWriter writer) throws PrintException {
      integralRange.exportState(EXPORT_INTEGRAL_RANGE, writer);
      elapsedTime.exportBatch(writer, EXPORT_INTEGRAL_TIME);
      writer.writeDouble(EXPORT_INTEGRAL_SUM, integralSum);
   }

   @Override
   public void importBatch(ResultReader reader) throws ImportException {
      integralRange.joinStateWithImportState(EXPORT_INTEGRAL_RANGE, reader);

      elapsedTime = Duration.importBatch(reader, EXPORT_INTEGRAL_TIME);
      integralSum = reader.readDouble(EXPORT_INTEGRAL_SUM);

      if (elapsedTime.isGreaterThan(Duration.ZERO)) {
         measuresSummation.update(integralSum / elapsedTime.toSeconds());
      } else {
         measuresSummation.update(Double.NaN);
      }
   }

}
