/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.statistics.median;

import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.parameters.reflection.ReflectionConstructable2;
import ikr.simlib.result.BatchExportable;
import ikr.simlib.result.ResultKeyword;
import ikr.simlib.result.ResultKeywords;
import ikr.simlib.result.ResultReader;
import ikr.simlib.result.ResultWriter;
import ikr.simlib.result.exceptions.ImportException;
import ikr.simlib.result.exceptions.PrintException;
import ikr.simlib.statistics.Range;
import ikr.simlib.statistics.Summation;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

/**
 * StdMedianStatistic provides an implementation for a median statistic. If the
 * parameter <em>maxBufferSize</em> is set to zero, all samples will be stored
 * internally and the median will be calculated at the end of the simulation.
 * Since this is very memory consuming, <code>maxBufferSize</code> can be set to
 * a number &gt; 0, which determines the maximum number of samples the statistic
 * should hold in memory. The median is then calculated at the end of the
 * simulation as an approximation that is based on all samples. Often, 10000
 * samples are sufficient for a good approximation.
 */
public class StdMedianStatistic extends MedianStatistic implements BatchExportable, ReflectionConstructable2<SimNode, Parameters> {

   private static final String MAX_BUFFER_SIZE = "MaxBufferSize";

   /** as required by {@link ReflectionConstructable} */
   public static StdMedianStatistic createInstance(SimNode ownNode, Parameters pars) {
      final int maxBufferSize = pars.get(ownNode, MAX_BUFFER_SIZE).asInteger();
      return new StdMedianStatistic(ownNode, maxBufferSize);
   }

   private static final String EXPORT_SAMPLE_SUM = ResultKeyword.asExport("samplesum");
   private static final String EXPORT_SAMPLE_RANGE = ResultKeyword.asExport("samplerange");

   /**
    * <code>ValuePair</code> is a pair of a double value and an integer weight.
    */
   private static class ValuePair {

      public double value;
      public int weight;

      public ValuePair(double value, int weight) {
         this.value = value;
         this.weight = weight;
      }
   }

   // Compares the values of a pair
   private static final Comparator<ValuePair> VALUE_COMPARATOR;

   static {
      VALUE_COMPARATOR = new Comparator<ValuePair>() {
         @Override
         public int compare(ValuePair a, ValuePair b) {
            return Double.compare(a.value, b.value);
         }
      };
   }

   private final int maxElements;
   private final List<ValuePair> samples;

   private final Summation measuresSum;
   private final Summation sampleSum;
   private final Range sampleRange;

   private long events;

   /**
    * Returns the value for the median calculated from fSamples.
    */
   protected double computeMedian() {
      double value = Double.NaN;

      if (samples.size() == 1) {
         value = samples.get(0).value;
      } else if (samples.size() == 0) {
         // do nothing
      } else {

         // sort the sample buffer by value
         Collections.sort(samples, VALUE_COMPARATOR);

         double uLastY = 0.0;
         long uLastX = 0;
         long uWeightSum = 0;
         double uHalf = (double) (sampleSum.getNumberOfSamples()) / 2 + 0.5;

         for (ValuePair it : samples) {
            uWeightSum = uWeightSum + it.weight;
            if (uWeightSum > uHalf) {
               // linear interpolation
               int uDeltaX = it.weight;
               double uDeltaY = it.value - uLastY;
               value = uLastY + uDeltaY * (uHalf - uLastX) / uDeltaX;
               break;
            }
            uLastX = uWeightSum;
            uLastY = it.value;
         }
      }

      return value;
   }

   public StdMedianStatistic(SimNode ownNode, int maxBufferSize) {
      super(ownNode);

      simNode.setExportHandler(this);

      // update depends on a even maximum buffer size
      if (maxBufferSize < 0 || maxBufferSize % 2 != 0) {
         throw new IllegalArgumentException("Maximum buffer size must be >= 0 and even");
      }

      maxElements = maxBufferSize;

      if (maxElements > 0) {
         samples = new ArrayList<ValuePair>();
      } else {
         samples = new LinkedList<ValuePair>();
      }
      sampleSum = new Summation();
      measuresSum = new Summation();
      sampleRange = new Range();

      resetStatistic();
   }

   public StdMedianStatistic(SimNode ownNode) {
      this(ownNode, 0);
   }

   @Override
   public void exportBatch(ResultWriter writer) throws PrintException {
      sampleSum.exportSum(EXPORT_SAMPLE_SUM, writer);
      sampleRange.exportState(EXPORT_SAMPLE_RANGE, writer);
      writer.writeDouble(ResultKeywords.BATCH_MEDIAN.toString(), getBatchMedian());
   }

   @Override
   public void importBatch(ResultReader reader) throws ImportException {
      sampleSum.importSum(EXPORT_SAMPLE_SUM, reader);
      sampleRange.joinStateWithImportState(EXPORT_SAMPLE_RANGE, reader);
      double median = reader.readDouble(ResultKeywords.BATCH_MEDIAN.toString());

      // Now let median statistic believe there is only one sample available.
      // We know this is the median value, since we just imported it. Then
      // ComputeMedian() below will return this value as the median.
      samples.clear();
      samples.add(new ValuePair(median, 1));

      computeMeasures(0);
   }

   @Override
   public long getBatchEvents() {
      return sampleSum.getNumberOfSamples();
   }

   @Override
   public double getBatchMedian() {
      return computeMedian();
   }

   @Override
   public long getEvents() {
      return events;
   }

   @Override
   public double getMaximum() {
      return sampleRange.getMaximum();
   }

   @Override
   public double getMedian() {
      return measuresSum.getMean();
   }

   @Override
   public double getMedianConfidenceInterval() {
      return measuresSum.getConfidenceInterval();
   }

   @Override
   public double getMinimum() {
      return sampleRange.getMinimum();
   }

   @Override
   public void update(double sample) {
      events++;

      // sample buffer is full
      if (maxElements > 0 && samples.size() == maxElements) {
         // sort the sample buffer by value
         Collections.sort(samples, VALUE_COMPARATOR);

         // the first iterator defines the read position, the second one
         // the write position
         Iterator<ValuePair> it = samples.iterator();
         Iterator<ValuePair> write_it = samples.iterator();
         // loop through all samples
         while (it.hasNext()) {
            ValuePair v = it.next();
            double uVal = v.value;
            int uWeight = v.weight;
            v = it.next();
            double uVal2 = v.value;
            int uWeight2 = v.weight;
            double uNewValue = (uVal * uWeight + uVal2 * uWeight2) / (uWeight + uWeight2);

            ValuePair wv = write_it.next();
            wv.value = uNewValue;
            wv.weight = uWeight + uWeight2;
         }

         // after reducing the buffers content, now adapt it's size
         samples.subList(samples.size() / 2, samples.size()).clear();
      }

      samples.add(new ValuePair(sample, 1));
      sampleSum.update(sample);
      sampleRange.update(sample);
   }

   @Override
   public void computeMeasures(int batchNumber) {
      measuresSum.update(computeMedian());
   }

   @Override
   public void resetBatchStatistic() {
      samples.clear();
      sampleSum.reset();
   }

   @Override
   public void resetStatistic() {
      resetBatchStatistic();
      measuresSum.reset();
      sampleRange.reset();
      events = 0;
   }
}
