/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.statistics.probability;

import ikr.simlib.model.SimNode;
import ikr.simlib.result.BatchExportable;
import ikr.simlib.result.ResultKeyword;
import ikr.simlib.result.ResultReader;
import ikr.simlib.result.ResultWriter;
import ikr.simlib.result.exceptions.ImportException;
import ikr.simlib.result.exceptions.PrintException;
import ikr.simlib.statistics.Summation;

/**
 * <code>StdProbabilityStatistic</code> extends
 * {@link ikr.simlib.statistics.probability.StdProbabilityStatistic} and
 * provides an implementation for a probability statistic.
 */
public class StdProbabilityStatistic extends ProbabilityStatistic implements BatchExportable {

   private static final String EXPORT_TOTAL_SUM = ResultKeyword.asExport("sumtotal");
   private static final String EXPORT_SPECIAL_SUM = ResultKeyword.asExport("sumspecial");

   private long sumTotal;
   private long sumSpecial;
   protected final Summation measuresSummation;

   public StdProbabilityStatistic(SimNode ownNode) {
      super(ownNode);

      simNode.setExportHandler(this);

      measuresSummation = new Summation();

      resetStatistic();
   }

   @Override
   public void exportBatch(ResultWriter writer) throws PrintException {
      writer.writeLong(EXPORT_TOTAL_SUM, sumTotal);
      writer.writeLong(EXPORT_SPECIAL_SUM, sumSpecial);
   }

   @Override
   public void importBatch(ResultReader reader) throws ImportException {
      sumTotal = reader.readLong(EXPORT_TOTAL_SUM);
      sumSpecial = reader.readLong(EXPORT_SPECIAL_SUM);

      computeMeasures(0);
   }

   @Override
   public double getBatchProbability() {
      if (sumTotal != 0)
         return (double) sumSpecial / (double) sumTotal;
      return 0;
   }

   @Override
   public double getConfidenceInterval() {
      return measuresSummation.getConfidenceInterval();
   }

   @Override
   public double getMean() {
      return measuresSummation.getMean();
   }

   @Override
   public void update(long number, boolean special) {
      sumTotal += number;
      if (special)
         sumSpecial += number;
   }

   @Override
   public void computeMeasures(int batchNumber) {
      if (sumTotal != 0)
         measuresSummation.update((double) sumSpecial / (double) sumTotal);
   }

   @Override
   public void resetBatchStatistic() {
      sumTotal = 0;
      sumSpecial = 0;
   }

   @Override
   public void resetStatistic() {
      resetBatchStatistic();
      measuresSummation.reset();
   }
}
