/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.statistics.rate;

import ikr.simlib.events.calendar.Calendar;
import ikr.simlib.events.time.Duration;
import ikr.simlib.events.time.PointInTime;
import ikr.simlib.model.SimNode;
import ikr.simlib.result.BatchExportable;
import ikr.simlib.result.ResultKeyword;
import ikr.simlib.result.ResultKeywords;
import ikr.simlib.result.ResultReader;
import ikr.simlib.result.ResultWriter;
import ikr.simlib.result.exceptions.ImportException;
import ikr.simlib.result.exceptions.PrintException;
import ikr.simlib.statistics.Summation;

/**
 * <code>StdRateStatistic</code> extends
 * {@link ikr.simlib.statistics.rate.RateStatistic} and provides an
 * implementation for a rate statistic. Thereby "rate" measures that rate in
 * e.g. units whereas "samplerate" considers the rate of samples (the amount
 * passed to <code>Update(amount)</code> is not considered)
 * <p>
 * For printing results it defines the following keywords:
 * <dl>
 * <dt>rate</dt>
 * <dd>Mean of all rates (amount/time) in all batches</dd>
 * <dt>cintrate</dt>
 * <dd>Confidence interval to mean rates</dd>
 * <dt>samplerate</dt>
 * <dd>Mean of all rates (samples/time) in all batches</dd>
 * <dt>cintsamplerate</dt>
 * <dd>Confidence interval to mean rates</dd>
 * </dl>
 */
public class StdRateStatistic extends RateStatistic implements BatchExportable {

   private static final String EXPORT_SUM = ResultKeyword.asExport(ResultKeywords.SUM.toString());
   private static final String EXPORT_NUMBER_OF_SAMPLES = ResultKeyword.asExport(ResultKeywords.NUMBER_OF_SAMPLES.toString());
   private static final String EXPORT_ELAPSED_TIME = ResultKeyword.asExport("elapsedtime");

   private double sum;
   private long noOfSamples;
   private PointInTime batchStartTime;
   private Duration elapsedTime;

   private final Summation rateMeasuresSummation;
   private final Summation sampleRateMeasuresSummation;

   public StdRateStatistic(SimNode ownNode) {
      super(ownNode);

      simNode.setExportHandler(this);

      rateMeasuresSummation = new Summation();
      sampleRateMeasuresSummation = new Summation();

      resetStatistic();
   }

   @Override
   public void exportBatch(ResultWriter writer) throws PrintException {
      elapsedTime.exportBatch(writer, EXPORT_ELAPSED_TIME);
      writer.writeDouble(EXPORT_SUM, sum);
      writer.writeLong(EXPORT_NUMBER_OF_SAMPLES, noOfSamples);
   }

   @Override
   public void importBatch(ResultReader reader) throws ImportException {
      elapsedTime = Duration.importBatch(reader, EXPORT_ELAPSED_TIME);
      sum = reader.readDouble(EXPORT_SUM);
      noOfSamples = reader.readLong(EXPORT_NUMBER_OF_SAMPLES);

      rateMeasuresSummation.update(getBatchRate());
      sampleRateMeasuresSummation.update(getBatchSampleRate());
   }

   @Override
   public double getRate() {
      return rateMeasuresSummation.getMean();
   }

   @Override
   public double getRateConfInterval() {
      return rateMeasuresSummation.getConfidenceInterval();
   }

   @Override
   public double getSampleRate() {
      return sampleRateMeasuresSummation.getMean();
   }

   @Override
   public double getSampleRateConfInterval() {
      return sampleRateMeasuresSummation.getConfidenceInterval();
   }

   @Override
   public void update(double amount) {
      sum += amount;
      noOfSamples++;
   }

   @Override
   public double getBatchRate() {
      double value;
      if (elapsedTime.isGreaterThan(Duration.ZERO)) {
         value = sum / elapsedTime.toSeconds();
      } else {
         value = Double.NaN;
      }
      return value;
   }

   @Override
   public double getBatchSampleRate() {
      double value;
      if (elapsedTime.isGreaterThan(Duration.ZERO)) {
         value = noOfSamples / elapsedTime.toSeconds();
      } else {
         value = Double.NaN;
      }
      return value;
   }

   @Override
   public void computeMeasures(int batchNumber) {
      elapsedTime = batchStartTime.until(Calendar.getInstance().getSystemTime());
      rateMeasuresSummation.update(getBatchRate());
      sampleRateMeasuresSummation.update(getBatchSampleRate());
   }

   @Override
   public void resetBatchStatistic() {
      elapsedTime = null;
      sum = 0;
      noOfSamples = 0;
      batchStartTime = Calendar.getInstance().getSystemTime();
   }

   @Override
   public void resetStatistic() {
      resetBatchStatistic();
      rateMeasuresSummation.reset();
      sampleRateMeasuresSummation.reset();
   }
}
