/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.statistics.sample;

import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.parameters.reflection.ReflectionConstructable2;
import ikr.simlib.statistics.Statistic;
import ikr.simlib.statistics.Statistics;
import ikr.simlib.statistics.Statistics.Predicate;
import ikr.simlib.statistics.UnweightedSampleCollector;
import ikr.simlib.statistics.conditional.ConditionalUnweightedSampleCollector;
import ikr.simlib.statistics.conditional.implementations.ConditionalStdSampleStatistic;
import ikr.simlib.statistics.support.BucketUtility;
import ikr.simlib.statistics.support.LinBucketUtility;

public abstract class CondMeanStatistic extends Statistic implements ReflectionConstructable2<SimNode, Parameters>, ConditionalUnweightedSampleCollector<Double, Double> {

   private CondMeanStatistic(SimNode ownNode) {
      super(ownNode);
   }

   private static final String ARRAY_SIZE = "ArraySize";
   private static final String LOWER_BOUND = "LowerBound";
   private static final String UPPER_BOUND = "UpperBound";

   /** as required by {@link ReflectionConstructable} */
   public static ConditionalUnweightedSampleCollector<Double, Double> createInstance(SimNode ownNode, Parameters pars) {
      final int arraySize = pars.get(ownNode, ARRAY_SIZE).asInteger();
      final double lowerBound = pars.get(ownNode, LOWER_BOUND).asDouble();
      final double upperBound = pars.get(ownNode, UPPER_BOUND).asDouble();
      return create(ownNode, lowerBound, upperBound, arraySize);
   }

   /**
    * Convenience method which creates something that closely resembles an
    * object of class {@link CondMeanStatistic} as present in previous versions
    * of this library.
    * <p>
    * Values are recorded in an overall statistic and in one out of a set of
    * statistics which depend on the key (or condentiator).
    * <p>
    * The difference to the old implementation is that the mean of the bucket is
    * not printed to the result file (it does not make sense for all types of
    * buckets, e.g. logarithmic ones).
    * 
    * @param ownNode
    *           a {@link SimNode}, which serves as the parent for the overall
    *           and the bucket statistics
    * @param bucketUtility
    *           a class defining the buckets
    * @return the combined statistic
    */
   @SuppressWarnings("unchecked")
   public static ConditionalUnweightedSampleCollector<Double, Double> create(SimNode ownNode, final BucketUtility bucketUtility) {
      final UnweightedSampleCollector<Double> overallStatistic = new StdSampleStatistic(ownNode.createChildNode("Overall"));
      final ConditionalUnweightedSampleCollector<Double, Double> condStat = Statistics.filterByKey(new ConditionalStdSampleStatistic<Double>(ownNode, Double.class, bucketUtility), new Predicate<Double>() {
         @Override
         public boolean test(Double input) {
            return bucketUtility.isInRange(input);
         }
      });
      // this is not unsafe, but only Java version 1.7 or newer recognize that
      return Statistics.combine(Statistics.asConditionalStatistic(overallStatistic), condStat);
   }

   /** @see #create(SimNode, BucketUtility) */
   public static ConditionalUnweightedSampleCollector<Double, Double> create(SimNode ownNode, double min, double max, int noOfBuckets) {
      return create(ownNode, new LinBucketUtility(min, max, noOfBuckets));
   }

}