/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.statistics.sample;

import ikr.simlib.model.SimNode;
import ikr.simlib.result.BatchExportable;
import ikr.simlib.result.ResultKeyword;
import ikr.simlib.result.ResultReader;
import ikr.simlib.result.ResultWriter;
import ikr.simlib.result.exceptions.ImportException;
import ikr.simlib.result.exceptions.PrintException;
import ikr.simlib.statistics.Range;
import ikr.simlib.statistics.Summation;

/**
 * <code>StdSampleStatistic</code> provides an implementation for a sample
 * statistic. See {@link ikr.simlib.statistics.sample.SampleStatistic}
 */
public class StdSampleStatistic extends SampleStatistic implements BatchExportable {

   private static final String EXPORT_SAMPLE_SUM = ResultKeyword.asExport("samplesum");
   private static final String EXPORT_SAMPLE_RANGE = ResultKeyword.asExport("samplerange");

   private final Summation measuresSum;
   private final Summation sampleSum;
   private final Summation deviationSum;
   private final Summation coefficientSum;
   private final Range measuresRange;
   private final Range sampleRange;
   private final Range deviationRange;
   private final Range coefficientRange;
   private long events;

   public StdSampleStatistic(SimNode ownNode) {
      super(ownNode);

      simNode.setExportHandler(this);

      measuresSum = new Summation();
      sampleSum = new Summation();
      deviationSum = new Summation();
      coefficientSum = new Summation();
      measuresRange = new Range();
      sampleRange = new Range();
      deviationRange = new Range();
      coefficientRange = new Range();

      resetStatistic();
   }

   @Override
   public void exportBatch(ResultWriter writer) throws PrintException {
      sampleSum.exportSum(EXPORT_SAMPLE_SUM, writer);
      sampleRange.exportState(EXPORT_SAMPLE_RANGE, writer);
   }

   @Override
   public void importBatch(ResultReader reader) throws ImportException {
      sampleSum.importSum(EXPORT_SAMPLE_SUM, reader);
      sampleRange.joinStateWithImportState(EXPORT_SAMPLE_RANGE, reader);

      events += sampleSum.getNumberOfSamples();

      computeMeasures(0);
   }

   @Override
   public void update(double sample) {
      sampleSum.update(sample);
      sampleRange.update(sample);
      events++;
   }

   @Override
   public double getBatchCoefficient() {
      return sampleSum.getCoefficientOfVariation();
   }

   @Override
   public long getBatchEvents() {
      return sampleSum.getNumberOfSamples();
   }

   @Override
   public double getBatchMean() {
      return sampleSum.getMean();
   }

   @Override
   public double getBatchVariance() {
      return sampleSum.getVariance();
   }

   @Override
   public double getCoefficient() {
      return coefficientSum.getMean();
   }

   @Override
   public double getCoefficientConfidenceInterval() {
      return coefficientSum.getConfidenceInterval();
   }

   @Override
   public double getDeviation() {
      return deviationSum.getMean();
   }

   @Override
   public double getDeviationConfidenceInterval() {
      return deviationSum.getConfidenceInterval();
   }

   @Override
   public double getMaximum() {
      return sampleRange.getMaximum();
   }

   @Override
   public double getMean() {
      return measuresSum.getMean();
   }

   @Override
   public double getMeanConfidenceInterval() {
      return measuresSum.getConfidenceInterval();
   }

   @Override
   public double getMinimum() {
      return sampleRange.getMinimum();
   }

   @Override
   public double getBatchStandardDeviation() {
      return sampleSum.getStandardDeviation();
   }

   @Override
   public long getEvents() {
      return events;
   }

   @Override
   public void computeMeasures(int batchNumber) {
      double auxTerm;

      auxTerm = sampleSum.getMean();
      measuresSum.update(auxTerm);
      measuresRange.update(auxTerm);

      auxTerm = sampleSum.getStandardDeviation();
      deviationSum.update(auxTerm);
      deviationRange.update(auxTerm);

      auxTerm = sampleSum.getCoefficientOfVariation();
      coefficientSum.update(auxTerm);
      coefficientRange.update(auxTerm);
   }

   @Override
   public void resetBatchStatistic() {
      sampleSum.reset();
   }

   @Override
   public void resetStatistic() {
      resetBatchStatistic();
      measuresSum.reset();
      deviationSum.reset();
      coefficientSum.reset();
      sampleRange.reset();
      measuresRange.reset();
      deviationRange.reset();
      coefficientRange.reset();
      events = 0;
   }

}
