/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.support;

/**
 * This class contains support functions for vector math.
 * 
 */
public final class Vectors {
   /**
    * Adds each element from two vectors and returns the result as a new vector.
    * 
    * @return the new vector containing the per element sum
    */
   public static double[] add(double[] vecA, double[] vecB) {
      if (vecA.length != vecB.length) {
         throw new IllegalArgumentException("Vector lengths are not equal");
      }

      final int N = vecA.length;

      final double[] result = new double[N];

      for (int n = 0; n < N; n++) {
         result[n] = vecA[n] + vecB[n];
      }

      return result;
   }

   /**
    * Creates a new vector containing the larger elements from both vectors,
    * that is <code>r[n] = max(a[n], b[n])</code>.
    * 
    * @return the new vector containing the larger elements
    */
   public static double[] max(double[] vecA, double[] vecB) {
      if (vecA.length != vecB.length) {
         throw new IllegalArgumentException("Vector lengths are not equal");
      }

      final int N = vecA.length;

      final double[] result = new double[N];

      for (int n = 0; n < N; n++) {
         result[n] = Math.max(vecA[n], vecB[n]);
      }

      return result;
   }

   /**
    * Creates a new vector containing the smaller elements from both vectors,
    * that is <code>r[n] = min(a[n], b[n])</code>.
    * 
    * @return the new vector containing the smaller elements
    */
   public static double[] min(double[] vecA, double[] vecB) {
      if (vecA.length != vecB.length) {
         throw new IllegalArgumentException("Vector lengths are not equal");
      }

      final int N = vecA.length;

      final double[] result = new double[N];

      for (int n = 0; n < N; n++) {
         result[n] = Math.min(vecA[n], vecB[n]);
      }

      return result;
   }

   /**
    * Calculates the sum of a vector.
    * 
    * @return the sum of all elements
    */
   public static double sum(double[] vec) {
      double sum = 0;

      for (double v : vec) {
         sum += v;
      }

      return sum;
   }

   /**
    * Calculates the (geometric) length of a vector.
    * 
    * @return the geometric length of the vector
    */
   public static double length(double[] vec) {
      double sum = 0;

      for (double v : vec) {
         sum += v * v;
      }

      return Math.sqrt(sum);
   }

   /**
    * Normalizes all elements of a vector.
    */
   public static void normalize(double[] vec) {
      final double sum = sum(vec);

      if (sum == 0) {
         throw new IllegalArgumentException("Sum of vector mustn't be 0");
      }

      final int N = vec.length;

      for (int n = 0; n < N; n++) {
         vec[n] *= sum;
      }
   }
}
