/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.untested.entities.generators;

import ikr.simlib.entities.generators.Generator;
import ikr.simlib.events.time.PointInTime;
import ikr.simlib.factories.Factory;
import ikr.simlib.messages.Message;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;
import ikr.simlib.untested.trace.TraceFile;

/**
 * TraceGenerator generates messages at timestamps, which are given in a trace
 * file. Further the length of the message is provided by the trace file. The
 * type of the message depends on the default message, handled by the according
 * message factory. <code>TraceGenerator</code> has two
helper classes, namely {@link ikr.simlib.untested.trace.TraceFile} and {@link ikr.simlib.untested.trace.TraceItem}. The trace file must have have following
format:<br>
hh:mm:ss.sss length<br>
The time information can either be absolute or relative to the preeceding event (= interarrival
time), this can be set by the boolean attribute <i>absoluteTime</i> in the constructor, or in the
parameter file when the generator is parsed. Further the behaviour at the end of the trace file
can be determined. If <i>wrapFile</i> is set to true, the trace file is reread from the beginning, but
in the case of absolute time only the IAT between two succeeding events is considered. In the
other case the simulation interrupts with an error, if the end of the trace file is reached. Comments
are initiated by a leading '#'.
 */
public class TraceGenerator extends Generator {
   
   private static final String TRACE_FILE = "TraceFile";
   
   /** as required by {@link ReflectionConstructable} */
   public static TraceGenerator createInstance(SimNode ownNode, Parameters pars) {
      String traceFileName = pars.get(ownNode, TRACE_FILE).asString();
      return new TraceGenerator(traceFileName, ownNode, null);
   }

   protected TraceFile traceFile;
   protected int msgCounter;

   public TraceGenerator(String filename, SimNode ownNode, Factory<Message> factory, boolean activateOnStart) {
      this(filename, ownNode, true, false, factory, activateOnStart);
   }

   public TraceGenerator(String filename, SimNode ownNode, Factory<Message> factory) {
      this(filename, ownNode, factory, true);
   }

   public TraceGenerator(String filename, SimNode ownNode, boolean absoluteTime, boolean wrapFile, Factory<Message> factory, boolean activateOnStart) {
      super(ownNode, factory, activateOnStart);

      traceFile = new TraceFile(filename, absoluteTime, wrapFile);
   }

   public TraceGenerator(String filename, SimNode ownNode, boolean absoluteTime, boolean wrapFile, Factory<Message> factory) {
      this(filename, ownNode, absoluteTime, wrapFile, factory, true);
   }

   @Override
   protected Message createMessage() {
      msgCounter++;
      Message tempMes = super.createMessage();
      tempMes.setLength(traceFile.getNextLength());
      return tempMes;
   }

   @Override
   protected PointInTime getNextEventTime(PointInTime currentTime) {
      return traceFile.getNextEventTime(currentTime);
   }
}
