/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.untested.statistics.correlation;

import ikr.simlib.model.SimNode;
import ikr.simlib.result.BatchExportable;
import ikr.simlib.result.ResultKeyword;
import ikr.simlib.result.ResultReader;
import ikr.simlib.result.ResultWriter;
import ikr.simlib.result.exceptions.ImportException;
import ikr.simlib.result.exceptions.PrintException;
import ikr.simlib.statistics.Summation;

/**
 * <code>StdAutoCorrelationStatistic</code> provides a concrete implementation
 * of an auto-correlation statistic. The lag can be defined as input parameter.
 * 
 * <p>
 * The following keywords are defined for printing the results:
 * </p>
 * 
 * <dl>
 * <dt>coeff</dt>
 * <dd>coefficients of auto-correlation</dd>
 * <dt>coeffcint</dt>
 * <dd>coefficients of auto-correlation +/- confidence interval of corr. coeff.</dd>
 * <dt>coeff</dt>
 * <dd>auto-covariances</dd>
 * <dt>coeffcint</dt>
 * <dd>auto-covariances +/- confidence interval of covariance</dd>
 * <dt>lag</dt>
 * <dd>lag (look-behind range)</dd>
 * </dl>
 * 
 * <p>
 * A more detailed description can be found in the student thesis of Luis Lopez
 * (No 1535) or in the dissertaion of Juergen Enssle (LDB 28226).
 * </p>
 */
public class StdAutoCorrelationStatistic extends AutoCorrelationStatistic implements BatchExportable {

   private static final String EXPORT_XY_SAMPLE_SUM = ResultKeyword.asExport("xysamplesum");
   private static final String EXPORT_X_SAMPLE_SUM = ResultKeyword.asExport("xsamplesum");

   protected int lag;

   /**
    * Last <code>lag</code> samples
    */
   protected double[] sampleArray;

   protected Summation xSampleSummation;
   protected Summation[] xySampleSummation;
   protected int[] noOfSamples;
   protected int currentPos;

   protected Summation[] coeffMeasures;
   protected Summation[] covarianceMeasures;

   public StdAutoCorrelationStatistic(int lag, SimNode ownNode) {
      super(ownNode);

      if (lag <= 0) {
         throw new IllegalArgumentException("Lag must be at least 1");
      }

      this.lag = lag;
      sampleArray = new double[lag];
      xSampleSummation = new Summation();
      xySampleSummation = new Summation[lag];
      noOfSamples = new int[lag];
      currentPos = 0;
      coeffMeasures = new Summation[lag];
      covarianceMeasures = new Summation[lag];

      for (int i = 0; i < lag; i++) {
         sampleArray[i] = Double.NaN;

         xySampleSummation[i] = new Summation();
         coeffMeasures[i] = new Summation();
         covarianceMeasures[i] = new Summation();
      }

      resetStatistic();
   }

   @Override
   public void exportBatch(ResultWriter writer) throws PrintException {
      xSampleSummation.exportSum(EXPORT_X_SAMPLE_SUM, writer);

      for (int n = 0; n < xySampleSummation.length; n++) {
         xySampleSummation[n].exportSum(EXPORT_XY_SAMPLE_SUM + n, writer);
      }
   }

   @Override
   public void importBatch(ResultReader reader) throws ImportException {
      xSampleSummation.importSum(EXPORT_X_SAMPLE_SUM, reader);

      for (int n = 0; n < xySampleSummation.length; n++) {
         xySampleSummation[n].importSum(EXPORT_XY_SAMPLE_SUM + n, reader);
      }

      computeMeasures(0);
   }

   @Override
   public double getCoefficient(int index) {
      return coeffMeasures[index].getMean();
   }

   @Override
   public double getCoefficientConfidenceInterval(int index) {
      return coeffMeasures[index].getConfidenceInterval();
   }

   @Override
   public double getCovariance(int index) {
      return covarianceMeasures[index].getMean();
   }

   @Override
   public double getCovarianceConfidenceInterval(int index) {
      return covarianceMeasures[index].getConfidenceInterval();
   }

   @Override
   public int getLag() {
      return lag;
   }

   @Override
   public double getNoOfSamples(int index) {
      return noOfSamples[index];
   }

   @Override
   public void resetSampleArray() {
      for (int i = 0; (i < lag) && !Double.isNaN(sampleArray[i]); i++) {
         sampleArray[i] = Double.NaN;
      }
      currentPos = 0;
   }

   @Override
   public void update(double sample, boolean useValue) {
      if (useValue) {
         for (int k = 0, arrayPos = (currentPos - 1 + lag) % lag; (k < lag) && !Double.isNaN(sampleArray[arrayPos]); k++, arrayPos = (arrayPos - 1 + lag) % lag) {
            xySampleSummation[k].update(sample * sampleArray[arrayPos]);
         }
         xSampleSummation.update(sample);
      }
      sampleArray[currentPos] = sample;
      currentPos = (currentPos + 1) % lag;
   }

   @Override
   public void computeMeasures(int batchNumber) {
      double meanX = xSampleSummation.getMean();
      double sigmaX = xSampleSummation.getStandardDeviation();
      for (int k = 0; (k < lag) && (xySampleSummation[k].getNumberOfSamples() > 0); k++) {
         double meanXY = xySampleSummation[k].getMean();

         double covariance = meanXY - meanX * meanX;
         covarianceMeasures[k].update(covariance);

         if (sigmaX > 0) {
            double coeffOfCorr = covariance / (sigmaX * sigmaX);
            coeffMeasures[k].update(coeffOfCorr);
         }
         noOfSamples[k] += xySampleSummation[k].getNumberOfSamples();
      }
   }

   @Override
   public void resetBatchStatistic() {
      for (int i = 0; i < lag; i++) {
         xySampleSummation[i].reset();
      }
      xSampleSummation.reset();
   }

   @Override
   public void resetStatistic() {
      resetBatchStatistic();
      for (int i = 0; i < lag; i++) {
         coeffMeasures[i].reset();
         covarianceMeasures[i].reset();
         noOfSamples[i] = 0;
      }
      currentPos = 0;
   }

}
