/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.untested.statistics.correlation;

import ikr.simlib.model.SimNode;
import ikr.simlib.result.BatchExportable;
import ikr.simlib.result.ResultKeyword;
import ikr.simlib.result.ResultReader;
import ikr.simlib.result.ResultWriter;
import ikr.simlib.result.exceptions.ImportException;
import ikr.simlib.result.exceptions.PrintException;
import ikr.simlib.statistics.Summation;

/**
 * The class <code>StdCorrelationStatistic</code> provides an implementation to
 * calculate the coefficient of correlation and the covariance between 2 series
 * of samples.
 * <p>
 * Following keywords are defined for printing the results:
 * </p>
 * <dl>
 * <dt>coeff</dt>
 * <dd>coefficient of correlation</dd>
 * <dt>cintcoeff</dt>
 * <dd>confidence interval of corr. coeff.</dd>
 * <dt>pmcintcoeff</dt>
 * <dd>cintcoeff in +/- notation</dd>
 * <dt>cov</dt>
 * <dd>covariance</dd>
 * <dt>cintcov</dt>
 * <dd>confidence interval of covariance</dd>
 * <dt>pmcintcov</dt>
 * <dd>cintcov in +/- notation</dd>
 * </dl>
 */
public class StdCorrelationStatistic extends CorrelationStatistic implements BatchExportable {

   private static final String EXPORT_XY_SAMPLE_SUM = ResultKeyword.asExport("xysamplesum");
   private static final String EXPORT_X_SAMPLE_SUM = ResultKeyword.asExport("xsamplesum");
   private static final String EXPORT_Y_SAMPLE_SUM = ResultKeyword.asExport("ysamplesum");

   protected double xySampleSummation;
   protected final Summation xSampleSummation;
   protected final Summation ySampleSummation;

   protected final Summation coeffMeasures;
   protected final Summation covarianceMeasures;

   public StdCorrelationStatistic(SimNode ownNode) {
      super(ownNode);

      simNode.setExportHandler(this);

      xSampleSummation = new Summation();
      ySampleSummation = new Summation();

      coeffMeasures = new Summation();
      covarianceMeasures = new Summation();

      resetStatistic();
   }

   @Override
   public void exportBatch(ResultWriter writer) throws PrintException {
      writer.writeDouble(EXPORT_XY_SAMPLE_SUM, xySampleSummation);
      xSampleSummation.exportSum(EXPORT_X_SAMPLE_SUM, writer);
      ySampleSummation.exportSum(EXPORT_Y_SAMPLE_SUM, writer);
   }

   @Override
   public void importBatch(ResultReader reader) throws ImportException {
      xySampleSummation = reader.readDouble(EXPORT_XY_SAMPLE_SUM);
      xSampleSummation.importSum(EXPORT_X_SAMPLE_SUM, reader);
      ySampleSummation.importSum(EXPORT_Y_SAMPLE_SUM, reader);

      computeMeasures(0);
   }

   @Override
   public double getCoefficient() {
      return coeffMeasures.getMean();
   }

   @Override
   public double getCoefficientConfidenceInterval() {
      return coeffMeasures.getConfidenceInterval();
   }

   @Override
   public double getCovariance() {
      return covarianceMeasures.getMean();
   }

   @Override
   public double getCovarianceConfidenceInterval() {
      return covarianceMeasures.getConfidenceInterval();
   }

   @Override
   public void update(double xSample, double ySample) {
      xSampleSummation.update(xSample);
      ySampleSummation.update(ySample);
      xySampleSummation += xSample * ySample;
   }

   @Override
   public void computeMeasures(int batchNumber) {
      long noOfSamples = xSampleSummation.getNumberOfSamples();

      if (noOfSamples > 0) {
         double meanX = xSampleSummation.getMean();
         double meanY = ySampleSummation.getMean();
         double meanXY = xySampleSummation / noOfSamples;
         double sigmaX = xSampleSummation.getStandardDeviation();
         double sigmaY = ySampleSummation.getStandardDeviation();

         double covariance = (meanXY - meanX * meanY) * noOfSamples / (noOfSamples - 1);
         covarianceMeasures.update(covariance);

         if (sigmaX * sigmaY > 0) {
            double coeffOfCorr = covariance / (sigmaX * sigmaY);
            coeffMeasures.update(coeffOfCorr);
         }
      }
   }

   @Override
   public void resetBatchStatistic() {
      xSampleSummation.reset();
      ySampleSummation.reset();
      xySampleSummation = 0;
   }

   @Override
   public void resetStatistic() {
      resetBatchStatistic();
      coeffMeasures.reset();
      covarianceMeasures.reset();
   }
}
