/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.untested.statistics.fairness;

import ikr.simlib.exceptions.SimulationException;
import ikr.simlib.model.SimNode;
import ikr.simlib.result.BatchExportable;
import ikr.simlib.result.PrintResultWriter;
import ikr.simlib.result.Printable;
import ikr.simlib.result.ResultKeyword;
import ikr.simlib.result.ResultKeywords;
import ikr.simlib.result.ResultReader;
import ikr.simlib.result.ResultType;
import ikr.simlib.result.ResultWriter;
import ikr.simlib.result.exceptions.ImportException;
import ikr.simlib.result.exceptions.PrintException;
import ikr.simlib.statistics.Statistic;
import ikr.simlib.statistics.Summation;

import java.io.FileNotFoundException;
import java.io.PrintWriter;

public class FairnessStatistic extends Statistic implements BatchExportable, Printable {

   private static final String LIST = ResultKeyword.asExport(ResultKeywords.LIST.toString());

   private static final String NUMBER_OF_BLOCK = ResultKeyword.asExport("noofblock");
   private static final String BLOCK_COV_SUM = ResultKeyword.asExport("blockcovsum");
   private static final String BLOCK_FI_SUM = ResultKeyword.asExport("blockfisum");

   protected String traceFileName;
   protected PrintWriter traceFile;

   protected final int ldMinBlockSize;
   protected final int ldMaxBlockSize;

   protected final int minBlockSize;
   protected final Summation[] covBatchMeans;
   /** Fairness index */
   protected final Summation[] fiBatchMeans;
   protected final double[] blockCoVSums;
   protected final double[] blockFISums;
   protected final long[] noOfBlocks;

   public FairnessStatistic(int ldMinBlockSize, int ldMaxBlockSize, String traceFileName, SimNode ownNode) {
      super(ownNode);

      simNode.setPrintHandler(this);
      simNode.setExportHandler(this);

      if (ldMaxBlockSize >= 32) {
         throw new SimulationException("maximum block size must be < 2^32");
      }

      if (ldMinBlockSize > ldMaxBlockSize) {
         throw new SimulationException("maximum block size must be >= minimum block size");
      }

      this.ldMinBlockSize = ldMinBlockSize;
      this.ldMaxBlockSize = ldMaxBlockSize;
      this.traceFileName = traceFileName;
      minBlockSize = 1 << ldMinBlockSize;
      covBatchMeans = new Summation[ldMaxBlockSize - ldMinBlockSize + 1];
      fiBatchMeans = new Summation[ldMaxBlockSize - ldMinBlockSize + 1];
      blockCoVSums = new double[ldMaxBlockSize - ldMinBlockSize + 1];
      blockFISums = new double[ldMaxBlockSize - ldMinBlockSize + 1];
      noOfBlocks = new long[ldMaxBlockSize - ldMinBlockSize + 1];

      for (int i = 0; i <= ldMaxBlockSize - ldMinBlockSize; i++) {
         covBatchMeans[i] = new Summation();
         fiBatchMeans[i] = new Summation();
      }

      resetStatistic();
   }

   @Override
   public void printResults(ResultType type, PrintResultWriter writer) throws PrintException {
      final int maxElements = ldMaxBlockSize - ldMinBlockSize + 1;
      int elements = 0;

      switch (type) {
      case BATCH:
         elements = 0;
         while (elements < maxElements && !Double.isNaN(getBatchCoV(elements))) {
            elements++;
         }

         writer.writeStartArray(ResultKeywords.BATCH_COV_LIST.toString());
         for (int i = 0; i < elements; i++) {
            writer.writeStartArrayItem(i);
            final int blockSize = (int) Math.pow(minBlockSize, i);
            writer.writeDouble(ResultKeywords.BATCH_BLOCKSIZE, blockSize);
            writer.writeDouble(ResultKeywords.BATCH_COV, getBatchCoV(i));
            writer.writeEndArrayItem();
         }
         writer.writeEndArray();
         break;

      case FINAL:
         elements = 0;
         while (elements < maxElements && !Double.isNaN(getCoVMean(elements))) {
            elements++;
         }
         writer.writeStartArray(ResultKeywords.COV_LIST.toString());
         for (int i = 0; i < elements; i++) {
            writer.writeStartArrayItem(i);
            final int blockSize = (int) Math.pow(minBlockSize, i);
            writer.writeDouble(ResultKeywords.BLOCKSIZE, blockSize);
            writer.writeDouble(ResultKeywords.COV_MEAN, getCoVMean(i));
            writer.writeDouble(ResultKeywords.COV_CINT, getCoVConfInterval(i));
            writer.writeEndArrayItem();
         }
         writer.writeEndArray();

         elements = 0;
         while (elements < maxElements && !Double.isNaN(getFIMean(elements))) {
            elements++;
         }
         writer.writeStartArray(ResultKeywords.FI_LIST.toString());
         for (int i = 0; i < elements; i++) {
            writer.writeStartArrayItem(i);
            final int blockSize = (int) Math.pow(minBlockSize, i);
            writer.writeDouble(ResultKeywords.BLOCKSIZE, blockSize);
            writer.writeDouble(ResultKeywords.FI_MEAN, getFIMean(i));
            writer.writeDouble(ResultKeywords.FI_CINT, getFIConfInterval(i));
            writer.writeEndArrayItem();
         }
         writer.writeEndArray();
         break;
      }
   }

   @Override
   public void exportBatch(ResultWriter writer) throws PrintException {
      final int length = ldMaxBlockSize - ldMinBlockSize;

      writer.writeStartArray(LIST);
      for (int i = 0; i <= length; i++) {
         writer.writeStartArrayItem(i);

         writer.writeDouble(BLOCK_COV_SUM, blockCoVSums[i]);
         writer.writeDouble(BLOCK_FI_SUM, blockFISums[i]);
         writer.writeLong(NUMBER_OF_BLOCK, noOfBlocks[i]);

         writer.writeEndArrayItem();
      }
      writer.writeEndArray();
   }

   @Override
   public void importBatch(ResultReader reader) throws ImportException {
      final int length = ldMaxBlockSize - ldMinBlockSize;

      reader.readStartArray(LIST);
      for (int i = 0; i <= length; i++) {
         reader.readStartArrayItem(i);

         blockCoVSums[i] = reader.readDouble(BLOCK_COV_SUM);
         blockFISums[i] = reader.readDouble(BLOCK_FI_SUM);
         noOfBlocks[i] = reader.readLong(NUMBER_OF_BLOCK);

         reader.readEndArrayItem();
      }
      reader.readEndArray();
   }

   @Override
   public void handleStopTransientPhase() {
      super.handleStopTransientPhase();

      if (traceFile == null && !traceFileName.equals("")) {
         // FIXME: OMG what a hack
         if (traceFileName.equals("__cout")) {
            traceFile = new PrintWriter(System.out, true);
         } else {
            try {
               traceFile = new PrintWriter(traceFileName);
            } catch (FileNotFoundException fnfe) {
               throw new SimulationException("Trace file \"" + traceFileName + "\"not found");
            }
         }
      }
   }

   @Override
   public void computeMeasures(int batchNumber) {
      for (int i = 0; i <= ldMaxBlockSize - ldMinBlockSize && noOfBlocks[i] > 0; i++) {

         if (!Double.isNaN(blockCoVSums[i])) {
            covBatchMeans[i].update(blockCoVSums[i] / noOfBlocks[i]);
            fiBatchMeans[i].update(blockFISums[i] / noOfBlocks[i]);
         }
      }
   }

   @Override
   public void resetBatchStatistic() {
      for (int i = 0; i <= ldMaxBlockSize - ldMinBlockSize; i++) {
         noOfBlocks[i] = 0;
         blockCoVSums[i] = 0;
         blockFISums[i] = 0;
      }
   }

   @Override
   public void resetStatistic() {
      resetBatchStatistic();
      for (int i = 0; i <= ldMaxBlockSize - ldMinBlockSize; i++) {
         covBatchMeans[i].reset();
         fiBatchMeans[i].reset();
      }
   }

   public double getCoVMean(int ldBlockSize) {
      if (ldBlockSize > ldMaxBlockSize - ldMinBlockSize) {
         throw new SimulationException("block size out of range");
      }

      return covBatchMeans[ldBlockSize].getMean();
   }

   public double getFIMean(int ldBlockSize) {
      if (ldBlockSize > ldMaxBlockSize - ldMinBlockSize) {
         throw new SimulationException("block size out of range");
      }

      return fiBatchMeans[ldBlockSize].getMean();
   }

   public double getCoVConfInterval(int ldBlockSize) {
      if (ldBlockSize > ldMaxBlockSize - ldMinBlockSize) {
         throw new SimulationException("block size out of range");
      }

      return covBatchMeans[ldBlockSize].getConfidenceInterval();
   }

   public double getFIConfInterval(int ldBlockSize) {
      if (ldBlockSize > ldMaxBlockSize - ldMinBlockSize) {
         throw new SimulationException("block size out of range");
      }

      return fiBatchMeans[ldBlockSize].getConfidenceInterval();
   }

   public double getBatchCoV(int ldBlockSize) {
      if (ldBlockSize > ldMaxBlockSize - ldMinBlockSize) {
         throw new SimulationException("block size out of range");
      }

      if (noOfBlocks[ldBlockSize] > 0) {
         return blockCoVSums[ldBlockSize] / noOfBlocks[ldBlockSize];
      } else {
         return Double.NaN;
      }
   }
}
