/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.untested.statistics.variancetime;

import ikr.simlib.events.time.Duration;
import ikr.simlib.events.time.PointInTime;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.parameters.reflection.ReflectionConstructable;

/**
 * <code>TVTRateStatistic</code> provides an implementation to estimate the VT
 * plot of rates.
 */
public class VTRateStatistic extends VTStatistic {

   private static final String LD_MIN_BLOCK_SIZE = "LdMinBlockSize";
   private static final String LD_MAX_BLOCK_SIZE = "LdMaxBlockSize";
   private static final String TRACE_FILE = "TraceFile";
   private static final String TIME_BASE = "TimeBase";

   /** as required by {@link ReflectionConstructable} */
   public static VTRateStatistic createInstance(SimNode ownNode, Parameters pars) {
      final int ldMinBlockSize = pars.get(ownNode, LD_MIN_BLOCK_SIZE).asInteger();
      final int ldMaxBlockSize = pars.get(ownNode, LD_MAX_BLOCK_SIZE).asInteger();
      final String traceFileName = pars.get(ownNode, TRACE_FILE).asString();
      final Duration timeBase = pars.get(ownNode, TIME_BASE).asInstance(Duration.class, ownNode.createChildNode(TIME_BASE), pars);
      return new VTRateStatistic(timeBase, ldMinBlockSize, ldMaxBlockSize, traceFileName, ownNode);
   }

   protected final Duration timeBase;

   protected Duration nextEndOfInterval;
   protected int noOfIntervals;
   protected final double[] blockSums;

   public VTRateStatistic(Duration timeBase, int ldMinBlockSize, int ldMaxBlockSize, String traceFileName, SimNode ownNode) {
      super(ldMinBlockSize, ldMaxBlockSize, traceFileName, ownNode);

      this.timeBase = timeBase;
      nextEndOfInterval = timeBase.multiplyBy(minBlockSize);
      noOfIntervals = 0;
      blockSums = new double[ldMaxBlockSize - ldMinBlockSize + 1];
   }

   public void update(PointInTime currentTime, double weight) {
      int i, blockSize;

      blockSums[0] += weight;
      while (currentTime.isAfter(PointInTime.ZERO.plus(nextEndOfInterval))) {
         noOfIntervals += minBlockSize;
         for (i = 0, blockSize = minBlockSize; (i <= ldMaxBlockSize - ldMinBlockSize) && (noOfIntervals % blockSize == 0); i++, blockSize *= 2) {
            double blockMean = blockSums[i] / timeBase.multiplyBy(blockSize).toSeconds();
            blockMeansSummations[i].update(blockMean);
            if (traceFile != null) {
               traceFile.printf("VTBlock%d %g %g%n", i + ldMinBlockSize, currentTime, blockMean);
            }
            if (i < ldMaxBlockSize - ldMinBlockSize) {
               blockSums[i + 1] += blockSums[i];
            }
            blockSums[i] = 0;
         }
         nextEndOfInterval = nextEndOfInterval.plus(timeBase.multiplyBy(minBlockSize));
      }
   }

   public final void update(PointInTime currentTime) {
      update(currentTime, 1);
   }
}
